#include "opengl.h"
#include "wl_def.h"
#include "id_us.h"
#include "console.h"

static char con_lines[32][45];	// buffer
static int con_active;			// console is visible
static int con_current;			// current line

static void Con_ProcessCommand();
static void Con_Print(char* text);

static void Con_Version();

command_t con_commands[] = {
	{ "VERSION", Con_Version },
	{ "", 0 }
};

cvar_t con_vars[] = {
	{ "R_ALPHAREF", cvar_float, &r_alpharef },
	{ "", cvar_float, 0 }
};

// ========================================================

static void Con_Version()
{
//	Con_Print("Version %4.2f\n", VERSION);
	Con_Print("Compiled: "__TIME__" "__DATE__);
}

static void Con_Print(char* text)
{
	int i;

	for (i = 31; i > 1; i--)
		strcpy(con_lines[i], con_lines[i-1]);
	strcpy(con_lines[1], text);
}

static void Con_ProcessCommand()
{
	int i, argc = 0;
	char args[2][64], *ch;
	command_t* cmd = con_commands;
	cvar_t* var = con_vars;

	ch = &con_lines[0][2];

	for (i = 2; i < 45; i++)
	{
		if (con_lines[0][i] == 0 || con_lines[0][i] == ' ')
		{
			strcpy(args[argc], ch);
			args[argc][&con_lines[0][i]-ch] = 0;
			argc++;
			ch = &con_lines[0][i]+1;
		}

		if (argc == 2 || (con_lines[0][i] == 0))
			break;
	}

	if (argc == 0)
		return;

	for (i = 31; i > 0; i--)
		strcpy(con_lines[i], con_lines[i-1]);

	while (cmd->func)
	{
		if (strcmp(cmd->name, args[0]) == 0)
		{
			cmd->func();
			return;
		}
		cmd++;
	}

	while (var->value)
	{
		if (strcmp(var->name, args[0]) == 0)
		{
			if (argc == 1)
			{
				char str[45];

				switch (var->type)
				{
				case cvar_float:
					sprintf(str, "\"%s\" IS %f", var->name, *(float*)var->value);
				break;

				case cvar_int:
					sprintf(str, "\"%s\" IS %d", var->name, *(int*)var->value);
				break;
				}
				Con_Print(str);
				return;
			}
			else
			{
				switch (var->type)
				{
				case cvar_float:
					{
						float tmp;

						if (sscanf(args[1], "%f", &tmp) == 1)
						{
							char str[45];

							*(float*)var->value = tmp;
							sprintf(str, "\"%s\" SET TO %f", var->name, tmp);
							Con_Print(str);
							return;
						}
					} break;

				case cvar_int:
					{
						int tmp;

						if (sscanf(args[1], "%d", &tmp) == 1)
						{
							char str[45];

							*(int*)var->value = tmp;
							sprintf(str, "\"%s\" SET TO %d", var->name, tmp);
							Con_Print(str);
							return;
						}
					} break;
				}

			}
		}
		var++;
	}

	Con_Print("Unknown Command");
}

// ========================================================

void Con_Init()
{
	con_current = 0;
	con_active = 0;
	LoadFonts();
	memset(con_lines, 0, sizeof(con_lines));
	strcpy(con_lines[0], "] ");
	Con_Print("WolfGL Console");
}

void Con_Draw()
{
	int i;

	if (con_active == 0)
		return;

	fontnumber = 0;
	SETFONTCOLOR(TEXTCOLOR, BKGDCOLOR);
	GL_Bar(0,0,320,100, 0);

	for (i = 0; i < 10; i++)
	{
		PrintX = 4;
		PrintY = (9-i)*10;
		US_Print(con_lines[i]);
	}
}

int Con_Key(int key)
{
	static wbool shift_down = false;
	byte c;
	int i, len;

	if (key == sc_Tilde)
	{
		con_active = !con_active;
		return 1;
	}

	if (!con_active)
		return 0;

	if (key & 0x80)
	{
		key &= 0x7f;

		if ((key == sc_LShift) || (key == sc_RShift))
			shift_down = false;
		return 1;
	}
	else
	{
		if ((key == sc_LShift) || (key == sc_RShift))
			shift_down = true;
	}

	if (key == sc_Enter)
	{
		Con_ProcessCommand();
		strcpy(con_lines[0], "] ");
		con_current = 0;
		return 1;
	}

	if (shift_down)
		c = ShiftNames[key];
	else
		c = ASCIINames[key];

	if ((c >= 'a') && (c <= 'z'))
		c -= 'a' - 'A';

	if (key == sc_LeftArrow)
	{
		len = strlen(con_lines[0]);
		if (len > 2)
			con_lines[0][len-1] = 0;
	}
	else if (key == sc_UpArrow)
	{
		for (i = 1; i < 32; i++)
			if (con_lines[i][0] == ']')
			{
				strcpy(con_lines[0], con_lines[i]);
				break;
			}
	}
	else if (key == sc_DownArrow)
	{
		strcpy(con_lines[0], "] ");
	}
	else if (c)
	{
		len = strlen(con_lines[0]);
		if (len < 44)
		{
			con_lines[0][len] = c;
			con_lines[0][len+1] = 0;
		}
	}

	return 1;
}
