
/*
=============================================================================

Id Software Caching Manager
---------------------------

Must be started BEFORE the memory manager, because it needs to get the headers
loaded into the data segment

=============================================================================
*/

#include <stdio.h>
#include <string.h>
#include "system.h"
#include "id_heads.h"
#include "sound.h"

#define THREEBYTEGRSTARTS

//	LOCAL CONSTANTS

// 0-255 is a character, > is a pointer to a node
typedef struct {
	unsigned long bit0, bit1;
} huffnode;

typedef struct {
	unsigned short RLEWtag;
	long headeroffsets[100];
	byte *tileinfo;
} mapfiletype;


//	GLOBAL VARIABLES

byte 		*tinf;
int			mapon;

unsigned short	*mapsegs[MAPPLANES];
maptype			*mapheaderseg[NUMMAPS];
byte			*audiosegs[NUMSNDCHUNKS];
void			*grsegs[NUMCHUNKS];

byte	grneeded[NUMCHUNKS];
byte	ca_levelbit, ca_levelnum;

FILE	*profilehandle;

char	audioname[13]="audio.";

//	LOCAL VARIABLES

// Need a string, not constant to change cache files
char extension[5];
char gheadname[10]="vgahead.";
char gfilename[10]="vgagraph.";
char gdictname[10]="vgadict.";
char mheadname[10]="maphead.";
char mfilename[10]="maptemp.";
char aheadname[10]="audiohed.";
char afilename[10]="audiot.";

void CA_CannotOpen(char *string);

long		*grstarts;
long		*audiostarts;

huffnode	grhuffman[255];

FILE* grhandle;
FILE* maphandle;
FILE* audiohandle;

long		chunkcomplen, chunkexplen;
SDMode		oldsoundmode;

void CAL_CarmackExpand (unsigned short *source, unsigned short *dest, unsigned short length);

#ifdef THREEBYTEGRSTARTS
#define FILEPOSSIZE	3
static long GRFILEPOS(int c)
{
	long value;
	int	offset;

	offset = c*3;

	value = *(long*)(((byte*)grstarts)+offset);

	value &= 0x00ffffffl;

	if (value == 0xffffffl)
		value = -1;

	return value;
};
#else
#define FILEPOSSIZE	4
#define	GRFILEPOS(c) (grstarts[c])
#endif

//	Gets the length of an explicit length chunk (not tiles)
//	The file pointer is positioned so the compressed data can be read in next.
void CAL_GetGrChunkLength (int chunk)
{
	fseek(grhandle, GRFILEPOS(chunk), SEEK_SET);
	fread(&chunkexplen, sizeof(chunkexplen), 1, grhandle);
	chunkcomplen = GRFILEPOS(chunk+1)-GRFILEPOS(chunk)-4;
}

//	Read from a file to a pointer
wbool CA_FarRead(FILE* handle, byte *dest, long length)
{
	if (length > 0xffff)
		Quit ("CA_FarRead doesn't support 64K reads yet!");

	if (fread(dest, length, 1, handle) != 1)
		return false;

	return true;
}

//	Write from a file to a pointer
wbool CA_FarWrite (FILE* handle, byte *source, long length)
{
	if (length>0xffffl)
		Quit ("CA_FarWrite doesn't support 64K reads yet!");

	if (fwrite(source, length, 1, handle) != 1)
		return false;

	return true;
}

//	Writes a file from a memory buffer
wbool CA_WriteFile (char *filename, void *ptr, long length)
{
	FILE* handle;

	handle = fopen(filename, "wb");

	if (handle == 0)
		return false;

	if (!CA_FarWrite(handle, ptr, length))
	{
		fclose(handle);
		return false;
	}
	fclose(handle);
	return true;
}

//	Allocate space for and load a file
wbool CA_LoadFile (char *filename, memptr *ptr)
{
	FILE* handle;
	long size;

	if ((handle = fopen(filename, "rb")) == 0)
		return false;

	size = S_FileLength(handle);
	MM_GetPtr(ptr,size);
	if (!CA_FarRead(handle, *ptr, size))
	{
		fclose(handle);
		return false;
	}

	fclose(handle);
	return true;
}

// ============================================================================
//	COMPRESSION routines, see JHUFF.C for more
// ============================================================================

//	Goes through a huffman table and changes the 256-511 node numbers to the
//	actual address of the node.  Must be called before CAL_HuffExpand
void CAL_OptimizeNodes (huffnode *table)
{
	huffnode *node;
	int i;

	node = table;

	for (i = 0; i < 255; i++)
	{
		if (node->bit0 >= 256)
			node->bit0 = (unsigned long)(table+(node->bit0-256));
		if (node->bit1 >= 256)
			node->bit1 = (unsigned long)(table+(node->bit1-256));
		node++;
	}
}

// Length is the length of the EXPANDED data. If screenhack,
// the data is decompressed in four planes directly to the screen
void CAL_HuffExpand (byte *source, byte *dest, long length, huffnode *hufftable, wbool screenhack)
{
	huffnode *headptr;
	byte* end;
	unsigned char ch, cl;
	huffnode *node;

	// head node is allways node 254
	headptr = hufftable+254;
	// normalize
	source++;
	source--;
	dest++;
	dest--;
	end = dest+length;

	node = headptr;
	ch = *source;
	source++;
	cl = 1;

	do
	{
		if ((ch & cl) != 0)
			node = (huffnode*)node->bit1;
		else
			node = (huffnode*)node->bit0;

		if ((int)node < 256)
		{
			*dest = (byte)((int)node);
			dest++;
			node = headptr;
		}

		cl <<= 1;
		if (cl == 0)
		{
			ch = *source;
			source++;
			cl = 1;
		}
	} while (dest != end);
}

#define NEARTAG	0xa7
#define FARTAG	0xa8

//	Length is the length of the EXPANDED data
void CAL_CarmackExpand (unsigned short *source, unsigned short *dest, unsigned short length)
{
	unsigned short ch,chhigh,count,offset;
	unsigned short *copyptr, *inptr, *outptr;

	length/=2;

	inptr = source;
	outptr = dest;

	while (length)
	{
		ch = *inptr++;
		chhigh = ch>>8;
		if (chhigh == NEARTAG)
		{
			count = ch&0xff;
			if (!count)
			{
				// have to insert a word containing the tag byte
				ch |= *((unsigned char*)inptr)++;
				*outptr++ = ch;
				length--;
			}
			else
			{
				offset = *((unsigned char*)inptr)++;
				copyptr = outptr - offset;
				length -= count;
				while (count--)
					*outptr++ = *copyptr++;
			}
		}
		else if (chhigh == FARTAG)
		{
			count = ch&0xff;
			if (!count)
			{
				// have to insert a word containing the tag byte
				ch |= *((unsigned char*)inptr)++;
				*outptr++ = ch;
				length --;
			}
			else
			{
				offset = *inptr++;
				copyptr = dest + offset;
				length -= count;
				while (count--)
					*outptr++ = *copyptr++;
			}
		}
		else
		{
			*outptr++ = ch;
			length --;
		}
	}
}

//	length is EXPANDED length
void CA_RLEWexpand (unsigned short *source, unsigned short *dest, long length, unsigned short rlewtag)
{
	unsigned short value, count, i;
	unsigned short *end;

	end = dest + (length)/2;

	// expand it
	do
	{
		value = *source++;
		if (value != rlewtag)
			// uncompressed
			*dest++=value;
		else
		{
			// compressed string
			count = *source++;
			value = *source++;
			for (i=1;i<=count;i++)
				*dest++ = value;
		}
	} while (dest < end);
}

//=============================================================================
//					 CACHE MANAGER ROUTINES
//=============================================================================

void CAL_SetupGrFile()
{
	char fname[13];
	FILE* handle;
	memptr compseg;
	int i;
	unsigned short sh;

	// load ???dict.ext (huffman dictionary for graphics files)
	strcpy(fname,gdictname);
	strcat(fname,extension);

	if ((handle = fopen(fname, "rb")) == 0)
		CA_CannotOpen(fname);

	for (i = 0; i < 255; i++)
	{
		fread(&sh, 2, 1, handle);
		grhuffman[i].bit0 = sh;
		fread(&sh, 2, 1, handle);
		grhuffman[i].bit1 = sh;
	}

	fclose(handle);
	CAL_OptimizeNodes (grhuffman);

	// load the data offsets from ???head.ext
	MM_GetPtr ((memptr)&grstarts,(NUMCHUNKS+1)*FILEPOSSIZE);

	strcpy(fname,gheadname);
	strcat(fname,extension);

	if ((handle = fopen(fname, "rb")) == 0)
		CA_CannotOpen(fname);

	CA_FarRead(handle, (memptr)grstarts, (NUMCHUNKS+1)*FILEPOSSIZE);

	fclose(handle);


	// Open the graphics file, leaving it open until the game is finished
	strcpy(fname,gfilename);
	strcat(fname,extension);

	grhandle = fopen(fname, "rb");
	if (grhandle == 0)
		CA_CannotOpen(fname);

	// load the pic and sprite headers into the arrays in the data segment
	MM_GetPtr((memptr)&pictable,NUMPICS*sizeof(pictabletype));
	// position file pointer
	CAL_GetGrChunkLength(STRUCTPIC);
	MM_GetPtr(&compseg,chunkcomplen);
	CA_FarRead (grhandle,compseg,chunkcomplen);
	CAL_HuffExpand (compseg, (byte*)pictable,NUMPICS*sizeof(pictabletype),grhuffman,false);
	MM_FreePtr(&compseg);
}

//==========================================================================

void CAL_SetupMapFile()
{
	int	i;
	FILE* handle;
	long length,pos;
	char fname[13];

	// load maphead.ext (offsets and tileinfo for map file)
	strcpy(fname,mheadname);
	strcat(fname,extension);

	if ((handle = fopen(fname, "rb")) == 0)
		CA_CannotOpen(fname);

	length = S_FileLength(handle);
	MM_GetPtr ((memptr)&tinf,length);
	CA_FarRead(handle, tinf, length);
	fclose(handle);

	// open the data file
#ifdef CARMACIZED
	strcpy(fname,"gamemaps.");
	strcat(fname,extension);

	if ((maphandle = fopen(fname, "rb")) == 0)
		CA_CannotOpen(fname);
#else
	strcpy(fname,mfilename);
	strcat(fname,extension);

	if ((maphandle = fopen(fname, "rb")) == 0)
		CA_CannotOpen(fname);
#endif

	// load all map header
	for (i=0;i<NUMMAPS;i++)
	{
		pos = *((long*)(tinf+2+i*4));

//		pos = ((mapfiletype*)tinf)->headeroffsets[i];
		// $FFFFFFFF start is a sparse map
		if (pos < 0)
			continue;

		MM_GetPtr((memptr)&mapheaderseg[i],sizeof(maptype));
		MM_SetLock((memptr)&mapheaderseg[i],true);
		fseek(maphandle, pos, SEEK_SET);

//		CA_FarRead(maphandle, (memptr)mapheaderseg[i], sizeof(maptype));
		fread(mapheaderseg[i]->planestart, 12, 1, maphandle);
		fread(mapheaderseg[i]->planelength, 6, 1, maphandle);
		fread(&mapheaderseg[i]->width, 2, 1, maphandle);
		fread(&mapheaderseg[i]->height, 2, 1, maphandle);
		fread(mapheaderseg[i]->name, 16, 1, maphandle);
	}

	// allocate space for 3 64*64 planes
	for (i=0;i<MAPPLANES;i++)
	{
		MM_GetPtr ((memptr)&mapsegs[i],64*64*2);
		MM_SetLock ((memptr)&mapsegs[i],true);
	}
}

//==========================================================================

void CAL_SetupAudioFile()
{
	FILE* handle;
	long length;
	char fname[13];

	// load maphead.ext (offsets and tileinfo for map file)
	strcpy(fname, aheadname);
	strcat(fname, extension);

	if ((handle = fopen(fname, "rb")) == 0)
		CA_CannotOpen(fname);

	length = S_FileLength(handle);
	MM_GetPtr ((memptr)&audiostarts, length);
	CA_FarRead(handle, (byte*)audiostarts, length);
	fclose(handle);

	// open the data file
	strcpy(fname,afilename);
	strcat(fname,extension);

	if ((audiohandle = fopen(fname, "rb")) == 0)
		CA_CannotOpen(fname);
}

//==========================================================================

//	Open all files and load in headers
void CA_Startup()
{
#ifdef PROFILE
	remove("PROFILE.TXT");
	profilehandle = fopen("PROFILE.TXT", "wt");
#endif

	CAL_SetupMapFile();
	CAL_SetupGrFile();
	CAL_SetupAudioFile();

	mapon = -1;
	ca_levelbit = 1;
	ca_levelnum = 0;
}

//==========================================================================

//	Closes all files
void CA_Shutdown()
{
#ifdef PROFILE
	if (profilehandle)
		fclose(profilehandle);
#endif

	if (maphandle)
		fclose(maphandle);
	if (grhandle)
		fclose(grhandle);
	if (audiohandle)
		fclose(audiohandle);
}

//===========================================================================

void CA_CacheAudioChunk (int chunk)
{
	long	pos,compressed;

	if (audiosegs[chunk])
	{
		// allready in memory
		MM_SetPurge ((memptr)&audiosegs[chunk],0);
		return;
	}

	// load the chunk into a buffer, either the miscbuffer if it fits, or allocate
	// a larger buffer
	pos = audiostarts[chunk];
	compressed = audiostarts[chunk+1]-pos;

	fseek(audiohandle, pos, SEEK_SET);

	MM_GetPtr ((memptr)&audiosegs[chunk],compressed);
	if (mmerror)
		return;

	CA_FarRead(audiohandle, audiosegs[chunk], compressed);
}

//===========================================================================

//	Purges all sounds, then loads all new ones (mode switch)
void CA_LoadAllSounds()
{
	unsigned short start = 0, i;

	switch (oldsoundmode)
	{
	case sdm_Off:
		goto cachein;
	case sdm_PC:
		start = STARTPCSOUNDS;
		break;
	case sdm_AdLib:
		start = STARTADLIBSOUNDS;
		break;
	}

	// make purgable
	for (i=0;i<NUMSOUNDS;i++,start++)
		if (audiosegs[start])
			MM_SetPurge ((memptr)&audiosegs[start],3);

cachein:

	switch (SoundMode)
	{
	case sdm_Off:
		return;
	case sdm_PC:
		start = STARTPCSOUNDS;
		break;
	case sdm_AdLib:
		start = STARTADLIBSOUNDS;
		break;
	}

	for (i=0;i<NUMSOUNDS;i++,start++)
		CA_CacheAudioChunk (start);

//	for (i=0;i<NUMSOUNDS;i++)
//		CacheSound(i);

	oldsoundmode = SoundMode;
}

//===========================================================================

//	Does whatever is needed with a pointer to a compressed chunk
void CAL_ExpandGrChunk(int chunk, byte *source)
{
	long expanded;

	if (chunk >= STARTTILE8 && chunk < STARTEXTERNS)
	{
		// expanded sizes of tile8/16/32 are implicit

#define BLOCK		64
#define MASKBLOCK	128

	        // tile 8s are all in one chunk!
		if (chunk<STARTTILE8M)
			expanded = BLOCK*NUMTILE8;
		else if (chunk<STARTTILE16)
			expanded = MASKBLOCK*NUMTILE8M;
		// all other tiles are one/chunk
		else if (chunk<STARTTILE16M)
			expanded = BLOCK*4;
		else if (chunk<STARTTILE32)
			expanded = MASKBLOCK*4;
		else if (chunk<STARTTILE32M)
			expanded = BLOCK*16;
		else
			expanded = MASKBLOCK*16;
	}
	else
	{
		// everything else has an explicit size longword
		expanded = *(long*)source;
		source += 4;
	}

	// allocate final space, decompress it, and free bigbuffer
	// Sprites need to have shifts made and various other junk
	MM_GetPtr (&grsegs[chunk],expanded);
	if (mmerror)
		return;
	CAL_HuffExpand (source,grsegs[chunk],expanded,grhuffman,false);
}

//	Makes sure a given chunk is in memory, loading it if needed
void CA_CacheGrChunk (int chunk)
{
	long	pos,compressed;
	memptr	bigbufferseg;
	byte	*source;
	int		next;

	// make sure it doesn't get removed
	grneeded[chunk] |= ca_levelbit;
	if (grsegs[chunk])
	{
		// allready in memory
		MM_SetPurge (&grsegs[chunk],0);
		return;
	}

	// load the chunk into a buffer, either the miscbuffer if it fits, or allocate
	// a larger buffer
	pos = GRFILEPOS(chunk);
	
	// $FFFFFFFF start is a sparse tile
	if (pos < 0)
	  return;

	next = chunk +1;
	// skip past any sparse tiles
	while (GRFILEPOS(next) == -1)
		next++;

	compressed = GRFILEPOS(next)-pos;

	fseek(grhandle,pos,SEEK_SET);

	if (compressed<=BUFFERSIZE)
	{
		CA_FarRead(grhandle,bufferseg,compressed);
		source = bufferseg;
	}
	else
	{
		MM_GetPtr(&bigbufferseg,compressed);
		MM_SetLock (&bigbufferseg,true);
		CA_FarRead(grhandle,bigbufferseg,compressed);
		source = bigbufferseg;
	}

	CAL_ExpandGrChunk (chunk,source);

	if (compressed>BUFFERSIZE)
		MM_FreePtr(&bigbufferseg);
}

//==========================================================================

//	Decompresses a chunk from disk straight onto the screen
void CA_CacheScreen(int chunk)
{
	long	pos, compressed, expanded;
	memptr	bigbufferseg;
	byte	*source;
	int		next;

	// load the chunk into a buffer
	pos = GRFILEPOS(chunk);
	next = chunk +1;
	while (GRFILEPOS(next) == -1)		// skip past any sparse tiles
		next++;
	compressed = GRFILEPOS(next)-pos;

	fseek(grhandle, pos, SEEK_SET);

	MM_GetPtr(&bigbufferseg,compressed);
	MM_SetLock (&bigbufferseg,true);
	CA_FarRead(grhandle,bigbufferseg,compressed);
	source = bigbufferseg;

	expanded = *(long*)source;
	source += 4;			// skip over length
/*
	// allocate final space, decompress it, and free bigbuffer
	// Sprites need to have shifts made and various other junk
	CAL_HuffExpand (source,MK_FP(SCREENSEG,bufferofs),expanded,grhuffman,true);
	VW_MarkUpdateBlock (0,0,319,199);
*/	MM_FreePtr(&bigbufferseg);
}

//==========================================================================

//	WOLF: This is specialized for a 64*64 map size
void CA_CacheMap (int mapnum)
{
	long	pos,compressed;
	int		plane;
	memptr	*dest,bigbufferseg;
	unsigned short size;
	unsigned short *source;
#ifdef CARMACIZED
	memptr	buffer2seg;
	long	expanded;
#endif

	mapon = mapnum;

	// load the planes into the allready allocated buffers
	size = 64*64*2;

	for (plane = 0; plane < MAPPLANES; plane++)
	{
		pos = mapheaderseg[mapnum]->planestart[plane];
		compressed = mapheaderseg[mapnum]->planelength[plane];

		dest = (memptr)&mapsegs[plane];

		fseek(maphandle, pos, SEEK_SET);
		if (compressed <= BUFFERSIZE)
			source = bufferseg;
		else
		{
			MM_GetPtr(&bigbufferseg,compressed);
			MM_SetLock (&bigbufferseg,true);
			source = bigbufferseg;
		}

		CA_FarRead(maphandle, (byte*)source, compressed);

#ifdef CARMACIZED
		// unhuffman, then unRLEW
		// The huffman'd chunk has a two byte expanded length first
		// The resulting RLEW chunk also does, even though it's not really
		// needed
		expanded = *source;
		source++;
		MM_GetPtr (&buffer2seg, expanded);
		CAL_CarmackExpand (source, (unsigned short*)buffer2seg, (unsigned short)expanded);
		CA_RLEWexpand (((unsigned short*)buffer2seg)+1,
			*dest, size, ((mapfiletype*)tinf)->RLEWtag);
		MM_FreePtr (&buffer2seg);
#else
		// unRLEW, skipping expanded length
		CA_RLEWexpand(source+1, *dest, size, ((mapfiletype*)tinf)->RLEWtag);
#endif

		if (compressed>BUFFERSIZE)
			MM_FreePtr(&bigbufferseg);
	}
}

//===========================================================================

//	Goes up a bit level in the needed lists and clears it out.
//	Everything is made purgable
void CA_UpLevel ()
{
	int	i;

	if (ca_levelnum==7)
		Quit ("CA_UpLevel: Up past level 7!");

	for (i=0;i<NUMCHUNKS;i++)
		if (grsegs[i])
			MM_SetPurge (&(memptr)grsegs[i],3);
	ca_levelbit<<=1;
	ca_levelnum++;
}

//===========================================================================

//	Goes down a bit level in the needed lists and recaches
//	everything from the lower level
void CA_DownLevel (void)
{
	if (!ca_levelnum)
		Quit ("CA_DownLevel: Down past level 0!");
	ca_levelbit>>=1;
	ca_levelnum--;
	CA_CacheMarks();
}

//===========================================================================

//	Clears out all the marks at the current level
void CA_ClearMarks (void)
{
	int i;

	for (i=0;i<NUMCHUNKS;i++)
		grneeded[i]&=~ca_levelbit;
}

//===========================================================================

//	Clears out all the marks on all the levels
void CA_ClearAllMarks (void)
{
	memset (grneeded,0,sizeof(grneeded));
	ca_levelbit = 1;
	ca_levelnum = 0;
}

//===========================================================================

//	Make everything possible purgable
void CA_SetAllPurge()
{
	int i;

	// free sounds
	for (i = 0; i < NUMSNDCHUNKS; i++)
		if (audiosegs[i])
			MM_SetPurge ((memptr)&audiosegs[i],3);

	// free graphics
	CA_ClearMarks();

	for (i = 0; i < NUMCHUNKS; i++)
		if (grsegs[i])
			MM_SetPurge (&(memptr)grsegs[i],3);
}

//===========================================================================

#define MAXEMPTYREAD	1024

void CA_CacheMarks()
{
	int 	i,next,numcache;
	long	pos,endpos,nextpos,nextendpos,compressed;
	long	bufferstart,bufferend;
	byte	*source;
	memptr	bigbufferseg;

	numcache = 0;

	// go through and make everything not needed purgable
	for (i=0;i<NUMCHUNKS;i++)
		if (grneeded[i]&ca_levelbit)
		{
		  // its allready in memory, make sure it stays there!
		  if (grsegs[i])
		    MM_SetPurge(&grsegs[i],0);
		  else
		    numcache++;
		}
		else
		{
		  // not needed, so make it purgeable
		  if (grsegs[i])
				MM_SetPurge(&grsegs[i],3);
		}

	// nothing to cache!
	if (!numcache)
		return;

	// go through and load in anything still needed
	bufferstart = bufferend = 0;

	for (i=0;i<NUMCHUNKS;i++)
		if ( (grneeded[i]&ca_levelbit) && !grsegs[i])
		{
			pos = GRFILEPOS(i);
			if (pos<0)
				continue;

			next = i +1;
			while (GRFILEPOS(next) == -1)
				next++;

			compressed = GRFILEPOS(next)-pos;
			endpos = pos+compressed;

			if (compressed<=BUFFERSIZE)
			{
				if (bufferstart<=pos
				&& bufferend>= endpos)
				{
					// data is allready in buffer
					source = (byte*)bufferseg+(pos-bufferstart);
				}
				else
				{
					// load buffer with a new block from disk
					// try to get as many of the needed blocks in as possible
					while ( next < NUMCHUNKS )
					{
						while (next < NUMCHUNKS &&
						!(grneeded[next]&ca_levelbit && !grsegs[next]))
							next++;
						if (next == NUMCHUNKS)
							continue;

						nextpos = GRFILEPOS(next);
						while (GRFILEPOS(++next) == -1)
							;
						nextendpos = GRFILEPOS(next);
						if (nextpos - endpos <= MAXEMPTYREAD
						&& nextendpos-pos <= BUFFERSIZE)
							endpos = nextendpos;
						else
							next = NUMCHUNKS;
					}

					fseek(grhandle, pos, SEEK_SET);
					CA_FarRead(grhandle,bufferseg,endpos-pos);
					bufferstart = pos;
					bufferend = endpos;
					source = bufferseg;
				}
			}
			else
			{
				// big chunk, allocate temporary buffer
				MM_GetPtr(&bigbufferseg,compressed);
				if (mmerror)
					return;
				MM_SetLock (&bigbufferseg,true);
				fseek(grhandle, pos, SEEK_SET);
				CA_FarRead(grhandle,bigbufferseg,compressed);
				source = bigbufferseg;
			}

			CAL_ExpandGrChunk (i,source);
			if (mmerror)
				return;

			if (compressed>BUFFERSIZE)
				MM_FreePtr(&bigbufferseg);

		}
}

void CA_CannotOpen(char *string)
{
	char str[30];

	strcpy(str,"Can't open ");
	strcat(str,string);
	strcat(str,"!\n");
	Quit(str);
}
