//
//	ID Engine
//	ID_SD.c - Sound Manager for Wolfenstein 3D
//	v1.2
//	By Jason Blochowiak
//

//
//	This module handles dealing with generating sound on the appropriate
//		hardware
//
//	Globals:
//		For User Mgr:
//			SoundBlasterPresent - SoundBlaster card present?
//			AdLibPresent - AdLib card present?
//			SoundMode - What device is used for sound effects
//				(Use SM_SetSoundMode() to set)
//			MusicMode - What device is used for music
//				(Use SM_SetMusicMode() to set)
//			DigiMode - What device is used for digitized sound effects
//				(Use SM_SetDigiDevice() to set)
//
//		For Cache Mgr:
//			NeedsDigitized - load digitized sounds?
//			NeedsMusic - load music?
//

#include <memory.h>
#include <malloc.h>
#include "id_heads.h"
#include "sound.h"

#ifdef	nil
#undef	nil
#endif
#define	nil	0

#define	SDL_SoundFinished()	{SoundNumber = SoundPriority = 0;}

//	Global variables
wbool		AdLibPresent;
wbool		SoundBlasterPresent,
			NeedsDigitized,
			NeedsMusic;
SDMode		SoundMode;
SMMode		MusicMode;
SDSMode		DigiMode;
longword	TicCount;
unsigned long	*SoundTable;
int			DigiMap[LASTSOUND];

//	Internal variables
static	wbool		SD_Started;
static	soundnames	SoundNumber;
		word		SoundPriority;

		int			LeftPosition, RightPosition;
		long		LocalTime;

		word		NumDigi;
		word		*DigiList;

/*
//	SoundBlaster variables
static	wbool					sbNoCheck,sbNoProCheck;
static	byte					sbOldIntMask = -1;
static	volatile byte			huge *sbNextSegPtr;
static	int						sbLocation = -1,sbInterrupt = 7,sbIntVec = 0xf,
								sbIntVectors[] = {-1,-1,0xa,0xb,-1,0xd,-1,0xf,-1,-1,-1};
static	volatile longword		sbNextSegLen;
static	volatile SampledSound	huge *sbSamples;
static	void interrupt			(*sbOldIntHand)(void);
static	byte					sbpOldFMMix,sbpOldVOCMix;

//	AdLib variables
byte			*alSound;
word			alBlock;
longword		alLengthLeft;
longword		alTimeCount;
Instrument		alZeroInst;

// This table maps channel numbers to carrier and modulator op cells
static	byte			carriers[9] =  { 3, 4, 5,11,12,13,19,20,21},
						modifiers[9] = { 0, 1, 2, 8, 9,10,16,17,18};
*/
//	Sequencer variables
wbool			sqActive;
/*static	word			alFXReg;
static	ActiveTrack		*tracks[sqMaxTracks],
						mytracks[sqMaxTracks];
static	word			sqMode,sqFadeStep;
		word			*sqHack, *sqHackPtr,sqHackLen,sqHackSeqLen;
		long			sqHackTime;
*/

/*
//
//	SoundBlaster code
//

///////////////////////////////////////////////////////////////////////////
//
//	SDL_SBPlaySeg() - Plays a chunk of sampled sound on the SoundBlaster
//	Insures that the chunk doesn't cross a bank boundary, programs the DMA
//	 controller, and tells the SB to start doing DMA requests for DAC
//
///////////////////////////////////////////////////////////////////////////
static longword
SDL_SBPlaySeg(volatile byte huge *data,longword length)
{
	unsigned		datapage;
	longword		dataofs,uselen;

	uselen = length;
	datapage = FP_SEG(data) >> 12;
	dataofs = ((FP_SEG(data) & 0xfff) << 4) + FP_OFF(data);
	if (dataofs >= 0x10000)
	{
		datapage++;
		dataofs -= 0x10000;
	}

	if (dataofs + uselen > 0x10000)
		uselen = 0x10000 - dataofs;

	uselen--;

	// Program the DMA controller
asm	pushf
asm	cli
	outportb(0x0a,sbDMA | 4);					// Mask off DMA on channel sbDMA
	outportb(0x0c,0);							// Clear byte ptr flip-flop to lower byte
	outportb(0x0b,0x49);						// Set transfer mode for D/A conv
	outportb(sbDMAa2,(byte)dataofs);			// Give LSB of address
	outportb(sbDMAa2,(byte)(dataofs >> 8));		// Give MSB of address
	outportb(sbDMAa1,(byte)datapage);			// Give page of address
	outportb(sbDMAa3,(byte)uselen);				// Give LSB of length
	outportb(sbDMAa3,(byte)(uselen >> 8));		// Give MSB of length
	outportb(0x0a,sbDMA);						// Re-enable DMA on channel sbDMA

	// Start playing the thing
	sbWriteDelay();
	sbOut(sbWriteCmd,0x14);
	sbWriteDelay();
	sbOut(sbWriteData,(byte)uselen);
	sbWriteDelay();
	sbOut(sbWriteData,(byte)(uselen >> 8));
asm	popf

	return(uselen + 1);
}

//	SDL_SBService() - Services the SoundBlaster DMA interrupt
static void interrupt
SDL_SBService(void)
{
	longword	used;

	sbIn(sbDataAvail);	// Ack interrupt to SB

	if (sbNextSegPtr)
	{
		used = SDL_SBPlaySeg(sbNextSegPtr,sbNextSegLen);
		if (sbNextSegLen <= used)
			sbNextSegPtr = nil;
		else
		{
			sbNextSegPtr += used;
			sbNextSegLen -= used;
		}
	}
	else
	{
		SDL_SBStopSample();
		SDL_DigitizedDone();
	}

	outportb(0x20,0x20);	// Ack interrupt
}

///////////////////////////////////////////////////////////////////////////
//
//	SDL_SBPlaySample() - Plays a sampled sound on the SoundBlaster. Sets up
//		DMA to play the sound
//
///////////////////////////////////////////////////////////////////////////
static void
SDL_SBPlaySample(byte huge *data,longword len)
{
	longword	used;

	SDL_SBStopSample();

asm	pushf
asm	cli

	used = SDL_SBPlaySeg(data,len);
	if (len <= used)
		sbNextSegPtr = nil;
	else
	{
		sbNextSegPtr = data + used;
		sbNextSegLen = len - used;
	}

	// Save old interrupt status and unmask ours
	sbOldIntMask = inportb(0x21);
	outportb(0x21,sbOldIntMask & ~(1 << sbInterrupt));

	sbWriteDelay();
	sbOut(sbWriteCmd,0xd4);						// Make sure DSP DMA is enabled

asm	popf
}

//	Sets the attenuation levels for the left and right
//	channels by using the mixer chip on the SB Pro. This hits a hole in
//	the address map for normal SBs.
static void SDL_PositionSBP(int leftpos,int rightpos)
{
	byte	v;

	leftpos = 15 - leftpos;
	rightpos = 15 - rightpos;
	v = ((leftpos & 0x0f) << 4) | (rightpos & 0x0f);

asm	pushf
asm	cli

	sbOut(sbpMixerAddr,sbpmVoiceVol);
	sbOut(sbpMixerData,v);

asm	popf
}
*/

//	Stuff for digitized sounds
memptr SDL_LoadDigiSegment(word page)
{
	memptr	addr;

	addr = PM_GetSoundPage(page);
	PM_SetPageLock(PMSoundStart + page,pml_Locked);

	return(addr);
}

void SDL_PlayDigiSegment(memptr addr,word len)
{
/*	switch (DigiMode)
	{
	case sds_SoundBlaster:
		SDL_SBPlaySample(addr,len);
		break;
	}
*/
}

void SD_StopDigitized()
{
/*
	switch (DigiMode)
	{
	case sds_SoundBlaster:
		SDL_SBStopSample();
		break;
	}
*/
}

void SD_Poll()
{
	S_UpdateSound();
}

void SD_SetPosition(int leftpos,int rightpos)
{
	if ((leftpos < 0) || (leftpos > 15)||
		(rightpos < 0) || (rightpos > 15) ||
		((leftpos == 15) && (rightpos == 15)))
		Quit("SD_SetPosition: Illegal position");
/*
	switch (DigiMode)
	{
	case sds_SoundBlaster:
		SDL_PositionSBP(leftpos,rightpos);
		break;
	}
*/
}

void CacheSound(word which)
{
	word DigiLeft, DigiPage, b;
	char* buf;
	memptr addr;
	long len;

	DigiPage = DigiList[(which * 2) + 0];
	DigiLeft = DigiList[(which * 2) + 1];

	b = 0;
	buf = malloc(DigiLeft);

	do 
	{
		len = (DigiLeft >= PMPageSize)? PMPageSize : (DigiLeft % PMPageSize);
		addr = SDL_LoadDigiSegment(DigiPage++);

		memcpy(buf+b, addr, len);
		DigiLeft -= len;
		b += len;
	} while (DigiLeft);

	S_LoadSound(which, buf, b);

	free(buf);
}

void SD_PlayDigitized(word which, int leftpos, int rightpos)
{
	if (!DigiMode)
		return;

	if (which >= NumDigi)
		Quit("SD_PlayDigitized: bad sound number");

	SD_SetPosition(leftpos, rightpos);

	CacheSound(which);

	S_PlaySound(which, -leftpos + rightpos);
}

void SD_SetDigiDevice(SDSMode mode)
{
	if (mode == DigiMode)
		return;

	SD_StopDigitized();

	if (((mode == sds_SoundBlaster) && SoundBlasterPresent) ||
		(mode == sds_Off))
		DigiMode = mode;
}

void SDL_SetupDigi()
{
	memptr	list;
	word	*p, pg;
	int		i;

	PM_UnlockMainMem();
	MM_GetPtr(&list,PMPageSize);
	PM_CheckMainMem();
	p = (word*)(PM_GetPage(ChunksInFile - 1));
	memcpy((void*)list, (void*)p, PMPageSize);
	pg = PMSoundStart;
	for (i = 0;i < PMPageSize / (sizeof(word) * 2);i++,p += 2)
	{
		if (pg >= ChunksInFile - 1)
			break;
		pg += (p[1] + (PMPageSize - 1)) / PMPageSize;
	}
	PM_UnlockMainMem();
	MM_GetPtr((memptr *)&DigiList,i * sizeof(word) * 2);
	memcpy((void*)DigiList,(void*)list,i * sizeof(word) * 2);
	MM_FreePtr(&list);
	NumDigi = i;

	for (i = 0;i < LASTSOUND;i++)
		DigiMap[i] = -1;
}

// 	AdLib Code
/*
///////////////////////////////////////////////////////////////////////////
//
//	SDL_ALStopSound() - Turns off any sound effects playing through the
//		AdLib card
//
///////////////////////////////////////////////////////////////////////////
static void
SDL_ALStopSound(void)
{
asm	pushf
asm	cli

	(long)alSound = 0;
	alOut(alFreqH + 0,0);

asm	popf
}

static void
SDL_AlSetFXInst(Instrument far *inst)
{
	byte		c,m;

	m = modifiers[0];
	c = carriers[0];
	alOut(m + alChar,inst->mChar);
	alOut(m + alScale,inst->mScale);
	alOut(m + alAttack,inst->mAttack);
	alOut(m + alSus,inst->mSus);
	alOut(m + alWave,inst->mWave);
	alOut(c + alChar,inst->cChar);
	alOut(c + alScale,inst->cScale);
	alOut(c + alAttack,inst->cAttack);
	alOut(c + alSus,inst->cSus);
	alOut(c + alWave,inst->cWave);

	// Note: Switch commenting on these lines for old MUSE compatibility
//	alOut(alFeedCon,inst->nConn);
	alOut(alFeedCon,0);
}

///////////////////////////////////////////////////////////////////////////
//
//	SDL_ALPlaySound() - Plays the specified sound on the AdLib card
//
///////////////////////////////////////////////////////////////////////////
static void
SDL_ALPlaySound(AdLibSound far *sound)
{
	Instrument	far *inst;
	byte		huge *data;

	SDL_ALStopSound();

asm	pushf
asm	cli

	alLengthLeft = sound->common.length;
	data = sound->data;
	data++;
	data--;
	alSound = (byte far *)data;
	alBlock = ((sound->block & 7) << 2) | 0x20;
	inst = &sound->inst;

	if (!(inst->mSus | inst->cSus))
	{
	asm	popf
		Quit("SDL_ALPlaySound() - Bad instrument");
	}

	SDL_AlSetFXInst(&alZeroInst);	// DEBUG
	SDL_AlSetFXInst(inst);

asm	popf
}

//	Shuts down the AdLib card for sound effects
static void SDL_ShutAL(void)
{
asm	pushf
asm	cli

	alOut(alEffects,0);
	alOut(alFreqH + 0,0);
	SDL_AlSetFXInst(&alZeroInst);
	alSound = 0;

asm	popf
}

//	Totally shuts down the AdLib card
static void SDL_CleanAL(void)
{
	int	i;

asm	pushf
asm	cli

	alOut(alEffects,0);
	for (i = 1;i < 0xf5;i++)
		alOut(i,0);

asm	popf
}

///////////////////////////////////////////////////////////////////////////
//
//	SDL_StartAL() - Starts up the AdLib card for sound effects
//
///////////////////////////////////////////////////////////////////////////
static void
SDL_StartAL(void)
{
	alFXReg = 0;
	alOut(alEffects,alFXReg);
	SDL_AlSetFXInst(&alZeroInst);
}

///////////////////////////////////////////////////////////////////////////
//
//	SDL_DetectAdLib() - Determines if there's an AdLib (or SoundBlaster
//		emulating an AdLib) present
//
///////////////////////////////////////////////////////////////////////////
static wbool
SDL_DetectAdLib(void)
{
	byte	status1,status2;
	int		i;

	alOut(4,0x60);	// Reset T1 & T2
	alOut(4,0x80);	// Reset IRQ
	status1 = readstat();
	alOut(2,0xff);	// Set timer 1
	alOut(4,0x21);	// Start timer 1
#if 0
	SDL_Delay(TimerDelay100);
#else
asm	mov	dx,0x388
asm	mov	cx,100
usecloop:
asm	in	al,dx
asm	loop usecloop
#endif

	status2 = readstat();
	alOut(4,0x60);
	alOut(4,0x80);

	if (((status1 & 0xe0) == 0x00) && ((status2 & 0xe0) == 0xc0))
	{
		for (i = 1;i <= 0xf5;i++)	// Zero all the registers
			alOut(i,0);

		alOut(1,0x20);	// Set WSE=1
		alOut(8,0);		// Set CSM=0 & SEL=0

		return(true);
	}
	else
		return(false);
}
*/

//	turns off whatever device was being used for sound fx
static void SDL_ShutDevice(void)
{
/*	switch (SoundMode)
	{
	case sdm_AdLib:
		SDL_ShutAL();
		break;
	}
*/	SoundMode = sdm_Off;
}

//	totally shuts down all sound devices
static void SDL_CleanDevice()
{
/*	if ((SoundMode == sdm_AdLib) || (MusicMode == smm_AdLib))
		SDL_CleanAL();
*/
}

//	turns on whatever device is to be used for sound fx
static void SDL_StartDevice(void)
{
/*	switch (SoundMode)
	{
	case sdm_AdLib:
		SDL_StartAL();
		break;
	}
*/	SoundNumber = SoundPriority = 0;
}

//	Public routines

//	Sets which sound hardware to use for sound effects
wbool SD_SetSoundMode(SDMode mode)
{
	wbool result = false;

	SD_StopSound();

	if ((mode == sdm_AdLib) && !AdLibPresent)
		mode = sdm_Off;

	switch (mode)
	{
	case sdm_Off:
		NeedsDigitized = false;
		result = true;
		break;
	case sdm_PC:
		break;
	case sdm_AdLib:
		if (AdLibPresent)
		{
			NeedsDigitized = false;
			result = true;
		}
		break;
	}

	if (result && (mode != SoundMode))
	{
		SDL_ShutDevice();
		SoundMode = mode;
		SDL_StartDevice();
	}
	else
		SoundTable = (unsigned long*)(&audiosegs[STARTADLIBSOUNDS]);

	return result;
}

//	sets the device to use for background music
wbool SD_SetMusicMode(SMMode mode)
{
	wbool result = false;

	SD_FadeOutMusic();
	while (SD_MusicPlaying())
		;

	if (mode == smm_Off)
	{
		NeedsMusic = false;
		result = true;
	}
	else
	{
/*	case smm_AdLib:
		if (AdLibPresent)
		{
			NeedsMusic = true;
			result = true;
		}
*/	}

	if (result)
		MusicMode = mode;

	return result;
}


//	starts up the Sound Mgr
//	Detects all additional sound hardware and installs my ISR
void SD_Startup()
{
	if (SD_Started)
		return;

	LocalTime = TicCount 
/*		= alTimeCount
*/		= 0;

	SD_SetSoundMode(sdm_Off);
	SD_SetMusicMode(smm_Off);

	SoundBlasterPresent = DigitalDevice;

/*
	AdLibPresent = SDL_DetectAdLib();
	if (AdLibPresent && !sbNoCheck)
	{
		int port = -1;
		char *env = getenv("BLASTER");
		if (env)
		{
			long temp;
			while (*env)
			{
				while (isspace(*env))
					env++;

				switch (toupper(*env))
				{
				case 'A':
					temp = strtol(env + 1,&env,16);
					if
					(
						(temp >= 0x210)
					&&	(temp <= 0x260)
					&&	(!(temp & 0x00f))
					)
							port = (temp - 0x200) >> 4;
						else
							Quit("SD_Startup: Unsupported address value in BLASTER");
						break;
					case 'I':
						temp = strtol(env + 1,&env,10);
						if
						(
							(temp >= 0)
						&&	(temp <= 10)
						&&	(sbIntVectors[temp] != -1)
						)
						{
							sbInterrupt = temp;
							sbIntVec = sbIntVectors[sbInterrupt];
						}
						else
							Quit("SD_Startup: Unsupported interrupt value in BLASTER");
						break;
					case 'D':
						temp = strtol(env + 1,&env,10);
						if ((temp == 0) || (temp == 1) || (temp == 3))
							SDL_SBSetDMA(temp);
						else
							Quit("SD_Startup: Unsupported DMA value in BLASTER");
						break;
					default:
						while (isspace(*env))
							env++;
						while (*env && !isspace(*env))
							env++;
						break;
					}
				}
			}
			SoundBlasterPresent = SDL_DetectSoundBlaster(port);
		}
	}

	if (SoundBlasterPresent)
		SDL_StartSB();
*/
	SDL_SetupDigi();

	SD_Started = true;
}

//	turns off whatever sound hardware was active
void SD_Shutdown()
{
	if (!SD_Started)
		return;

	SD_MusicOff();
	SD_StopSound();
	SDL_ShutDevice();
	SDL_CleanDevice();
/*
	if (SoundBlasterPresent)
		SDL_ShutSB();
*/
	SD_Started = false;
}

//	Sets up a stereo imaging location for the next
//	sound to be played. Each channel ranges from 0 to 15.
void SD_PositionSound(int leftvol, int rightvol)
{
	LeftPosition = leftvol;
	RightPosition = rightvol;
}

//	the specified sound on the appropriate hardware
wbool SD_PlaySound(soundnames sound)
{
	SoundCommon	*s;
	int	lp,rp;

	lp = LeftPosition;
	rp = RightPosition;
	LeftPosition = 0;
	RightPosition = 0;

	if (sound == -1)
		return false;

	s = (void*)((unsigned long)(SoundTable[sound]));
	if ((SoundMode != sdm_Off) && !s)
		Quit("SD_PlaySound() - Uncached sound");

	if ((DigiMode != sds_Off) && (DigiMap[sound] != -1))
	{
		SD_PlayDigitized(DigiMap[sound],lp,rp);

		return true;
	}

	if (SoundMode == sdm_Off)
		return false;
	if (!s->length)
		Quit("SD_PlaySound() - Zero length sound");
	if (s->priority < SoundPriority)
		return false;
/*
	switch (SoundMode)
	{
	case sdm_AdLib:
		SDL_ALPlaySound((void far *)s);
		break;
	}
*/
	SoundNumber = sound;
	SoundPriority = s->priority;

	return false;
}

//	returns the sound number that's playing, or 0 if
//	no sound is playing
word SD_SoundPlaying()
{
	wbool result = false;
/*
	switch (SoundMode)
	{
	case sdm_AdLib:
		result = alSound? true : false;
		break;
	}
*/
	if (result)
		return(SoundNumber);
	else
		return(false);
}

//	if a sound is playing, stops it
void SD_StopSound()
{
/*
	if (DigiPlaying)
		SD_StopDigitized();

	switch (SoundMode)
	{
	case sdm_AdLib:
		SDL_ALStopSound();
		break;
	}
*/
	SDL_SoundFinished();
}

//	waits until the current sound is done playing
void SD_WaitSoundDone()
{
	while (SD_SoundPlaying())
		;
}

//	turns on the sequencer
void SD_MusicOn()
{
	sqActive = true;
}

//	turns off the sequencer and any playing notes
void SD_MusicOff()
{
/*	word i;

	switch (MusicMode)
	{
	case smm_AdLib:
		alFXReg = 0;
		alOut(alEffects,0);
		for (i = 0;i < sqMaxTracks;i++)
			alOut(alFreqH + i + 1,0);
		break;
	}
*/
	sqActive = false;

}

//	starts playing the music pointed to
void SD_StartMusic(MusicGroup *music)
{
	SD_MusicOff();
/*
	if (MusicMode == smm_AdLib)
	{
		sqHackPtr = sqHack = music->values;
		sqHackSeqLen = sqHackLen = music->length;
		sqHackTime = 0;
		alTimeCount = 0;
		SD_MusicOn();
	}
*/
}

//	starts fading out the music. Call SD_MusicPlaying()
//	to see if the fadeout is complete
void SD_FadeOutMusic()
{
/*
	switch (MusicMode)
	{
	case smm_AdLib:
		// DEBUG - quick hack to turn the music off
		SD_MusicOff();
		break;
	}
*/
}

//	returns true if music is currently playing, false if not
wbool SD_MusicPlaying()
{
	wbool	result;

	switch (MusicMode)
	{
/*	case smm_AdLib:
		result = false;
		// DEBUG - not written
		break;
*/	default:
		result = false;
	}

	return(result);
}
