//
//	ID Engine
//	ID_US_1.c - User Manager - General routines
//	v1.1d1
//	By Jason Blochowiak
//	Hacked up for Catacomb 3D
//

//
//	This module handles dealing with user input & feedback
//
//	Depends on: Input Mgr, View Mgr, some variables from the Sound, Caching,
//		and Refresh Mgrs, Memory Mgr for background save/restore
//
//	Globals:
//		ingame - Flag set by game indicating if a game is in progress
//      abortgame - Flag set if the current game should be aborted (if a load
//			game fails)
//		loadedgame - Flag set if a game was loaded
//		abortprogram - Normally nil, this points to a terminal error message
//			if the program needs to abort
//		PrintX, PrintY - Where the User Mgr will print (global coords)
//		WindowX,WindowY,WindowW,WindowH - The dimensions of the current
//			window
//

#include <stdlib.h>
#include <string.h>
#include "id_heads.h"
#include "system.h"
#include "opengl.h"

extern int _argc;
extern char** _argv;

//	Global variables
char		*abortprogram;
wbool		NoWait;
word		PrintX, PrintY;
word		WindowX, WindowY, WindowW, WindowH;

//	Internal variables
static	char		*ParmStrings[] = {"TEDLEVEL","NOWAIT"};
static	wbool		US_Started;

void	(*USL_MeasureString)(char*, word*, word*) = VW_MeasurePropString;

HighScore	Scores[MaxScores] =
{
	{"id software-'92",10000,1},
	{"Adrian Carmack",10000,1},
	{"John Carmack",10000,1},
	{"Kevin Cloud",10000,1},
	{"Tom Hall",10000,1},
	{"John Romero",10000,1},
	{"Jay Wilbur",10000,1}
};
/*
//	Public routines

//	Handles the Abort/Retry/Fail sort of errors passed from DOS.
int USL_HardError(word errval,int ax,int bp,int si)
{
#define IGNORE  0
#define RETRY   1
#define	ABORT   2
extern	void	ShutdownId(void);

static	char		buf[32];
static	WindowRec	wr;
		int			di;
		char		c,*s,*t;


	di = _DI;

	if (ax < 0)
		s = "Device Error";
	else
	{
		if ((di & 0x00ff) == 0)
			s = "Drive ~ is Write Protected";
		else
			s = "Error on Drive ~";
		for (t = buf;*s;s++,t++)	// Can't use sprintf()
			if ((*t = *s) == '~')
				*t = (ax & 0x00ff) + 'A';
		*t = '\0';
		s = buf;
	}

	c = peekb(0x40,0x49);	// Get the current screen mode
	if ((c < 4) || (c == 7))
		goto oh_kill_me;

	// DEBUG - handle screen cleanup

	US_SaveWindow(&wr);
	US_CenterWindow(30,3);
	US_CPrint(s);
	US_CPrint("(R)etry or (A)bort?");
	VW_UpdateScreen();
	IN_ClearKeysDown();

asm	sti	// Let the keyboard interrupts come through

	while (true)
	{
		switch (IN_WaitForASCII())
		{
		case key_Escape:
		case 'a':
		case 'A':
			goto oh_kill_me;
			break;
		case key_Return:
		case key_Space:
		case 'r':
		case 'R':
			US_ClearWindow();
			VW_UpdateScreen();
			US_RestoreWindow(&wr);
			return(RETRY);
			break;
		}
	}

oh_kill_me:
	abortprogram = s;
	ShutdownId();
	fprintf(stderr,"Terminal Error: %s\n",s);
	if (tedlevel)
		fprintf(stderr,"You launched from TED. I suggest that you reboot...\n");

	return(ABORT);
#undef	IGNORE
#undef	RETRY
#undef	ABORT
}
*/

//	Starts the User Mgr
void US_Startup()
{
	int	i, n;

	if (US_Started)
		return;
/*
	harderr(USL_HardError);	// Install the fatal error handler
*/
	// Initialize the random number generator
	US_InitRndT(true);

	// Check for TED launching here
	for (i = 1;i < _argc;i++)
	{
		n = US_CheckParm(_argv[i],ParmStrings);
		switch(n)
		{
		 case 0:
		   tedlevelnum = atoi(_argv[i + 1]);
		   if (tedlevelnum >= 0)
		     tedlevel = true;
		   break;

		 case 1:
		   NoWait = true;
		   break;
		}
	}

// LZ
//tedlevelnum = 0;
//tedlevel = true;
//NoWait = true;

	US_Started = true;
}

//	Shuts down the User Mgr
void US_Shutdown()
{
	if (!US_Started)
		return;

	US_Started = false;
}

//	checks to see if a string matches one of a set of strings.
//	The check is case insensitive. The routine returns the index
//	of the string that matched, or -1 if no matches were found
int US_CheckParm(char *parm, char **strings)
{
	char cp, cs, *p, *s;
	int i;

	// Skip non-alphas
	while (!isalpha(*parm))
		parm++;

	for (i = 0;*strings && **strings;i++)
	{
		for (s = *strings++,p = parm,cs = cp = 0;cs == cp;)
		{
			cs = *s++;
			if (!cs)
				return(i);
			cp = *p++;

			if (isupper(cs))
				cs = tolower(cs);
			if (isupper(cp))
				cp = tolower(cp);
		}
	}
	return(-1);
}

//	Window/Printing routines

//	Prints a string in the current window. Newlines are supported.
void US_Print(char *str)
{
	char	c, *se, *s, tmp[128];
	word	w, h;

	strcpy(tmp, str);
	s = tmp;

	while (*s)
	{
		se = s;
		while ((c = *se) && (c != '\n'))
			se++;
		*se = 0;

		USL_MeasureString(s,&w,&h);
		px = PrintX;
		py = PrintY;
		GL_DrawPropString(s);

		s = se;
		if (c)
		{
			*se = c;
			s++;

			PrintX = WindowX;
			PrintY += h;
		}
		else
			PrintX += w;
	}
}

//	Prints an unsigned long
void US_PrintUnsigned(longword n)
{
  /*
	char buffer[32];

	US_Print(ultoa(n,buffer,10));
*/
}

/*
//	Prints a string in the center of the given rect
void USL_PrintInCenter(char *s, Rect r)
{
	word	w,h,
			rw,rh;

	USL_MeasureString(s,&w,&h);
	rw = r.lr.x - r.ul.x;
	rh = r.lr.y - r.ul.y;

	px = r.ul.x + ((rw - w) / 2);
	py = r.ul.y + ((rh - h) / 2);
	GL_DrawPropString(s);
}

//	Prints a string centered in the current window.
void US_PrintCentered(char *s)
{
	Rect	r;

	r.ul.x = WindowX;
	r.ul.y = WindowY;
	r.lr.x = r.ul.x + WindowW;
	r.lr.y = r.ul.y + WindowH;

	USL_PrintInCenter(s,r);
}
*/

//	Prints a string centered on the current line and
//	advances to the next line. Newlines are not supported.
void US_CPrintLine(char *s)
{
	word w, h;

	USL_MeasureString(s,&w,&h);

	if (w > WindowW)
		Quit("US_CPrintLine() - String exceeds width");
	px = WindowX + ((WindowW - w) / 2);
	py = PrintY;
	GL_DrawPropString(s);
	PrintY += h;
}

//	Prints a string in the current window. Newlines are supported.
void US_CPrint(char *str)
{
	char c, *se, *s, tmp[128];

	strcpy(tmp, str);
	s = tmp;

	while (*s)
	{
		se = s;
		while ((c = *se) && (c != '\n'))
			se++;
		*se = '\0';

		US_CPrintLine(s);

		s = se;
		if (c)
		{
			*se = c;
			s++;
		}
	}
}
/*
//	Clears the current window to white and homes the
//	cursor
void US_ClearWindow()
{
	VWB_Bar(WindowX,WindowY,WindowW,WindowH,WHITE);
	PrintX = WindowX;
	PrintY = WindowY;
}

//	Draws a frame and sets the current window parms
void US_DrawWindow(word x,word y,word w,word h)
{
	word i,sx,sy,sw,sh;

	WindowX = x * 8;
	WindowY = y * 8;
	WindowW = w * 8;
	WindowH = h * 8;

	PrintX = WindowX;
	PrintY = WindowY;

	sx = (x - 1) * 8;
	sy = (y - 1) * 8;
	sw = (w + 1) * 8;
	sh = (h + 1) * 8;

	US_ClearWindow();

	VWB_DrawTile8(sx,sy,0),VWB_DrawTile8(sx,sy + sh,5);
	for (i = sx + 8;i <= sx + sw - 8;i += 8)
		VWB_DrawTile8(i,sy,1),VWB_DrawTile8(i,sy + sh,6);
	VWB_DrawTile8(i,sy,2),VWB_DrawTile8(i,sy + sh,7);

	for (i = sy + 8;i <= sy + sh - 8;i += 8)
		VWB_DrawTile8(sx,i,3),VWB_DrawTile8(sx + sw,i,4);
}

//	Generates a window of a given width & height in the
//	middle of the screen
void US_CenterWindow(word w,word h)
{
	US_DrawWindow(((MaxX / 8) - w) / 2,((MaxY / 8) - h) / 2,w,h);
}

//	Saves the current window parms into a record for
//	later restoration
void US_SaveWindow(WindowRec *win)
{
	win->x = WindowX;
	win->y = WindowY;
	win->w = WindowW;
	win->h = WindowH;

	win->px = PrintX;
	win->py = PrintY;
}

//	Sets the current window parms to those held in the
//	record
void US_RestoreWindow(WindowRec *win)
{
	WindowX = win->x;
	WindowY = win->y;
	WindowW = win->w;
	WindowH = win->h;

	PrintX = win->px;
	PrintY = win->py;
}
*/
//	Input routines

//	XORs the I-bar text cursor. Used by US_LineInput()
static void USL_XORICursor(int x,int y,char *s,word cursor)
{
	static	wbool status;		// VGA doesn't XOR...
	char	buf[MaxString];
	int		temp;
	word	w, h;

	strcpy(buf,s);
	buf[cursor] = '\0';
	USL_MeasureString(buf,&w,&h);

	px = x + w - 1;
	py = y;
	if (status^=1)
		GL_DrawPropString("\x80");
	else
	{
		temp = fontcolor;
		fontcolor = backcolor;
		GL_DrawPropString("\x80");
		fontcolor = temp;
	}
}

//	Gets a line of user input at (x,y), the string defaults
//	to whatever is pointed at by def. Input is restricted to maxchars
//	chars or maxwidth pixels wide. If the user hits escape (and escok is
//	true), nothing is copied into buf, and false is returned. If the
//	user hits return, the current string is copied into buf, and true is
//	returned
wbool US_LineInput(int x,int y,char *buf,char *def,wbool escok, int maxchars,int maxwidth, void (*draw)())
{
	wbool		redraw,cursorvis,cursormoved,done,result = false;
	ScanCode	sc;
	char		temp,c,s[MaxString],olds[MaxString];
	word		i,cursor,w,h,len;
	longword	lasttime;

	if (def)
		strcpy(s,def);
	else
		*s = '\0';
	*olds = '\0';
	cursor = strlen(s);
	cursormoved = redraw = true;

	cursorvis = done = false;
	lasttime = TicCount;
	LastASCII = key_None;
	LastScan = sc_None;

	while (!done)
	{
		if (cursorvis)
			USL_XORICursor(x,y,s,cursor);

		S_ProcessEvents();
		TicCount += S_GetNewTics();
		sc = LastScan;
		LastScan = sc_None;
		c = LastASCII;
		LastASCII = key_None;

		switch (sc)
		{
		case sc_LeftArrow:
			if (cursor)
				cursor--;
			c = key_None;
			cursormoved = true;
			break;
		case sc_RightArrow:
			if (s[cursor])
				cursor++;
			c = key_None;
			cursormoved = true;
			break;
		case sc_Home:
			cursor = 0;
			c = key_None;
			cursormoved = true;
			break;
		case sc_End:
			cursor = strlen(s);
			c = key_None;
			cursormoved = true;
			break;

		case sc_Return:
			strcpy(buf,s);
			done = true;
			result = true;
			c = key_None;
			break;
		case sc_Escape:
			if (escok)
			{
				done = true;
				result = false;
			}
			c = key_None;
			break;

		case sc_BackSpace:
			if (cursor)
			{
				strcpy(s + cursor - 1,s + cursor);
				cursor--;
				redraw = true;
			}
			c = key_None;
			cursormoved = true;
			break;
		case sc_Delete:
			if (s[cursor])
			{
				strcpy(s + cursor,s + cursor + 1);
				redraw = true;
			}
			c = key_None;
			cursormoved = true;
			break;

		case 0x4c:	// Keypad 5
		case sc_UpArrow:
		case sc_DownArrow:
		case sc_PgUp:
		case sc_PgDn:
		case sc_Insert:
			c = key_None;
			break;
		}

		if (c)
		{
			len = strlen(s);
			USL_MeasureString(s,&w,&h);

			if (isprint(c)
			&&	(len < MaxString - 1)
			&&	((!maxchars) || (len < maxchars))
			&&	((!maxwidth) || (w < maxwidth)))
			{
				for (i = len + 1;i > cursor;i--)
					s[i] = s[i - 1];
				s[cursor++] = c;
				redraw = true;
			}
		}

		if (redraw)
		{
#ifdef SPEAR
#define BKGDCOLOR	0x9d
#else
#define BKGDCOLOR	0x2d
#endif
			draw();
			GL_Bar(x-1,y,175-24-16,9,BKGDCOLOR);

			px = x;
			py = y;
			fontnumber = 0;
			temp = fontcolor;
			fontcolor = backcolor;
			GL_DrawPropString(olds);
			fontcolor = temp;
			strcpy(olds,s);

			px = x;
			py = y;
			GL_DrawPropString(s);

			if (cursorvis)
				USL_XORICursor(x,y,s,cursor);

			S_FinishDisplay();

			redraw = false;
		}

		if (cursormoved)
		{
			cursorvis = false;
			lasttime = TicCount - TickBase;
			cursormoved = false;
		}

		if (TicCount - lasttime > TickBase / 2)
		{
			lasttime = TicCount;
			cursorvis ^= true;
			redraw = true;
		}

/*		VW_UpdateScreen();
*/	}

	if (cursorvis)
		USL_XORICursor(x,y,s,cursor);

	if (!result)
	{
		px = x;
		py = y;
		GL_DrawPropString(olds);
	}

/*	VW_UpdateScreen();
*/
	IN_ClearKeysDown();
	return result;
}

//	Assembly portion of the User Mgr. This is just John Carmack's table
//	driven pseudo-random number generator, and we put it in the User Mgr
//	because we couldn't figure out where it should go


//============================================================================
//                           RANDOM ROUTINES
//============================================================================


static unsigned char rndindex;

static unsigned char rndtable[256] = {
	  0,   8, 109, 220, 222, 241, 149, 107,  75, 248, 254, 140,  16,  66,
	 74,  21, 211,  47,  80, 242, 154,  27, 205, 128, 161,  89,  77,  36,
	 95, 110,  85,  48, 212, 140, 211, 249,  22,  79, 200,  50,  28, 188,
	 52, 140, 202, 120,  68, 145,  62,  70, 184, 190,  91, 197, 152, 224,
	149, 104,  25, 178, 252, 182, 202, 182, 141, 197,   4,  81, 181, 242,
	145,  42,  39, 227, 156, 198, 225, 193, 219,  93, 122, 175, 249,   0,
	175, 143,  70, 239,  46, 246, 163,  53, 163, 109, 168, 135,   2, 235,
	 25,  92,  20, 145, 138,  77,  69, 166,  78, 176, 173, 212, 166, 113,
	 94, 161,  41,  50, 239,  49, 111, 164,  70,  60,   2,  37, 171,  75,
	136, 156,  11,  56,  42, 146, 138, 229,  73, 146,  77,  61,  98, 196,
	135, 106,  63, 197, 195,  86,  96, 203, 113, 101, 170, 247, 181, 113,
	 80, 250, 108,   7, 255, 237, 129, 226,  79, 107, 112, 166, 103, 241,
	 24, 223, 239, 120, 198,  58,  60,  82, 128,   3, 184,  66, 143, 224,
	145, 224,  81, 206, 163,  45,  63,  90, 168, 114,  59,  33, 159,  95,
	 28, 139, 123,  98, 125, 196,  15,  70, 194, 253,  54,  14, 109, 226,
	 71,  17, 161,  93, 186,  87, 244, 138,  20,  52, 123, 251,  26,  36,
	 17,  46,  52, 231, 232,  76,  31, 221,  84,  37, 216, 165, 212, 106,
	197, 242,  98,  43,  39, 175, 254, 145, 190,  84, 118, 222, 187, 136,
	120, 163, 236, 249
};

//	Init table based RND generator
//	if randomize is false, the counter is set to 0
void US_InitRndT (wbool randomize)
{
	if (randomize)
	{
/*
__asm	mov	ax,SEG rndtable
__asm	mov	es,ax

__asm	mov	ah,2ch
__asm	int	21h		// GetSystemTime
__asm	and	dx,0ffh

__asm	mov	[es:rndindex],dx
*/
rndindex = 0;
	}
	else
		rndindex = 0;
}

//	Return a random # between 0-255
int US_RndT ()
{
	rndindex++;
	return rndtable[rndindex];
}
