#include <stdio.h>
#include <stdlib.h>
#include "system.h"
#include "opengl.h"
#include "wl_def.h"

//==========================================================================

static GLuint grtextures[NUMCHUNKS];
GLuint* texture_index = NULL;	// wall and sprites

//==========================================================================

void GL_Startup()
{
	memset(grtextures, 0, sizeof(grtextures));
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glDepthFunc(GL_LEQUAL);
	glEnable(GL_DEPTH_TEST);

	MM_GetPtr ((memptr)&texture_index, sizeof(GLuint)*PMSoundStart);
	memset(texture_index, 0, sizeof(GLuint)*PMSoundStart);

	printf("OpenGL version ");
	printf(glGetString(GL_VERSION));
	printf("\n");
	printf(glGetString(GL_RENDERER));
	printf("\n");
}

void LoadFonts()
{
	fontstruct *font;
	int x, y, i, j;
	int tx, ty;
	byte buffer[128][256], *source;
	byte loaded = 0;

	if (loaded)
		return;
	else
		loaded = 1;

	for (i = 0; i < 2; i++)
	{
		memset(buffer, 0, sizeof(buffer));
		CA_CacheGrChunk(STARTFONT+i);
		font = (fontstruct*)grsegs[STARTFONT+i];
		tx = ty = 0;

		for (j = 0; j < 256; j++)
		{
			if (font->location[j] == 0)
				continue;

			if ((tx + font->width[j]) > 255)
			{
				tx = 0;
				ty += font->height;

				if (ty > 127)
					break;
			}

			fonts[i].coords[j][0] = (float)tx/256;
			fonts[i].coords[j][1] = (float)(tx+font->width[j])/256;
			fonts[i].coords[j][2] = (float)ty/128;
			fonts[i].coords[j][3] = (float)(ty+font->height)/128;

			source = ((byte*)font)+font->location[j];

			for (x = tx; x < tx+font->width[j]; x++)
			{
				for (y = ty+font->height-1; y >= ty; y--)
				{
					buffer[y][x] = (*source != 0) ? 255 : 0;
					source += font->width[j];
				}
				source -= font->width[j]*font->height-1;
			}

			tx += font->width[j];
		}

		glGenTextures(1, &fonts[i].tex);
		glBindTexture(GL_TEXTURE_2D, fonts[i].tex);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_LINEAR_MIPMAP_NEAREST);
		gluBuild2DMipmaps(GL_TEXTURE_2D, GL_INTENSITY4, 256, 128,
			GL_LUMINANCE, GL_UNSIGNED_BYTE, buffer);
	}
}

void UnLoadFonts()
{
	glDeleteTextures(1, &fonts[0].tex);
	glDeleteTextures(1, &fonts[1].tex);
}

void GL_Shutdown()
{

}

//==========================================================================

void GL_CacheTexture(int chunknum)
{
	int	i, x, y, picnum = chunknum - STARTPICS;
	unsigned short width, height;
	long sx, sy;
	byte *source, *buffer;

	CA_CacheGrChunk (chunknum);
	source = grsegs[chunknum];
	sx = width = pictable[picnum].width;
	sy = height = pictable[picnum].height;

	for (i = 0; ((i < 16) && (sx != 0)); i++)
		sx = sx >> 1;
	sx = (i != 0) ? 1 << i : 1;

	for (i = 0; ((i < 16) && (sy != 0)); i++)
		sy = sy >> 1;
	sy = (i != 0) ? 1 << i : 1;

	buffer = malloc(sx*sy*3);

	for (i = 0; i < 4; i++)
	for (y = height-1; y >= 0; y--)
	for (x = 0; x < width/4; x++)
	{
		buffer[(y*sx+x*4+i)*3] = gamepal[(*source)*3];
		buffer[(y*sx+x*4+i)*3+1] = gamepal[(*source)*3+1];
		buffer[(y*sx+x*4+i)*3+2] = gamepal[(*source)*3+2];
		source++;
	}

	UNCACHEGRCHUNK(chunknum);

	for (y = height; y < sy; y++)
	for (x = 0; x < sx; x++)
	{
		buffer[(y*sx+x)*3] = buffer[((y-1)*sx+x)*3];
		buffer[(y*sx+x)*3+1] = buffer[((y-1)*sx+x)*3+1];
		buffer[(y*sx+x)*3+2] = buffer[((y-1)*sx+x)*3+2];
	}

	for (x = width; x < sx; x++)
	for (y = 0; y < height; y++)
	{
		buffer[(y*sx+x)*3] = buffer[(y*sx+x-1)*3];
		buffer[(y*sx+x)*3+1] = buffer[(y*sx+x-1)*3+1];
		buffer[(y*sx+x)*3+2] = buffer[(y*sx+x-1)*3+2];
	}

	glGenTextures(1, &grtextures[chunknum]);
	glBindTexture(GL_TEXTURE_2D, grtextures[chunknum]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_NEAREST);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, sx, sy,
		GL_RGB, GL_UNSIGNED_BYTE, buffer);
	free(buffer);
}

void GL_UnCacheTexture(int chunknum)
{
	glDeleteTextures(1, &grtextures[chunknum]);
	grtextures[chunknum] = 0;
}

//==========================================================================

void GL_FadeOut(int red, int green, int blue, int steps, void (*func)())
{
	int time;
	float clr[4] = { (float)red/255, (float)green/255, (float)blue/255, 0.0f };
	time = steps*2;

	while (time > 0)
	{
		func();

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, screen_width, 0, screen_height, -1, 1);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glColor4fv(clr);
		glBegin(GL_QUADS);
		glVertex2i(0, 0);
		glVertex2i(screen_width, 0);
		glVertex2i(screen_width, screen_height);
		glVertex2i(0, screen_height);
		glEnd();

		glDisable(GL_BLEND);
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		clr[3] = 2.0f - (float)time/steps;
		time -= S_GetNewTics();
		S_FinishDisplay();
	}

	screenfaded = true;
}

void GL_FadeIn(int red, int green, int blue, int steps, void (*func)())
{
	float clr[4] = { (float)red/255, (float)green/255, (float)blue/255, 1.0f };
	int time = steps*2;

	while (time > 0)
	{
		func();

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, screen_width, 0, screen_height, -1, 1);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glColor4fv(clr);
		glBegin(GL_QUADS);
		glVertex2i(0, 0);
		glVertex2i(screen_width, 0);
		glVertex2i(screen_width, screen_height);
		glVertex2i(0, screen_height);
		glEnd();

		glDisable(GL_BLEND);
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		clr[3] = 0.5f*time/steps;
		time -= S_GetNewTics();
		S_FinishDisplay();
	}

	screenfaded = false;
}

//==========================================================================

void GL_DrawTexture(int l, int b, int chunknum)
{
	int	i, picnum = chunknum - STARTPICS;
	unsigned short width, height;
	long sx, sy;

	sx = width = pictable[picnum].width;
	sy = height = pictable[picnum].height;
	b = 200-b-height;

	for (i = 0; ((i < 16) && (sx != 0)); i++)
		sx = sx >> 1;
	sx = (i != 0) ? 1 << i : 1;

	for (i = 0; ((i < 16) && (sy != 0)); i++)
		sy = sy >> 1;
	sy = (i != 0) ? 1 << i : 1;

	glColor3f(1,1,1);
	glBindTexture(GL_TEXTURE_2D, grtextures[chunknum]);
	glEnable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2i(l, b);
	glTexCoord2f((float)width/sx,0);
	glVertex2i(l+width, b);
	glTexCoord2f((float)width/sx,(float)height/sy);
	glVertex2i(l+width, b+height);
	glTexCoord2f(0, (float)height/sy);
	glVertex2i(l, b+height);
	glEnd();
	glDisable(GL_TEXTURE_2D);
}

void GL_DrawPicture(int l, int b, int chunknum)
{
	int	i, x, y, picnum = chunknum - STARTPICS;
	unsigned short width, height;
	long sx, sy;
	byte *source, *buffer;
	GLuint tex;

	CA_CacheGrChunk (chunknum);
	source = grsegs[chunknum];
	sx = width = pictable[picnum].width;
	sy = height = pictable[picnum].height;
	b = 200-b-height;

	for (i = 0; ((i < 16) && (sx != 0)); i++)
		sx = sx >> 1;
	sx = (i != 0) ? 1 << i : 1;

	for (i = 0; ((i < 16) && (sy != 0)); i++)
		sy = sy >> 1;
	sy = (i != 0) ? 1 << i : 1;

	buffer = malloc(sx*sy*3);

	for (i = 0; i < 4; i++)
	for (y = height-1; y >= 0; y--)
	for (x = 0; x < width/4; x++)
	{
		buffer[(y*sx+x*4+i)*3] = gamepal[(*source)*3];
		buffer[(y*sx+x*4+i)*3+1] = gamepal[(*source)*3+1];
		buffer[(y*sx+x*4+i)*3+2] = gamepal[(*source)*3+2];
		source++;
	}

	UNCACHEGRCHUNK(chunknum);

	for (y = height; y < sy; y++)
	for (x = 0; x < sx; x++)
	{
		buffer[(y*sx+x)*3] = buffer[((y-1)*sx+x)*3];
		buffer[(y*sx+x)*3+1] = buffer[((y-1)*sx+x)*3+1];
		buffer[(y*sx+x)*3+2] = buffer[((y-1)*sx+x)*3+2];
	}

	for (x = width; x < sx; x++)
	for (y = 0; y < height; y++)
	{
		buffer[(y*sx+x)*3] = buffer[(y*sx+x-1)*3];
		buffer[(y*sx+x)*3+1] = buffer[(y*sx+x-1)*3+1];
		buffer[(y*sx+x)*3+2] = buffer[(y*sx+x-1)*3+2];
	}

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_NEAREST);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, sx, sy,
		GL_RGB, GL_UNSIGNED_BYTE, buffer);
	free(buffer);

	glColor3f(1,1,1);
	glEnable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2i(l, b);
	glTexCoord2f((float)width/sx,0);
	glVertex2i(l+width, b);
	glTexCoord2f((float)width/sx,(float)height/sy);
	glVertex2i(l+width, b+height);
	glTexCoord2f(0, (float)height/sy);
	glVertex2i(l, b+height);
	glEnd();
	glDisable(GL_TEXTURE_2D);

	glDeleteTextures(1, &tex);
}

void GL_DrawLatch(int l, int b, int picnum)
{
	int	i, x, y;
	unsigned short width, height;
	long sx, sy;
	byte *source, *buffer;
	GLuint tex;

	sx = width = pictable[picnum-STARTPICS].width;
	sy = height = pictable[picnum-STARTPICS].height;
	source = (byte*)latchpics[2+picnum-LATCHPICS_LUMP_START];
	b = 200-b-height;
	l *= 8;

	for (i = 0; ((i < 16) && (sx != 0)); i++)
		sx = sx >> 1;
	sx = (i != 0) ? 1 << i : 1;

	for (i = 0; ((i < 16) && (sy != 0)); i++)
		sy = sy >> 1;
	sy = (i != 0) ? 1 << i : 1;

	buffer = malloc(sx*sy*3);

	for (y = 0; y < height; y++)
	for (x = 0; x < width; x++)
	{
		buffer[(y*sx+x)*3] = gamepal[(*source)*3];
		buffer[(y*sx+x)*3+1] = gamepal[(*source)*3+1];
		buffer[(y*sx+x)*3+2] = gamepal[(*source)*3+2];
		source++;
	}

	for (y = height; y < sy; y++)
	for (x = 0; x < sx; x++)
	{
		buffer[(y*sx+x)*3] = buffer[((y-1)*sx+x)*3];
		buffer[(y*sx+x)*3+1] = buffer[((y-1)*sx+x)*3+1];
		buffer[(y*sx+x)*3+2] = buffer[((y-1)*sx+x)*3+2];
	}

	for (x = width; x < sx; x++)
	for (y = 0; y < height; y++)
	{
		buffer[(y*sx+x)*3] = buffer[(y*sx+x-1)*3];
		buffer[(y*sx+x)*3+1] = buffer[(y*sx+x-1)*3+1];
		buffer[(y*sx+x)*3+2] = buffer[(y*sx+x-1)*3+2];
	}

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_NEAREST);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, sx, sy,
		GL_RGB, GL_UNSIGNED_BYTE, buffer);
	free(buffer);

	glColor3f(1,1,1);
	glEnable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2i(l, b);
	glTexCoord2f((float)width/sx,0);
	glVertex2i(l+width, b);
	glTexCoord2f((float)width/sx,(float)height/sy);
	glVertex2i(l+width, b+height);
	glTexCoord2f(0, (float)height/sy);
	glVertex2i(l, b+height);
	glEnd();
	glDisable(GL_TEXTURE_2D);

	glDeleteTextures(1, &tex);
}

void GL_DrawPropString(char *string)
{
	fontstruct *font;
	int x, y;
	byte ch;

	font = (fontstruct*)grsegs[STARTFONT+fontnumber];
	glBindTexture(GL_TEXTURE_2D, fonts[fontnumber].tex);

	glColor3f((float)gamepal[fontcolor*3]/255, (float)gamepal[fontcolor*3+1]/255, (float)gamepal[fontcolor*3+2]/255);
	glEnable(GL_TEXTURE_2D);
	glAlphaFunc(GL_GREATER, 0.4f);
	glEnable(GL_ALPHA_TEST);
	glBegin(GL_QUADS);

	x = px;
	y = 200-py-font->height;

	while ((ch = *string++)!=0)
	{
		glTexCoord2f(fonts[fontnumber].coords[ch][0], fonts[fontnumber].coords[ch][2]);
		glVertex2i(x, y);
		glTexCoord2f(fonts[fontnumber].coords[ch][1], fonts[fontnumber].coords[ch][2]);
		glVertex2i(x+font->width[ch], y);
		glTexCoord2f(fonts[fontnumber].coords[ch][1], fonts[fontnumber].coords[ch][3]);
		glVertex2i(x+font->width[ch], y+font->height);
		glTexCoord2f(fonts[fontnumber].coords[ch][0], fonts[fontnumber].coords[ch][3]);
		glVertex2i(x, y+font->height);
		x += font->width[ch];
	}

	glEnd();
	glDisable(GL_ALPHA_TEST);
	glDisable(GL_TEXTURE_2D);
}

void GL_Bar(int x, int y, int w, int h, int color)
{
	y = 200-y-h;

	glColor3f((float)gamepal[color*3]/255, (float)gamepal[color*3+1]/255, (float)gamepal[color*3+2]/255);
	glBegin(GL_QUADS);
	glVertex2i(x, y);
	glVertex2i(x+w, y);
	glVertex2i(x+w, y+h);
	glVertex2i(x, y+h);
	glEnd();
}

void GL_Hlin (int x1, int x2, int y, int color)
{
	y = 200-y;

	glColor3f((float)gamepal[color*3]/255, (float)gamepal[color*3+1]/255, (float)gamepal[color*3+2]/255);
	glBegin(GL_LINES);
	glVertex2i(x1, y);
	glVertex2i(x2, y);
	glEnd();
}

void GL_Vlin (int y1, int y2, int x, int color)
{
	y1 = 200-y1;
	y2 = 200-y2;

	glColor3f((float)gamepal[color*3]/255, (float)gamepal[color*3+1]/255, (float)gamepal[color*3+2]/255);
	glBegin(GL_LINES);
	glVertex2i(x, y1);
	glVertex2i(x, y2);
	glEnd();
}

void GL_Plot (int x, int y, int color)
{
	y = 200-y;

	glColor3f((float)gamepal[color*3]/255, (float)gamepal[color*3+1]/255, (float)gamepal[color*3+2]/255);
	glBegin(GL_POINTS);
	glVertex2i(x, y);
	glEnd();
}

//==========================================================================

static GLuint screentex[2];

// If it wasn't for 3dfx cards, one texture would be enough
void GL_CacheScreen(int chunknum, int palette)
{
	int	i, x, y;//, picnum = chunknum - STARTPICS;
	byte *source, *buffer, pal[768];
#ifdef SPEAR
	byte *title;
#endif

	if (palette == -1)
		memcpy(pal, gamepal, 768);
	else
	{
		CA_CacheGrChunk(palette);
		for (i = 0; i < 768; i++)
			pal[i] = ((byte*)grsegs[palette])[i] << 2;
		UNCACHEGRCHUNK(palette);
	}

	CA_CacheGrChunk (chunknum);
	MM_GetPtr((void*)&buffer, 256*256*3);
	source = grsegs[chunknum];
#ifdef SPEAR
	if (chunknum == TITLE1PIC)
	{
		CA_CacheGrChunk(TITLE2PIC);
		MM_GetPtr(&title, 320*200);
		for (i = 0; i < 4; i++)
		{
			memcpy(title+i*320*200/4, source+i*320*80/4, 80*320/4);
			memcpy(title+i*320*200/4+80*320/4, (byte*)grsegs[TITLE2PIC]+i*320*120/4, 120*320/4);
		}
		source = title;
		UNCACHEGRCHUNK(TITLE2PIC);
	}
#endif

	for (i = 0; i < 4; i++)
	for (y = 200-1; y >= 0; y--)
	{
		for (x = 0; x < 64; x++)
		{
			buffer[(y*256+x*4+i)*3] = pal[(*source)*3];
			buffer[(y*256+x*4+i)*3+1] = pal[(*source)*3+1];
			buffer[(y*256+x*4+i)*3+2] = pal[(*source)*3+2];
			source++;
		}
		source += 16;
	}

	glGenTextures(2, screentex);

	glBindTexture(GL_TEXTURE_2D, screentex[0]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_NEAREST);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 256, 256,
		GL_RGB, GL_UNSIGNED_BYTE, buffer);

	source = grsegs[chunknum];
#ifdef SPEAR
	if (chunknum == TITLE1PIC)
		source = title;
#endif

	for (i = 0; i < 4; i++)
	for (y = 200-1; y >= 0; y--)
	{
		source += 64;
		for (x = 0; x < 16; x++)
		{
			buffer[(y*64+x*4+i)*3] = pal[(*source)*3];
			buffer[(y*64+x*4+i)*3+1] = pal[(*source)*3+1];
			buffer[(y*64+x*4+i)*3+2] = pal[(*source)*3+2];
			source++;
		}
	}

	glBindTexture(GL_TEXTURE_2D, screentex[1]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_NEAREST);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 64, 256,
		GL_RGB, GL_UNSIGNED_BYTE, buffer);

#ifdef SPEAR
	if (chunknum == TITLE1PIC)
	{
		MM_FreePtr(&title);
	}
#endif
	UNCACHEGRCHUNK(chunknum);
	MM_FreePtr((void*)&buffer);
}

void GL_FreeScreen()
{
	if (screentex[0])
		glDeleteTextures(2, screentex);
}

void GL_DrawScreen()
{
	glColor3f(1,1,1);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, screentex[0]);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2i(0, 0);
	glTexCoord2f(1, 0);
	glVertex2i(256, 0);
	glTexCoord2f(1, (float)200/256);
	glVertex2i(256, 200);
	glTexCoord2f(0, (float)200/256);
	glVertex2i(0, 200);
	glEnd();

	glBindTexture(GL_TEXTURE_2D, screentex[1]);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2i(256, 0);
	glTexCoord2f(1, 0);
	glVertex2i(320, 0);
	glTexCoord2f(1, (float)200/256);
	glVertex2i(320, 200);
	glTexCoord2f(0, (float)200/256);
	glVertex2i(256, 200);
	glEnd();

	glDisable(GL_TEXTURE_2D);
}
