// WL_AGENT.C

#include <string.h>
#include "wl_def.h"

//=============================================================================
//						 LOCAL CONSTANTS
//=============================================================================

#define MOVESCALE		150l
#define BACKMOVESCALE	100l
#define ANGLESCALE		20

//=============================================================================
//						 GLOBAL VARIABLES
//=============================================================================

// player state info
wbool		running;
long		thrustspeed;

// player coordinates scaled to unsigned
unsigned short	plux,pluy;
int			anglefrac;
int			gotgatgun;

objtype		*LastAttacker;

//=============================================================================
//						 LOCAL VARIABLES
//=============================================================================


long	playerxmove,playerymove;

struct atkinf
{
	char tics,attack,frame;
		// attack is 1 for gun, 2 for knife
} attackinfo[4][14] =

{
{ {6,0,1},{6,2,2},{6,0,3},{6,-1,4} },
{ {6,0,1},{6,1,2},{6,0,3},{6,-1,4} },
{ {6,0,1},{6,1,2},{6,3,3},{6,-1,4} },
{ {6,0,1},{6,1,2},{6,4,3},{6,-1,4} },
};

int	strafeangle[9] = {0,90,180,270,45,135,225,315,0};

void DrawWeapon();
void GiveWeapon(int weapon);
void GiveAmmo(int ammo);

//===========================================================================

//----------

void Attack();
void Use();
void Search(objtype *ob);
void SelectWeapon();
void SelectItem();

//----------

wbool TryMove (objtype *ob);
void T_Player (objtype *ob);

void ClipMove (objtype *ob, long xmove, long ymove);


//=============================================================================
//						CONTROL STUFF
//=============================================================================


//	Keys 1-4 change weapons
void CheckWeaponChange()
{
	int	i;

	// must use knife with no ammo
	if (!gamestate.ammo)
		return;

	for (i=wp_knife ; i<=gamestate.bestweapon ; i++)
		if (buttonstate[bt_readyknife+i-wp_knife])
		{
			gamestate.weapon = gamestate.chosenweapon = i;
			DrawWeapon ();
			return;
		}
}

//	Takes controlx,controly, and buttonstate[bt_strafe]
//
//	Changes the player's angle and position
//
//	There is an angle hack because when going 70 fps, the roundoff becomes
//	significant
void ControlMovement (objtype *ob)
{
	long	oldx, oldy;
	int		angle, angleunits;

	thrustspeed = 0;

	oldx = player->x;
	oldy = player->y;

	// side to side move
	if (buttonstate[bt_strafe])
	{
		// strafing
		if (controlx > 0)
		{
			angle = ob->angle - ANGLES/4;
			if (angle < 0)
				angle += ANGLES;

			// move to left
			Thrust (angle,controlx*MOVESCALE);
		}
		else if (controlx < 0)
		{
			angle = ob->angle + ANGLES/4;
			if (angle >= ANGLES)
				angle -= ANGLES;

			// move to right
			Thrust (angle,-controlx*MOVESCALE);
		}
	}
	else
	{
		// not strafing
		anglefrac += controlx;
		angleunits = anglefrac/ANGLESCALE;
		anglefrac -= angleunits*ANGLESCALE;
		ob->angle -= angleunits;

		if (ob->angle >= ANGLES)
			ob->angle -= ANGLES;
		if (ob->angle < 0)
			ob->angle += ANGLES;
	}

	// forward/backwards move
	if (controly < 0)
	{
		// move forwards
		Thrust (ob->angle,-controly*MOVESCALE);
	}
	else if (controly > 0)
	{
		angle = ob->angle + ANGLES/2;
		if (angle >= ANGLES)
			angle -= ANGLES;

		// move backwards
		Thrust (angle,controly*BACKMOVESCALE);
	}

	// watching the BJ actor
	if (gamestate.victoryflag)
		return;

	// calculate total move
	playerxmove = player->x - oldx;
	playerymove = player->y - oldy;
}

//=============================================================================
//					STATUS WINDOW STUFF
//=============================================================================
#undef MAXLONG
#include "opengl.h"
extern GLuint statustex[2];

void StatusDrawPic(unsigned short left, unsigned short b, unsigned short picnum)
{
	int x, y;
	unsigned short wide, height, l;
	byte* source;
	byte tmp[64*64][3];

	wide = pictable[picnum-STARTPICS].width;
	height = pictable[picnum-STARTPICS].height;
	source = (byte*)latchpics[2+picnum-LATCHPICS_LUMP_START];
	l = left*8;
	b = 40-b-height;
	CheckStatusBar();

	if (l+wide > 255)
	{
		if (l < 256)
		{
			wide = l+wide-256;
			l = 0;
		}
		else
			l -= 256;

		for (y = 0; y < height; y++)
		for (x = 0; x < wide; x++)
		{
			tmp[y*wide+x][0] = gamepal[(*source)*3];
			tmp[y*wide+x][1] = gamepal[(*source)*3+1];
			tmp[y*wide+x][2] = gamepal[(*source)*3+2];
			source++;
		}

		glBindTexture(GL_TEXTURE_2D, statustex[1]);
		glTexSubImage2D(GL_TEXTURE_2D, 0, l, b, wide, height, 
			GL_RGB, GL_UNSIGNED_BYTE, tmp); 

		wide = pictable[picnum-STARTPICS].width;
		source = (byte*)latchpics[2+picnum-LATCHPICS_LUMP_START];
		l = left*8;
	}

	for (y = 0; y < height; y++)
	for (x = 0; x < wide; x++)
	{
		tmp[y*wide+x][0] = gamepal[(*source)*3];
		tmp[y*wide+x][1] = gamepal[(*source)*3+1];
		tmp[y*wide+x][2] = gamepal[(*source)*3+2];
		source++;
	}

	glBindTexture(GL_TEXTURE_2D, statustex[0]);
	glTexSubImage2D(GL_TEXTURE_2D, 0, l, b, wide, height, 
		GL_RGB, GL_UNSIGNED_BYTE, tmp); 
 
//	statusdirty = true;
}

void DrawFace()
{
	if (gamestate.health)
	{
		#ifdef SPEAR
		if (godmode)
			StatusDrawPic(17,4,GODMODEFACE1PIC+gamestate.faceframe);
		else
		#endif
		StatusDrawPic(17,4,FACE1APIC+3*((100-gamestate.health)/16)+gamestate.faceframe);
	}
	else
	{
#ifndef SPEAR
	 if (LastAttacker->obclass == needleobj)
	   StatusDrawPic(17,4,MUTANTBJPIC);
	 else
#endif
	   StatusDrawPic(17,4,FACE8APIC);
	}
}

//	Calls draw face if time to change

#define FACETICS	70

int	facecount;

void UpdateFace()
{
	if (SD_SoundPlaying() == GETGATLINGSND)
	  return;

	facecount += tics;
	if (facecount > US_RndT())
	{
		gamestate.faceframe = (US_RndT()>>6);
		if (gamestate.faceframe==3)
			gamestate.faceframe = 1;

		facecount = 0;
		DrawFace ();
	}
}

//	right justifies and pads with blanks
void LatchNumber (int x, int y, int width, long number)
{
	unsigned short length, c;
	char str[20];

	sprintf(str, "%d", number);
	length = strlen(str);

	while (length < width)
	{
		StatusDrawPic (x,y,N_BLANKPIC);
		x++;
		width--;
	}

	c= length <= width ? 0 : length-width;

	while (c<length)
	{
		StatusDrawPic (x,y,str[c]-'0'+ N_0PIC);
		x++;
		c++;
	}
}

void DrawHealth()
{
	LatchNumber (21,16,3,gamestate.health);
}

void TakeDamage(int points, objtype *attacker)
{
	LastAttacker = attacker;

	if (gamestate.victoryflag)
		return;
	if (gamestate.difficulty==gd_baby)
	  points>>=2;

	if (!godmode)
		gamestate.health -= points;

	if (gamestate.health<=0)
	{
		gamestate.health = 0;
		playstate = ex_died;
		killerobj = attacker;
	}

	StartDamageFlash (points);

	gotgatgun=0;

	DrawHealth ();
	DrawFace ();

	// MAKE BJ'S EYES BUG IF MAJOR DAMAGE!
	#ifdef SPEAR
	if (points > 30 && gamestate.health!=0 && !godmode)
	{
		StatusDrawPic (17,4,BJOUCHPIC);
		facecount = 0;
	}
	#endif
}

void HealSelf(int points)
{
	gamestate.health += points;
	if (gamestate.health>100)
		gamestate.health = 100;

	DrawHealth ();
	gotgatgun = 0;
	DrawFace ();
}

//===========================================================================

void DrawLevel()
{
#ifdef SPEAR
	if (gamestate.mapon == 20)
		LatchNumber (2,16,2,18);
	else
#endif
	LatchNumber (2,16,2,gamestate.mapon+1);
}

//===========================================================================

void DrawLives()
{
	LatchNumber (14,16,1,gamestate.lives);
}


void GiveExtraMan()
{
	if (gamestate.lives<9)
		gamestate.lives++;

	DrawLives ();
	SD_PlaySound (BONUS1UPSND);
}

//===========================================================================

void DrawScore()
{
	LatchNumber (6,16,6,gamestate.score);
}

void GivePoints(long points)
{
	gamestate.score += points;
	while (gamestate.score >= gamestate.nextextra)
	{
		gamestate.nextextra += EXTRAPOINTS;
		GiveExtraMan();
	}
	DrawScore();
}

//===========================================================================

void DrawWeapon()
{
	StatusDrawPic(32,8,KNIFEPIC+gamestate.weapon);
}

void DrawKeys()
{
	if (gamestate.keys & 1)
		StatusDrawPic(30,4,GOLDKEYPIC);
	else
		StatusDrawPic(30,4,NOKEYPIC);

	if (gamestate.keys & 2)
		StatusDrawPic(30,20,SILVERKEYPIC);
	else
		StatusDrawPic(30,20,NOKEYPIC);
}

void GiveWeapon (int weapon)
{
	GiveAmmo(6);

	if (gamestate.bestweapon<weapon)
		gamestate.bestweapon = gamestate.weapon
		= gamestate.chosenweapon = weapon;

	DrawWeapon();
}


//===========================================================================

void DrawAmmo()
{
	LatchNumber (27,16,2,gamestate.ammo);
}

void GiveAmmo(int ammo)
{
	// knife was out
	if (!gamestate.ammo)
	{
		if (!gamestate.attackframe)
		{
			gamestate.weapon = gamestate.chosenweapon;
			DrawWeapon ();
		}
	}

	gamestate.ammo += ammo;
	
	if (gamestate.ammo > 99)
		gamestate.ammo = 99;

	DrawAmmo();
}

//===========================================================================

void GiveKey(int key)
{
	gamestate.keys |= (1<<key);
	DrawKeys();
}

//=============================================================================
//							MOVEMENT
//=============================================================================

void GetBonus(statobj_t *check)
{
	switch (check->itemnumber)
	{
	case	bo_firstaid:
		if (gamestate.health == 100)
			return;

		SD_PlaySound (HEALTH2SND);
		HealSelf (25);
		break;

	case	bo_key1:
	case	bo_key2:
	case	bo_key3:
	case	bo_key4:
		GiveKey (check->itemnumber - bo_key1);
		SD_PlaySound (GETKEYSND);
		break;

	case	bo_cross:
		SD_PlaySound (BONUS1SND);
		GivePoints (100);
		gamestate.treasurecount++;
		break;
	case	bo_chalice:
		SD_PlaySound (BONUS2SND);
		GivePoints (500);
		gamestate.treasurecount++;
		break;
	case	bo_bible:
		SD_PlaySound (BONUS3SND);
		GivePoints (1000);
		gamestate.treasurecount++;
		break;
	case	bo_crown:
		SD_PlaySound (BONUS4SND);
		GivePoints (5000);
		gamestate.treasurecount++;
		break;

	case	bo_clip:
		if (gamestate.ammo == 99)
			return;

		SD_PlaySound (GETAMMOSND);
		GiveAmmo (8);
		break;
	case	bo_clip2:
		if (gamestate.ammo == 99)
			return;

		SD_PlaySound (GETAMMOSND);
		GiveAmmo (4);
		break;

#ifdef SPEAR
	case	bo_25clip:
		if (gamestate.ammo == 99)
		  return;

		SD_PlaySound (GETAMMOBOXSND);
		GiveAmmo (25);
		break;
#endif

	case	bo_machinegun:
		SD_PlaySound (GETMACHINESND);
		GiveWeapon (wp_machinegun);
		break;
	case	bo_chaingun:
		SD_PlaySound (GETGATLINGSND);
		GiveWeapon (wp_chaingun);

		StatusDrawPic(17,4,GOTGATLINGPIC);
		facecount = 0;
		gotgatgun = 1;
		break;

	case	bo_fullheal:
		SD_PlaySound (BONUS1UPSND);
		HealSelf (99);
		GiveAmmo (25);
		GiveExtraMan ();
		gamestate.treasurecount++;
		break;

	case	bo_food:
		if (gamestate.health == 100)
			return;

		SD_PlaySound (HEALTH1SND);
		HealSelf (10);
		break;

	case	bo_alpo:
		if (gamestate.health == 100)
			return;

		SD_PlaySound (HEALTH1SND);
		HealSelf (4);
		break;

	case	bo_gibs:
		if (gamestate.health >10)
			return;

		SD_PlaySound (SLURPIESND);
		HealSelf (1);
		break;

	case	bo_spear:
		spearflag = true;
		spearx = player->x;
		speary = player->y;
		spearangle = player->angle;
		playstate = ex_completed;
	}

	StartBonusFlash ();
	check->shapenum = -1;
			// remove from list
}

//	returns true if move ok
//	debug: use pointers to optimize
wbool TryMove (objtype *ob)
{
	int			xl,yl,xh,yh,x,y;
	objtype		*check;
	long		deltax,deltay;

	xl = (ob->x-PLAYERSIZE) >>TILESHIFT;
	yl = (ob->y-PLAYERSIZE) >>TILESHIFT;

	xh = (ob->x+PLAYERSIZE) >>TILESHIFT;
	yh = (ob->y+PLAYERSIZE) >>TILESHIFT;

	// check for solid walls
	for (y=yl;y<=yh;y++)
		for (x=xl;x<=xh;x++)
		{
			check = actorat[x][y];
			if (check && check<objlist)
				return false;
		}

	// check for actors
	if (yl>0)
		yl--;
	if (yh<MAPSIZE-1)
		yh++;
	if (xl>0)
		xl--;
	if (xh<MAPSIZE-1)
		xh++;

	for (y=yl;y<=yh;y++)
		for (x=xl;x<=xh;x++)
		{
			check = actorat[x][y];
			if (check > objlist
			&& (check->flags & FL_SHOOTABLE) )
			{
				deltax = ob->x - check->x;
				if (deltax < -MINACTORDIST || deltax > MINACTORDIST)
					continue;
				deltay = ob->y - check->y;
				if (deltay < -MINACTORDIST || deltay > MINACTORDIST)
					continue;

				return false;
			}
		}

	return true;
}

void ClipMove (objtype *ob, long xmove, long ymove)
{
	long	basex,basey;

	basex = ob->x;
	basey = ob->y;

	ob->x = basex+xmove;
	ob->y = basey+ymove;
	if (TryMove (ob))
		return;

	if (noclip && ob->x > 2*TILEGLOBAL && ob->y > 2*TILEGLOBAL &&
	ob->x < (((long)(mapwidth-1))<<TILESHIFT)
	&& ob->y < (((long)(mapheight-1))<<TILESHIFT) )
		return;
		// walk through walls

	if (!SD_SoundPlaying())
		SD_PlaySound (HITWALLSND);

	ob->x = basex+xmove;
	ob->y = basey;
	if (TryMove (ob))
		return;

	ob->x = basex;
	ob->y = basey+ymove;
	if (TryMove (ob))
		return;

	ob->x = basex;
	ob->y = basey;
}

//==========================================================================

void SpawnBJVictory();
 // From WL_ACT2.C

void VictoryTile()
{
#ifndef SPEAR
	SpawnBJVictory ();
#endif

	gamestate.victoryflag = true;
}

void Thrust (int angle, long speed)
{
	long xmove,ymove;
	unsigned short offset;


	// ZERO FUNNY COUNTER IF MOVED!
	#ifdef SPEAR
	if (speed)
		funnyticount = 0;
	#endif

	thrustspeed += speed;

	// moving bounds speed
	if (speed >= MINDIST*2)
		speed = MINDIST*2-1;

	xmove = speed*costable[angle];
	ymove = -speed*sintable[angle];

//	xmove = speed*cos(0.017453f*angle);
//	ymove = speed*sin(0.017453f*angle);

	ClipMove(player,xmove,ymove);

	player->tilex = player->x >> TILESHIFT;
		// scale to tile values
	player->tiley = player->y >> TILESHIFT;

	offset = farmapylookup[player->tiley]+player->tilex;
	player->areanumber = *(mapsegs[0] + offset) -AREATILE;

	if (*(mapsegs[1] + offset) == EXITTILE)
		VictoryTile ();
}


//=============================================================================
//								ACTIONS
//=============================================================================


void Cmd_Fire()
{
	buttonheld[bt_attack] = true;

	gamestate.weaponframe = 0;

	player->state = s_attack;

	gamestate.attackframe = 0;
	gamestate.attackcount =
		attackinfo[gamestate.weapon][gamestate.attackframe].tics;
	gamestate.weaponframe =
		attackinfo[gamestate.weapon][gamestate.attackframe].frame;
}

//===========================================================================

void Cmd_Use()
{
	int		checkx,checky,doornum,dir;
	wbool	elevatorok;

	// find which cardinal direction the player is facing
	if (player->angle < ANGLES/8 || player->angle > 7*ANGLES/8)
	{
		checkx = player->tilex + 1;
		checky = player->tiley;
		dir = di_east;
		elevatorok = true;
	}
	else if (player->angle < 3*ANGLES/8)
	{
		checkx = player->tilex;
		checky = player->tiley - 1;
		dir = di_north;
		elevatorok = false;
	}
	else if (player->angle < 5*ANGLES/8)
	{
		checkx = player->tilex - 1;
		checky = player->tiley;
		dir = di_west;
		elevatorok = true;
	}
	else
	{
		checkx = player->tilex;
		checky = player->tiley + 1;
		dir = di_south;
		elevatorok = false;
	}

	doornum = tilemap[checkx][checky];
	if (*(mapsegs[1]+farmapylookup[checky]+checkx) == PUSHABLETILE)
	{
		// pushable wall
		PushWall (checkx,checky,dir);
		return;
	}
//playstate = ex_completed;

	if (!buttonheld[bt_use] && doornum == ELEVATORTILE && elevatorok)
	{
		// use elevator
		buttonheld[bt_use] = true;

		tilemap[checkx][checky]++;
		// flip switch
		if (*(mapsegs[0]+farmapylookup[player->tiley]+player->tilex) == ALTELEVATORTILE)
			playstate = ex_secretlevel;
		else
			playstate = ex_completed;
		SD_PlaySound (LEVELDONESND);
		SD_WaitSoundDone();
	}
	else if (!buttonheld[bt_use] && doornum & 0x80)
	{
		buttonheld[bt_use] = true;
		OperateDoor (doornum & ~0x80);
	}
	else
		SD_PlaySound (DONOTHINGSND);
}


//=============================================================================
//						   PLAYER CONTROL
//=============================================================================


void SpawnPlayer (int tilex, int tiley, int dir)
{
	player->obclass = playerobj;
	player->active = true;
	player->tilex = tilex;
	player->tiley = tiley;
	player->areanumber =
		*(mapsegs[0] + farmapylookup[player->tiley]+player->tilex);
	player->x = ((long)tilex<<TILESHIFT)+TILEGLOBAL/2;
	player->y = ((long)tiley<<TILESHIFT)+TILEGLOBAL/2;
	player->state = s_player;
	player->angle = (1-dir)*90;
	if (player->angle<0)
		player->angle += ANGLES;
	player->flags = FL_NEVERMARK;

	// set some variables
	Thrust (0,0);

	InitAreas ();
}


//===========================================================================

//	Update player hands, and try to do damage when the proper frame is reached
void KnifeAttack(objtype *ob)
{
	objtype *check, *closest;
	long	dist;

	SD_PlaySound (ATKKNIFESND);	// actually fire
	dist = 0x7fffffff;
	closest = NULL;
	for (check = ob->next; check; check = check->next)
		if ((check->flags & FL_SHOOTABLE) &&
			(check->flags & FL_VISABLE) &&
			(abs (check->viewx-centerx) < shootdelta))
		{
			if (check->transx < dist)
			{
				dist = check->transx;
				closest = check;
			}
		}

	if (!closest || dist> 0x18000l)
	{
		// missed
		return;
	}

	// hit something
	DamageActor (closest,US_RndT() >> 4);
}

void GunAttack (objtype *ob)
{
	objtype *check,*closest,*oldclosest;
	int		damage;
	int		dx,dy,dist;
	long	viewdist;

	switch (gamestate.weapon)
	{
	case wp_pistol:
		SD_PlaySound (ATKPISTOLSND);
		break;
	case wp_machinegun:
		SD_PlaySound (ATKMACHINEGUNSND);
		break;
	case wp_chaingun:
		SD_PlaySound (ATKGATLINGSND);
		break;
	}

	madenoise = true;

	// find potential targets
	viewdist = 0x7fffffffl;
	closest = NULL;

//	while (1)
	{
		oldclosest = closest;

		for (check=ob->next ; check ; check=check->next)
			if ((check->flags & FL_SHOOTABLE)
				//&&
//				(check->flags & FL_VISABLE) &&
//				(abs(check->viewx - centerx) < shootdelta)
			)
//			{
		if (CheckLine(check))
//				if (check->transx < viewdist)
				{
//					viewdist = check->transx;
					closest = check;
			break;
				}
//			}

		if (closest == oldclosest)
			return; // no more targets, all missed

		// trace a line from player to enemey
//		if (CheckLine(closest))
//			break;
	}

	// hit something
	dx = abs(closest->tilex - player->tilex);
	dy = abs(closest->tiley - player->tiley);
	dist = dx>dy ? dx:dy;

	if (dist<2)
		damage = US_RndT() / 4;
	else if (dist<4)
		damage = US_RndT() / 6;
	else
	{
		if ( (US_RndT() / 12) < dist)
		// missed
			return;
		damage = US_RndT() / 6;
	}

	DamageActor (closest,damage);
}

//===========================================================================

void VictorySpin()
{
	long	desty;

	if (player->angle > 270)
	{
		player->angle -= tics * 3;
		if (player->angle < 270)
			player->angle = 270;
	}
	else if (player->angle < 270)
	{
		player->angle += tics * 3;
		if (player->angle > 270)
			player->angle = 270;
	}

	desty = (((long)player->tiley-5)<<TILESHIFT)-0x3000;

	if (player->y > desty)
	{
		player->y -= tics*4096;
		if (player->y < desty)
			player->y = desty;
	}
}


//===========================================================================

void T_Attack(objtype *ob)
{
	struct	atkinf	*cur;

	UpdateFace();

	// watching the BJ actor
	if (gamestate.victoryflag)
	{
		VictorySpin ();
		return;
	}

	if ( buttonstate[bt_use] && !buttonheld[bt_use] )
		buttonstate[bt_use] = false;

	if ( buttonstate[bt_attack] && !buttonheld[bt_attack])
		buttonstate[bt_attack] = false;

	ControlMovement (ob);

	// watching the BJ actor
	if (gamestate.victoryflag)
		return;

	// scale to fit in unsigned
	plux = player->x >> UNSIGNEDSHIFT;
	pluy = player->y >> UNSIGNEDSHIFT;

	// scale to tile values
	player->tilex = player->x >> TILESHIFT;
	player->tiley = player->y >> TILESHIFT;

	// change frame and fire
	gamestate.attackcount -= tics;
	while (gamestate.attackcount <= 0)
	{
		cur = &attackinfo[gamestate.weapon][gamestate.attackframe];
		switch (cur->attack)
		{
		case -1:
			ob->state = s_player;
			if (!gamestate.ammo)
			{
				gamestate.weapon = wp_knife;
				DrawWeapon();
			}
			else
			{
				if (gamestate.weapon != gamestate.chosenweapon)
				{
					gamestate.weapon = gamestate.chosenweapon;
					DrawWeapon ();
				}
			};
			gamestate.attackframe = gamestate.weaponframe = 0;
			return;

		case 4:
			if (!gamestate.ammo)
				break;
			if (buttonstate[bt_attack])
				gamestate.attackframe -= 2;
		case 1:
			if (!gamestate.ammo)
			{	
				// can only happen with chain gun
				gamestate.attackframe++;
				break;
			}
			GunAttack (ob);
			gamestate.ammo--;
			DrawAmmo ();
			break;

		case 2:
			KnifeAttack (ob);
			break;

		case 3:
			if (gamestate.ammo && buttonstate[bt_attack])
				gamestate.attackframe -= 2;
			break;
		}

		gamestate.attackcount += cur->tics;
		gamestate.attackframe++;
		gamestate.weaponframe =
			attackinfo[gamestate.weapon][gamestate.attackframe].frame;
	}

}



//===========================================================================

void T_Player(objtype *ob)
{
	// watching the BJ actor
	if (gamestate.victoryflag)
	{
		VictorySpin();
		return;
	}

	UpdateFace();
	CheckWeaponChange();

	if (buttonstate[bt_use])
		Cmd_Use();

	if (buttonstate[bt_attack] && !buttonheld[bt_attack])
		Cmd_Fire();

	ControlMovement(ob);

	// watching the BJ actor
	if (gamestate.victoryflag)
		return;

	// scale to fit in unsigned
	plux = player->x >> UNSIGNEDSHIFT;
	pluy = player->y >> UNSIGNEDSHIFT;

	// scale to tile values
	player->tilex = player->x >> TILESHIFT;
	player->tiley = player->y >> TILESHIFT;
}
