// WL_DRAW.C

#include <stdio.h>
#include <stdlib.h>
#include "system.h"
#include "opengl.h"
#include "wl_def.h"
#include "console.h"

//=============================================================================
//						 LOCAL CONSTANTS
//=============================================================================

// the door is the last picture before the sprites
#define DOORWALL	(PMSpriteStart-8)
#define TS (1.0f)
#define ACTORSIZE	0x4000

static unsigned char vgaCeiling[]=
{
#ifndef SPEAR
	0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0xbf,
	0x4e,0x4e,0x4e,0x1d,0x8d,0x4e,0x1d,0x2d,0x1d,0x8d,
	0x1d,0x1d,0x1d,0x1d,0x1d,0x2d,0xdd,0x1d,0x1d,0x98,

	0x1d,0x9d,0x2d,0xdd,0xdd,0x9d,0x2d,0x4d,0x1d,0xdd,
	0x7d,0x1d,0x2d,0x2d,0xdd,0xd7,0x1d,0x1d,0x1d,0x2d,
	0x1d,0x1d,0x1d,0x1d,0xdd,0xdd,0x7d,0xdd,0xdd,0xdd
#else
	0x6f,0x4f,0x1d,0xde,0xdf,0x2e,0x7f,0x9e,0xae,0x7f,
	0x1d,0xde,0xdf,0xde,0xdf,0xde,0xe1,0xdc,0x2e,0x1d,0xdc
#endif
};

//=============================================================================
//						 GLOBAL VARIABLES
//=============================================================================

// statusbar buffer (320 x 40)
unsigned char	statusbar1[64][256][3];
unsigned char	statusbar2[64][64][3];
wbool			statusdirty[2];
GLuint			statustex[2] = { 0, 0 };

float	r_alpharef = 0.75f;

long 	LastTicCount;
long 	frameon;
/*
unsigned short wallheight[MAXVIEWWIDTH];
*/
fixed	tileglobal	= TILEGLOBAL;
fixed	mindist		= MINDIST;

// math tables
int		pixelangle[MAXVIEWWIDTH];
long	finetangent[FINEANGLES/4];
float	sintable[ANGLES+ANGLES/4], *costable = sintable+(ANGLES/4);

// refresh variables
fixed	viewx, viewy;	// the focal point
int		viewangle;
float	viewsin, viewcos;

/*
void	TransformActor(objtype *ob);
int		CalcRotate(objtype *ob);
void	DrawScaleds();

// ray tracing variables
int		focaltx,focalty;
int		midangle,angle;

int		xtile,ytile;
long	xstep,ystep;
*/
int		horizwall[MAXWALLTILES], vertwall[MAXWALLTILES];

//============================================================================
//			   3 - D  DEFINITIONS
//============================================================================

GLuint arealists[NUMAREAS];
float doorverts[MAXDOORS*8*3];
float doorcoords[MAXDOORS*8*2];

static void CreateWallTexture(int num)
{
	int s, t;
	byte *loc, *texture, floorTexture[64][64][3];

	texture = PM_GetPage(num);

	for (t = 63; t >= 0; t--)
	for (s = 63; s >= 0; s--)
	{
		loc = &floorTexture[s][t][0];
		loc[0] = gamepal[(*texture)*3];
		loc[1] = gamepal[(*texture)*3+1];
		loc[2] = gamepal[(*texture)*3+2];
		texture++;
	}

	glGenTextures(1, &texture_index[num]);
	glBindTexture(GL_TEXTURE_2D, texture_index[num]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_NEAREST);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 64, 64,
		GL_RGB, GL_UNSIGNED_BYTE, floorTexture);
}

typedef struct wall_s
{
	int tex;
	float coords[4][2];
	float verts[4][3];
	struct wall_s* next;
} wall_t;

void CreateAreaList(int area)
{
	int x, y, texture, current = -1;
	float dx, dy;
	wall_t walls, *cw, *tmp;

	if (arealists[area] == 0)
		arealists[area] = glGenLists(1);
	walls.next = NULL;
	cw = &walls;

	// Find out which walls go into the linked list
	for (x = 0; x < MAPSIZE; x++)
	for (y = 0; y < MAPSIZE; y++)
	if (tilemap[x][y] == 0)
	{
		if (area != *(mapsegs[0] + farmapylookup[y]+x) - AREATILE)
			continue;

		if (pwallstate)
			if ((x == pwallx) && (y == pwally))
				continue;

		dx = x+0.5f;
		dy = y+0.5f;

		if ((x > 0) && (tilemap[x-1][y] != 0) && ((tilemap[x-1][y] & 0x80) == 0))
		{
			cw->next = (wall_t*)malloc(sizeof(wall_t));
			cw = cw->next;
			cw->tex = vertwall[tilemap[x-1][y] & ~0x40];
			cw->coords[0][0] = 0;
			cw->coords[0][1] = 0;
			cw->verts[0][0] = dx-TS/2;
			cw->verts[0][1] = dy-TS/2;
			cw->verts[0][2] = 0;
			cw->coords[1][0] = 1;
			cw->coords[1][1] = 0;
			cw->verts[1][0] = dx-TS/2;
			cw->verts[1][1] = dy+TS/2;
			cw->verts[1][2] = 0;
			cw->coords[2][0] = 1;
			cw->coords[2][1] = 1;
			cw->verts[2][0] = dx-TS/2;
			cw->verts[2][1] = dy+TS/2;
			cw->verts[2][2] = TS;
			cw->coords[3][0] = 0;
			cw->coords[3][1] = 1;
			cw->verts[3][0] = dx-TS/2;
			cw->verts[3][1] = dy-TS/2;
			cw->verts[3][2] = TS;
			cw->next = NULL;
		}

		if ((x < MAPSIZE-1) && (tilemap[x+1][y] != 0)  && ((tilemap[x+1][y] & 0x80) == 0))
		{
			cw->next = (wall_t*)malloc(sizeof(wall_t));
			cw = cw->next;
			cw->tex = vertwall[tilemap[x+1][y] & ~0x40];
			cw->coords[0][0] = 1;
			cw->coords[0][1] = 0;
			cw->verts[0][0] = dx+TS/2;
			cw->verts[0][1] = dy-TS/2;
			cw->verts[0][2] = 0;
			cw->coords[1][0] = 0;
			cw->coords[1][1] = 0;
			cw->verts[1][0] = dx+TS/2;
			cw->verts[1][1] = dy+TS/2;
			cw->verts[1][2] = 0;
			cw->coords[2][0] = 0;
			cw->coords[2][1] = 1;
			cw->verts[2][0] = dx+TS/2;
			cw->verts[2][1] = dy+TS/2;
			cw->verts[2][2] = TS;
			cw->coords[3][0] = 1;
			cw->coords[3][1] = 1;
			cw->verts[3][0] = dx+TS/2;
			cw->verts[3][1] = dy-TS/2;
			cw->verts[3][2] = TS;
			cw->next = NULL;
		}

		if ((y > 0) && (tilemap[x][y-1] != 0) && ((tilemap[x][y-1] & 0x80) == 0))
		{
			cw->next = (wall_t*)malloc(sizeof(wall_t));
			cw = cw->next;
			cw->tex = horizwall[tilemap[x][y-1] & ~0x40];
			cw->coords[0][0] = 1;
			cw->coords[0][1] = 0;
			cw->verts[0][0] = dx-TS/2;
			cw->verts[0][1] = dy-TS/2;
			cw->verts[0][2] = 0;
			cw->coords[1][0] = 0;
			cw->coords[1][1] = 0;
			cw->verts[1][0] = dx+TS/2;
			cw->verts[1][1] = dy-TS/2;
			cw->verts[1][2] = 0;
			cw->coords[2][0] = 0;
			cw->coords[2][1] = 1;
			cw->verts[2][0] = dx+TS/2;
			cw->verts[2][1] = dy-TS/2;
			cw->verts[2][2] = TS;
			cw->coords[3][0] = 1;
			cw->coords[3][1] = 1;
			cw->verts[3][0] = dx-TS/2;
			cw->verts[3][1] = dy-TS/2;
			cw->verts[3][2] = TS;
			cw->next = NULL;
		}

		if ((y < MAPSIZE-1) && (tilemap[x][y+1] != 0)  && ((tilemap[x][y+1] & 0x80) == 0))
		{
			cw->next = (wall_t*)malloc(sizeof(wall_t));
			cw = cw->next;
			cw->tex = horizwall[tilemap[x][y+1] & ~0x40];
			cw->coords[0][0] = 0;
			cw->coords[0][1] = 0;
			cw->verts[0][0] = dx-TS/2;
			cw->verts[0][1] = dy+TS/2;
			cw->verts[0][2] = 0;
			cw->coords[1][0] = 1;
			cw->coords[1][1] = 0;
			cw->verts[1][0] = dx+TS/2;
			cw->verts[1][1] = dy+TS/2;
			cw->verts[1][2] = 0;
			cw->coords[2][0] = 1;
			cw->coords[2][1] = 1;
			cw->verts[2][0] = dx+TS/2;
			cw->verts[2][1] = dy+TS/2;
			cw->verts[2][2] = TS;
			cw->coords[3][0] = 0;
			cw->coords[3][1] = 1;
			cw->verts[3][0] = dx-TS/2;
			cw->verts[3][1] = dy+TS/2;
			cw->verts[3][2] = TS;
			cw->next = NULL;
		}
	}

	if (walls.next == NULL)
	{
		if (arealists[area])
		{
			glDeleteLists(arealists[area], 1);
			arealists[area] = 0;
		}

		return;
	}

	glNewList(arealists[area], GL_COMPILE);

	while (walls.next)
	{
		texture = walls.next->tex;
		cw = &walls;
		glBindTexture(GL_TEXTURE_2D, texture_index[texture]);
		glBegin(GL_QUADS);

		while (cw)
		{
			if (cw->next == NULL)
				break;

			if (cw->next->tex == texture)
			{
				tmp = cw->next;

				glTexCoord2fv(tmp->coords[0]);
				glVertex3fv(tmp->verts[0]);
				glTexCoord2fv(tmp->coords[1]);
				glVertex3fv(tmp->verts[1]);
				glTexCoord2fv(tmp->coords[2]);
				glVertex3fv(tmp->verts[2]);
				glTexCoord2fv(tmp->coords[3]);
				glVertex3fv(tmp->verts[3]);

				cw->next = tmp->next;
				free(tmp);
			}
			else
				cw = cw->next;
		}

		glEnd();
	}
/*
	if (pwallstate)
	if (area == *(mapsegs[0] + farmapylookup[pwally]+pwallx) - AREATILE)
	{
		float delta = (float)pwallpos/63;
		x = pwallx;
		y = pwally;
		dx = x+0.5f;
		dy = y+0.5f;

		if ((x > 0) && (tilemap[x-1][y] != 0) && ((tilemap[x-1][y] & 0x80) == 0))
		{
			texture = texture_index[vertwall[tilemap[x-1][y] & ~0x40]];
			if (texture != current)
			{
				current = texture;
				glBindTexture(GL_TEXTURE_2D, texture);
			}
			glBegin(GL_QUADS);
			glTexCoord2f(0,0);
			glVertex3f(dx-TS/2,dy-TS/2,0);
			glTexCoord2f(1,0);
			glVertex3f(dx-TS/2,dy+TS/2,0);
			glTexCoord2f(1,1);
			glVertex3f(dx-TS/2,dy+TS/2,TS);
			glTexCoord2f(0,1);
			glVertex3f(dx-TS/2,dy-TS/2,TS);
			glEnd();
		}

		if ((x < MAPSIZE-1) && (tilemap[x+1][y] != 0)  && ((tilemap[x+1][y] & 0x80) == 0))
		{
			texture = texture_index[vertwall[tilemap[x+1][y] & ~0x40]];
			if (texture != current)
			{
				current = texture;
				glBindTexture(GL_TEXTURE_2D, texture);
			}
			glBegin(GL_QUADS);
			glTexCoord2f(1,0);
			glVertex3f(dx+TS/2,dy-TS/2,0);
			glTexCoord2f(0,0);
			glVertex3f(dx+TS/2,dy+TS/2,0);
			glTexCoord2f(0,1);
			glVertex3f(dx+TS/2,dy+TS/2,TS);
			glTexCoord2f(1,1);
			glVertex3f(dx+TS/2,dy-TS/2,TS);
			glEnd();
		}

		if ((y > 0) && (tilemap[x][y-1] != 0) && ((tilemap[x][y-1] & 0x80) == 0))
		{
			texture = texture_index[horizwall[tilemap[x][y-1] & ~0x40]];
			if (texture != current)
			{
				current = texture;
				glBindTexture(GL_TEXTURE_2D, texture);
			}
			glBegin(GL_QUADS);
			glTexCoord2f(1,0);
			glVertex3f(dx-TS/2,dy-TS/2,0);
			glTexCoord2f(0,0);
			glVertex3f(dx+TS/2,dy-TS/2,0);
			glTexCoord2f(0,1);
			glVertex3f(dx+TS/2,dy-TS/2,TS);
			glTexCoord2f(1,1);
			glVertex3f(dx-TS/2,dy-TS/2,TS);
			glEnd();
		}

		if ((y < MAPSIZE-1) && (tilemap[x][y+1] != 0)  && ((tilemap[x][y+1] & 0x80) == 0))
		{
			texture = texture_index[horizwall[tilemap[x][y+1] & ~0x40]];
			if (texture != current)
			{
				current = texture;
				glBindTexture(GL_TEXTURE_2D, texture);
			}
			glBegin(GL_QUADS);
			glTexCoord2f(0,0);
			glVertex3f(dx-TS/2,dy+TS/2,0);
			glTexCoord2f(1,0);
			glVertex3f(dx+TS/2,dy+TS/2,0);
			glTexCoord2f(1,1);
			glVertex3f(dx+TS/2,dy+TS/2,TS);
			glTexCoord2f(0,1);
			glVertex3f(dx-TS/2,dy+TS/2,TS);
			glEnd();
		}

		glPushMatrix();
		glTranslatef(dx, dy, 0);

		switch (pwalldir)
		{
			case di_east:
				texture = texture_index[vertwall[tilemap[x][y] & 63]];
				glRotatef(180, 0, 0, 1);
			break;

			case di_north:
				texture = texture_index[horizwall[tilemap[x][y] & 63]];
				glRotatef(90, 0, 0, 1);
			break;

			case di_west:
				texture = texture_index[vertwall[tilemap[x][y] & 63]];
			break;

			case di_south:
				texture = texture_index[horizwall[tilemap[x][y] & 63]];
				glRotatef(-90, 0, 0, 1);
			break;
		}

		glBindTexture(GL_TEXTURE_2D, texture);
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3f(TS/2-delta,-TS/2,0);
		glTexCoord2f(1,0);
		glVertex3f(TS/2-delta,+TS/2,0);
		glTexCoord2f(1,1);
		glVertex3f(TS/2-delta,+TS/2,TS);
		glTexCoord2f(0,1);
		glVertex3f(TS/2-delta,-TS/2,TS);
		glEnd();

		glPopMatrix();
	}
*/	glEndList();
}

void CreateTextures()
{
	int i, x, y, walltexture;

	if (texture_index == NULL)
	{
		MM_GetPtr ((memptr)&texture_index, sizeof(GLuint)*PMSoundStart);
		memset(texture_index, 0, sizeof(GLuint)*PMSoundStart);
	}

	for (x = 0; x < MAPSIZE; x++)
	for (y = 0; y < MAPSIZE; y++)
	if (tilemap[x][y] == 0)
	{
		if ((x > 0) && (tilemap[x-1][y] != 0) && ((tilemap[x-1][y] & 0x80) == 0))
		{
			walltexture = vertwall[tilemap[x-1][y] & ~0x40];
			if (texture_index[walltexture] == 0)
				CreateWallTexture(walltexture);
		}

		if ((x < MAPSIZE-1) && (tilemap[x+1][y] != 0)  && ((tilemap[x+1][y] & 0x80) == 0))
		{
			walltexture = vertwall[tilemap[x+1][y] & ~0x40];
			if (texture_index[walltexture] == 0)
				CreateWallTexture(walltexture);
		}

		if ((y > 0) && (tilemap[x][y-1] != 0) && ((tilemap[x][y-1] & 0x80) == 0))
		{
			walltexture = horizwall[tilemap[x][y-1] & ~0x40];
			if (texture_index[walltexture] == 0)
				CreateWallTexture(walltexture);
		}

		if ((y < MAPSIZE-1) && (tilemap[x][y+1] != 0)  && ((tilemap[x][y+1] & 0x80) == 0))
		{
			walltexture = horizwall[tilemap[x][y+1] & ~0x40];
			if (texture_index[walltexture] == 0)
				CreateWallTexture(walltexture);
		}
	}

	for (walltexture = DOORWALL; walltexture < DOORWALL+7; walltexture++)
	{
		CreateWallTexture(walltexture);

		if (walltexture == DOORWALL)
			walltexture++;
		if (walltexture == DOORWALL+4)
			walltexture++;
	}

	// Create walls
	for (i = 0; i < NUMAREAS; i++)
		CreateAreaList(i);

	// Create door arrays
	for (i = 0; i < doornum; i++)
	{
		float dx, dy;

		dx = doorobjlist[i].tilex+0.5f;
		dy = doorobjlist[i].tiley+0.5f;

		if (doorobjlist[i].vertical)
		{
			doorcoords[i*16] = 0;
			doorcoords[i*16+1] = 1;
			doorverts[i*24] = dx-TS/2;
			doorverts[i*24+1] = dy+TS/2;
			doorverts[i*24+2] = TS;
			doorcoords[i*16+2] = 0;
			doorcoords[i*16+3] = 0;
			doorverts[i*24+3] = dx-TS/2;
			doorverts[i*24+4] = dy+TS/2;
			doorverts[i*24+5] = 0;
			doorcoords[i*16+4] = 1;
			doorcoords[i*16+5] = 0;
			doorverts[i*24+6] = dx+TS/2;
			doorverts[i*24+7] = dy+TS/2;
			doorverts[i*24+8] = 0;
			doorcoords[i*16+6] = 1;
			doorcoords[i*16+7] = 1;
			doorverts[i*24+9] = dx+TS/2;
			doorverts[i*24+10] = dy+TS/2;
			doorverts[i*24+11] = TS;
			doorcoords[i*16+8] = 1;
			doorcoords[i*16+9] = 1;
			doorverts[i*24+12] = dx-TS/2;
			doorverts[i*24+13] = dy-TS/2;
			doorverts[i*24+14] = TS;
			doorcoords[i*16+10] = 1;
			doorcoords[i*16+11] = 0;
			doorverts[i*24+15] = dx-TS/2;
			doorverts[i*24+16] = dy-TS/2;
			doorverts[i*24+17] = 0;
			doorcoords[i*16+12] = 0;
			doorcoords[i*16+13] = 0;
			doorverts[i*24+18] = dx+TS/2;
			doorverts[i*24+19] = dy-TS/2;
			doorverts[i*24+20] = 0;
			doorcoords[i*16+14] = 0;
			doorcoords[i*16+15] = 1;
			doorverts[i*24+21] = dx+TS/2;
			doorverts[i*24+22] = dy-TS/2;
			doorverts[i*24+23] = TS;
		}
		else
		{
			doorcoords[i*16] = 1;
			doorcoords[i*16+1] = 1;
			doorverts[i*24] = dx+TS/2;
			doorverts[i*24+1] = dy-TS/2;
			doorverts[i*24+2] = TS;
			doorcoords[i*16+2] = 1;
			doorcoords[i*16+3] = 0;
			doorverts[i*24+3] = dx+TS/2;
			doorverts[i*24+4] = dy-TS/2;
			doorverts[i*24+5] = 0;
			doorcoords[i*16+4] = 0;
			doorcoords[i*16+5] = 0;
			doorverts[i*24+6] = dx+TS/2;
			doorverts[i*24+7] = dy+TS/2;
			doorverts[i*24+8] = 0;
			doorcoords[i*16+6] = 0;
			doorcoords[i*16+7] = 1;
			doorverts[i*24+9] = dx+TS/2;
			doorverts[i*24+10] = dy+TS/2;
			doorverts[i*24+11] = TS;
			doorcoords[i*16+8] = 0;
			doorcoords[i*16+9] = 1;
			doorverts[i*24+12] = dx-TS/2;
			doorverts[i*24+13] = dy-TS/2;
			doorverts[i*24+14] = TS;
			doorcoords[i*16+10] = 0;
			doorcoords[i*16+11] = 0;
			doorverts[i*24+15] = dx-TS/2;
			doorverts[i*24+16] = dy-TS/2;
			doorverts[i*24+17] = 0;
			doorcoords[i*16+12] = 1;
			doorcoords[i*16+13] = 0;
			doorverts[i*24+18] = dx-TS/2;
			doorverts[i*24+19] = dy+TS/2;
			doorverts[i*24+20] = 0;
			doorcoords[i*16+14] = 1;
			doorcoords[i*16+15] = 1;
			doorverts[i*24+21] = dx-TS/2;
			doorverts[i*24+22] = dy+TS/2;
			doorverts[i*24+23] = TS;
		}
	}
}

void UnloadTextures()
{
/*	int i;

	if (texture_index == NULL)
		return;

	for (i = 0; i < PMSoundStart; i++)
		if (texture_index[i] != 0)
		{
			glDeleteTextures(1, &texture_index[i]);
			texture_index[i] = 0;
		}
*/
}

void CreateSprite(int num)
{
	byte bmpbuff[64*64];
	byte alphamap[64*64];
	byte texture[64*64*4];
	
	int				idx, i, x, y;
	unsigned char	*sprite;
	unsigned short  linecmds, *cmdptr;
	short			*sprdata;
	t_compshape		*shape;
	
	sprite = PM_GetPage(num);
	shape = (t_compshape*)sprite;
	
	// set the texel index to the first texel
	i = ((((shape->rightpix-shape->leftpix)+1)*2)+4);
	// clear the buffers
	memset(bmpbuff,0,64*64);
	memset(alphamap,0,64*64);
	
	// setup a pointer to the column offsets
	cmdptr = shape->dataofs;
	for (x = shape->leftpix; x <= shape->rightpix; x++)
	{
		linecmds = *cmdptr;
		sprdata = (unsigned short *)(sprite+linecmds);
		idx = 0;
		while (sprdata[idx] != 0)
		{
			for (y = (sprdata[idx+2]/2); y < (sprdata[idx]/2); y++)
			{
				bmpbuff[((63-y)*64)+x] = sprite[i];
				alphamap[((63-y)*64)+x] = 255;
				i++;
			}
			idx += 3;
		}
		cmdptr++;
	}

	for (x = 0; x < 64*64; x++)
	{
		if (alphamap[x] == 255)
		{
			texture[x*4] = gamepal[bmpbuff[x]*3];
			texture[x*4+1] = gamepal[bmpbuff[x]*3+1];
			texture[x*4+2] = gamepal[bmpbuff[x]*3+2];
		}
		else
		{
			texture[x*4] = 0x80;
			texture[x*4+1] = 0x80;
			texture[x*4+2] = 0x80;
		}
		texture[x*4+3] = alphamap[x];
	}

	glGenTextures(1, &texture_index[num]);
	glBindTexture(GL_TEXTURE_2D, texture_index[num]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_NEAREST);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
//		GL_NEAREST);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 4, 64, 64,
		GL_RGBA, GL_UNSIGNED_BYTE, texture);
}

/*
========================
=
= TransformActor
=
= Takes paramaters:
=   gx,gy		: globalx/globaly of point
=
= globals:
=   viewx,viewy		: point of view
=   viewcos,viewsin	: sin/cos of viewangle
=   scale		: conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
========================
*/

void TransformActor(objtype *ob)
{
	fixed gx,gy,gxt,gyt,nx,ny;

	// translate point to view centered coordinates
	gx = ob->x-viewx;
	gy = ob->y-viewy;

	// calculate newx
	gxt = gx*viewcos;
	gyt = gy*viewsin;

	// fudge the shape forward a bit, because
	// the midpoint could put parts of the shape
	// into an adjacent wall
	nx = gxt-gyt-ACTORSIZE;

	// calculate newy
	gxt = gx*viewsin;
	gyt = gy*viewcos;
	ny = gyt+gxt;

	// calculate perspective ratio
	ob->transx = nx;
	ob->transy = ny;

	if (nx < mindist)
	{
		// too close, don't overflow the divide
		ob->viewheight = 0;
		return;
	}

	ob->viewx = centerx + ny*scale/nx;
	ob->viewheight = heightnumerator/(nx>>8);
}
/*
//==========================================================================

/*
========================
=
= TransformTile
=
= Takes paramaters:
=   tx,ty		: tile the object is centered in
=
= globals:
=   viewx,viewy		: point of view
=   viewcos,viewsin	: sin/cos of viewangle
=   scale		: conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
= Returns true if the tile is withing getting distance
=
========================
*/

wbool TransformTile (int tx, int ty, int *dispx, int *dispheight)
{
	fixed gx, gy, gxt, gyt, nx, ny;

	// translate point to view centered coordinates
	gx = ((long)tx<<TILESHIFT)+0x8000-viewx;
	gy = ((long)ty<<TILESHIFT)+0x8000-viewy;

	// calculate newx
	gxt = gx*viewcos;
	gyt = gy*viewsin;
	nx = gxt-gyt-0x2000;	// 0x2000 is size of object

	// calculate newy
	gxt = gx*viewsin;
	gyt = gy*viewcos;
	ny = gyt+gxt;

	// calculate perspective ratio
	if (nx < mindist)	// too close, don't overflow the divide
	{
		*dispheight = 0;
		return false;
	}

	*dispx = centerx + ny*scale/nx;
	*dispheight = heightnumerator/(nx>>8);

	// see if it should be grabbed
	if ((nx < TILEGLOBAL) && (ny > -TILEGLOBAL/2) &&
		(ny < TILEGLOBAL/2))
		return true;
	else
		return false;
}

//==========================================================================

int	CalcRotate (objtype *ob)
{
	int	angle, viewangle;

	// this isn't exactly correct, as it should vary by a trig value,
	// but it is close enough with only eight rotations
	viewangle = player->angle + (centerx - ob->viewx)/8;

	if (ob->obclass == rocketobj || ob->obclass == hrocketobj)
		angle =  (viewangle-180)- ob->angle;
	else
		angle =  (viewangle-180)- dirangle[ob->dir];

	angle+=ANGLES/16;
	while (angle>=ANGLES)
		angle-=ANGLES;
	while (angle<0)
		angle+=ANGLES;

	if (gamestates[ob->state].rotate == 2) // 2 rotation pain frame
		return 4*(angle/(ANGLES/2));       // seperated by 3 (art layout...)

	return angle/(ANGLES/8);
}

//	Draws all objects that are visable

#define MAXVISABLE	50

typedef struct
{
	int	viewx,
		viewheight,
		shapenum;
} visobj_t;

visobj_t	vislist[MAXVISABLE],*visptr,*visstep,*farthest;

void DrawScaleds()
{
	int 		i,least,numvisable,height;
	byte		*tilespot, *visspot;
	unsigned short	spotloc;
	float dx, dy;

	statobj_t	*statptr;
	objtype		*obj;

	visptr = &vislist[0];

	// place static objects
	for (statptr = &statobjlist[0] ; statptr !=laststatobj ; statptr++)
	{
		if ((visptr->shapenum = statptr->shapenum) == -1)
			continue;	// object has been deleted
/*
		if (!*statptr->visspot)
			continue;						// not visable
*/
		if (TransformTile (statptr->tilex,statptr->tiley,
			&visptr->viewx,&visptr->viewheight) && statptr->flags & FL_BONUS)
		{
			GetBonus (statptr);
			continue;
		}

		// to close to the object
		if (!visptr->viewheight)
			continue;

		// don't let it overflow
		if (visptr < &vislist[MAXVISABLE-1])	
			visptr++;
	}

	// place active objects
	for (obj = player->next;obj;obj=obj->next)
	{
		if (!(visptr->shapenum = gamestates[obj->state].shapenum))
			continue; // no shape

		// optimize: keep in struct?
		spotloc = (obj->tilex<<6)+obj->tiley;
		visspot = &spotvis[0][0]+spotloc;
		tilespot = &tilemap[0][0]+spotloc;

		// could be in any of the nine surrounding tiles
		if ((*visspot) ||
			(*(visspot-1) && !*(tilespot-1)) ||
			(*(visspot+1) && !*(tilespot+1)) ||
			(*(visspot-65) && !*(tilespot-65)) ||
			(*(visspot-64) && !*(tilespot-64)) ||
			(*(visspot-63) && !*(tilespot-63)) ||
			(*(visspot+65) && !*(tilespot+65)) ||
			(*(visspot+64) && !*(tilespot+64)) ||
			(*(visspot+63) && !*(tilespot+63)))
		{
			obj->active = true;
			TransformActor(obj);
			if (!obj->viewheight)
				continue; // too close or far away

			visptr->viewx = obj->viewx;
			visptr->viewheight = obj->viewheight;
			if (visptr->shapenum == -1)
				visptr->shapenum = obj->temp1; // special shape

			if (gamestates[obj->state].rotate)
				visptr->shapenum += CalcRotate (obj);

			if (visptr < &vislist[MAXVISABLE-1]) // don't let it overflow
				visptr++;
			obj->flags |= FL_VISABLE;
		}
		else
			obj->flags &= ~FL_VISABLE;
	}

	// draw from back to front
	numvisable = visptr-&vislist[0];

	if (!numvisable)
		return;
									// no visable objects
/*
	for (i = 0; i<numvisable; i++)
	{
		least = 32000;
		for (visstep=&vislist[0] ; visstep<visptr ; visstep++)
		{
			height = visstep->viewheight;
			if (height < least)
			{
				least = height;
				farthest = visstep;
			}
		}

		// draw farthest
		ScaleShape(farthest->viewx,farthest->shapenum,farthest->viewheight);

		farthest->viewheight = 32000;
	}
*/

// LZ

	for (statptr = &statobjlist[0] ; statptr !=laststatobj ; statptr++)
	{
		if (statptr->shapenum == -1)
			continue; // object has been deleted

		if (!areabyplayer[*(mapsegs[0] + farmapylookup[statptr->tiley]+statptr->tilex) - AREATILE])
			continue;
/*
		if (TransformTile (statptr->tilex,statptr->tiley
			,0,0) && statptr->flags & FL_BONUS)
		{
			GetBonus (statptr);
			continue;
		}
*/
		dx = (float)statptr->tilex+0.5f;
		dy = (float)statptr->tiley+0.5f;

		glPushMatrix();
		glTranslatef(dx, dy, 0);
		glRotated(-player->angle, 0,0,1);

		if (texture_index[PMSpriteStart+statptr->shapenum] == 0)
			CreateSprite(PMSpriteStart+statptr->shapenum);
		else
			glBindTexture(GL_TEXTURE_2D, texture_index[PMSpriteStart+statptr->shapenum]);

		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3f(0,-TS/2,0);
		glTexCoord2f(1,0);
		glVertex3f(0, TS/2,0);
		glTexCoord2f(1,1);
		glVertex3f(0, TS/2,TS);
		glTexCoord2f(0,1);
		glVertex3f(0,-TS/2,TS);
		glEnd();

		glPopMatrix();
	}

	for (obj = player->next; obj; obj = obj->next)
	{
		if (!areabyplayer[obj->areanumber])
			continue;
		if (!gamestates[obj->state].shapenum)
			continue;

		glPushMatrix();
		glTranslated((float)obj->x/(1<<16), (float)obj->y/(1<<16), 0);
		glRotated(-player->angle, 0,0,1);

		if (texture_index[PMSpriteStart+gamestates[obj->state].shapenum] == 0)
			CreateSprite(PMSpriteStart+gamestates[obj->state].shapenum);
		else
			glBindTexture(GL_TEXTURE_2D, texture_index[PMSpriteStart+gamestates[obj->state].shapenum]);

		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3f(0,-0.5f, 0);
		glTexCoord2f(1,0);
		glVertex3f(0, 0.5f, 0);
		glTexCoord2f(1,1);
		glVertex3f(0, 0.5f, 1.0f);
		glTexCoord2f(0,1);
		glVertex3f(0,-0.5f, 1.0f);
		glEnd();

		glPopMatrix();
	}
}

//==========================================================================

//	Draw the player's hands

int	weaponscale[NUMWEAPONS] = {SPR_KNIFEREADY,SPR_PISTOLREADY
	,SPR_MACHINEGUNREADY,SPR_CHAINREADY};
/*
void DrawPlayerWeapon (void)
{
	int	shapenum;

#ifndef SPEAR
	if (gamestate.victoryflag)
	{
		if (player->state == &s_deathcam && (TimeCount&32) )
			SimpleScaleShape(viewwidth/2,SPR_DEATHCAM,viewheight+1);
		return;
	}
#endif

	if (gamestate.weapon != -1)
	{
		shapenum = weaponscale[gamestate.weapon]+gamestate.weaponframe;
		SimpleScaleShape(viewwidth/2,shapenum,viewheight+1);
	}

	if (demorecord || demoplayback)
		SimpleScaleShape(viewwidth/2,SPR_DEMO,viewheight+1);
}
*/

//==========================================================================

void CalcTics()
{
//	long newtime;

	// calculate tics since last refresh for adaptive timing
//	if (lasttimecount > TimeCount)
//		TimeCount = lasttimecount;		// if the game was paused a LONG time

	do
	{
		tics = S_GetNewTics();
//		newtime = TimeCount;
//		tics = newtime-lasttimecount;
	} while (!tics);
			// make sure at least one tic passes

//	lasttimecount = newtime;

	if (tics > MAXTICS)
	{
//		TimeCount -= (tics-MAXTICS);
		tics = MAXTICS;
	}
}

//==========================================================================

void WallRefresh()
{
	int i, x, y, texture, current = -1;
	float dx, dy;
	static int lastpwall = 0;

	// set up variables for this view
	viewangle = player->angle;
//	midangle = viewangle*(FINEANGLES/ANGLES);
	viewsin = sintable[viewangle];
	viewcos = costable[viewangle];
	viewx = player->x - focallength*viewcos;
	viewy = player->y + focallength*viewsin;

//	focaltx = viewx>>TILESHIFT;
//	focalty = viewy>>TILESHIFT;

	if (pwallstate)
	{
		float delta = (float)pwallpos/63;
		x = pwallx;
		y = pwally;

		dx = x+0.5f;
		dy = y+0.5f;

		if ((x > 0) && (tilemap[x-1][y] != 0) && ((tilemap[x-1][y] & 0x80) == 0))
		{
			texture = texture_index[vertwall[tilemap[x-1][y] & ~0x40]];
			if (texture != current)
			{
				current = texture;
				glBindTexture(GL_TEXTURE_2D, texture);
			}
			glBegin(GL_QUADS);
			glTexCoord2f(0,0);
			glVertex3f(dx-TS/2,dy-TS/2,0);
			glTexCoord2f(1,0);
			glVertex3f(dx-TS/2,dy+TS/2,0);
			glTexCoord2f(1,1);
			glVertex3f(dx-TS/2,dy+TS/2,TS);
			glTexCoord2f(0,1);
			glVertex3f(dx-TS/2,dy-TS/2,TS);
			glEnd();
		}

		if ((x < MAPSIZE-1) && (tilemap[x+1][y] != 0)  && ((tilemap[x+1][y] & 0x80) == 0))
		{
			texture = texture_index[vertwall[tilemap[x+1][y] & ~0x40]];
			if (texture != current)
			{
				current = texture;
				glBindTexture(GL_TEXTURE_2D, texture);
			}
			glBegin(GL_QUADS);
			glTexCoord2f(1,0);
			glVertex3f(dx+TS/2,dy-TS/2,0);
			glTexCoord2f(0,0);
			glVertex3f(dx+TS/2,dy+TS/2,0);
			glTexCoord2f(0,1);
			glVertex3f(dx+TS/2,dy+TS/2,TS);
			glTexCoord2f(1,1);
			glVertex3f(dx+TS/2,dy-TS/2,TS);
			glEnd();
		}

		if ((y > 0) && (tilemap[x][y-1] != 0) && ((tilemap[x][y-1] & 0x80) == 0))
		{
			texture = texture_index[horizwall[tilemap[x][y-1] & ~0x40]];
			if (texture != current)
			{
				current = texture;
				glBindTexture(GL_TEXTURE_2D, texture);
			}
			glBegin(GL_QUADS);
			glTexCoord2f(1,0);
			glVertex3f(dx-TS/2,dy-TS/2,0);
			glTexCoord2f(0,0);
			glVertex3f(dx+TS/2,dy-TS/2,0);
			glTexCoord2f(0,1);
			glVertex3f(dx+TS/2,dy-TS/2,TS);
			glTexCoord2f(1,1);
			glVertex3f(dx-TS/2,dy-TS/2,TS);
			glEnd();
		}

		if ((y < MAPSIZE-1) && (tilemap[x][y+1] != 0)  && ((tilemap[x][y+1] & 0x80) == 0))
		{
			texture = texture_index[horizwall[tilemap[x][y+1] & ~0x40]];
			if (texture != current)
			{
				current = texture;
				glBindTexture(GL_TEXTURE_2D, texture);
			}
			glBegin(GL_QUADS);
			glTexCoord2f(0,0);
			glVertex3f(dx-TS/2,dy+TS/2,0);
			glTexCoord2f(1,0);
			glVertex3f(dx+TS/2,dy+TS/2,0);
			glTexCoord2f(1,1);
			glVertex3f(dx+TS/2,dy+TS/2,TS);
			glTexCoord2f(0,1);
			glVertex3f(dx-TS/2,dy+TS/2,TS);
			glEnd();
		}

		glPushMatrix();
		glTranslatef(dx, dy, 0);

		switch (pwalldir)
		{
			case di_east:
				texture = texture_index[vertwall[tilemap[x][y] & 63]];
				glRotatef(180, 0, 0, 1);
			break;

			case di_north:
				texture = texture_index[horizwall[tilemap[x][y] & 63]];
				glRotatef(90, 0, 0, 1);
			break;

			case di_west:
				texture = texture_index[vertwall[tilemap[x][y] & 63]];
			break;

			case di_south:
				texture = texture_index[horizwall[tilemap[x][y] & 63]];
				glRotatef(-90, 0, 0, 1);
			break;
		}

		if (texture != current)
		{
			current = texture;
			glBindTexture(GL_TEXTURE_2D, texture);
		}

		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3f(TS/2-delta,-TS/2,0);
		glTexCoord2f(1,0);
		glVertex3f(TS/2-delta,+TS/2,0);
		glTexCoord2f(1,1);
		glVertex3f(TS/2-delta,+TS/2,TS);
		glTexCoord2f(0,1);
		glVertex3f(TS/2-delta,-TS/2,TS);
		glEnd();

		glPopMatrix();
	}

	if (pwallstate)
	{
		lastpwall = *(mapsegs[0] + farmapylookup[pwally]+pwallx) - AREATILE + 1;
		CreateAreaList(lastpwall-1);
		CreateAreaList(player->areanumber);
	}
	else if (lastpwall)
	{
		CreateAreaList(lastpwall-1);
		CreateAreaList(player->areanumber);
		lastpwall = 0;
	}

	for (i = 0; i < NUMAREAS; i++)
		if (areabyplayer[i])
			glCallList(arealists[i]);

	glBindTexture(GL_TEXTURE_2D, texture_index[DOORWALL+3]);
	glVertexPointer(3, GL_FLOAT, 0, doorverts);
	glTexCoordPointer(2, GL_FLOAT, 0, doorcoords);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	// Doors
	for (i = 0; i < doornum; i++)
	{
		int num, area1, area2;
		unsigned short *map;

		map = mapsegs[0] + farmapylookup[doorobjlist[i].tiley] + doorobjlist[i].tilex;

		if (doorobjlist[i].vertical)
		{
			area1 =	*(map+1);
			area2 =	*(map-1);
		}
		else
		{
			area1 =	*(map-mapwidth);
			area2 =	*(map+mapwidth);
		}

		area1 -= AREATILE;
		area2 -= AREATILE;

		if (!areabyplayer[area1] && !areabyplayer[area2])
			continue;

		glDrawArrays(GL_QUADS, i*8, 8);
	}

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	for (i = 0; i < doornum; i++)
	{
		int num, area1, area2;
		unsigned short *map;

		map = mapsegs[0] + farmapylookup[doorobjlist[i].tiley] + doorobjlist[i].tilex;

		if (doorobjlist[i].vertical)
		{
			area1 =	*(map+1);
			area2 =	*(map-1);
		}
		else
		{
			area1 =	*(map-mapwidth);
			area2 =	*(map+mapwidth);
		}

		area1 -= AREATILE;
		area2 -= AREATILE;

		if (!areabyplayer[area1] && !areabyplayer[area2])
			continue;

		dx = doorobjlist[i].tilex+0.5f;
		dy = doorobjlist[i].tiley+0.5f;

		switch (doorobjlist[i].lock)
		{
		case dr_normal:
			num = DOORWALL;
			break;
		case dr_lock1:
		case dr_lock2:
		case dr_lock3:
		case dr_lock4:
			num = DOORWALL+6;
			break;
		case dr_elevator:
			num = DOORWALL+4;
			break;
		}

		glBindTexture(GL_TEXTURE_2D, texture_index[num]);
		if (doorobjlist[i].vertical)
		{
			dy += (float)doorposition[i]*TS/0xFFFF;
			glBegin(GL_QUADS);
			glTexCoord2f(1,1);
			glVertex3f(dx,dy-TS/2,TS);
			glTexCoord2f(1,0);
			glVertex3f(dx,dy-TS/2,0);
			glTexCoord2f(0,0);
			glVertex3f(dx,dy+TS/2,0);
			glTexCoord2f(0,1);
			glVertex3f(dx,dy+TS/2,TS);
			glEnd();
		}
		else
		{
			dx += (float)doorposition[i]*TS/0xFFFF;
			glBegin(GL_QUADS);
			glTexCoord2f(1,1);
			glVertex3f(dx-TS/2,dy,TS);
			glTexCoord2f(1,0);
			glVertex3f(dx-TS/2,dy,0);
			glTexCoord2f(0,0);
			glVertex3f(dx+TS/2,dy,0);
			glTexCoord2f(0,1);
			glVertex3f(dx+TS/2,dy,TS);
			glEnd();
		}
	}
}

void DrawStatusBar()
{
//	CheckStatusBar();

	glColor3f(1,1,1);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, statustex[0]);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2i(0,0);
	glTexCoord2f(1,0);
	glVertex2i(256, 0);
	glTexCoord2f(1,(float)40/64);
	glVertex2i(256, 40);
	glTexCoord2f(0, (float)40/64);
	glVertex2i(0, 40);
	glEnd();
	glBindTexture(GL_TEXTURE_2D, statustex[1]);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2i(256,0);
	glTexCoord2f(1,0);
	glVertex2i(320, 0);
	glTexCoord2f(1,(float)40/64);
	glVertex2i(320, 40);
	glTexCoord2f(0, (float)40/64);
	glVertex2i(256, 40);
	glEnd();
	glDisable(GL_TEXTURE_2D);
}

void CheckStatusBar()
{
	if (statustex[0] == 0)
	{
		statusdirty[0] = statusdirty[1] = true;
		glGenTextures(2, statustex);
	}

	if (statusdirty[0])
	{
		statusdirty[0] = false;
		glBindTexture(GL_TEXTURE_2D, statustex[0]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_LINEAR_MIPMAP_NEAREST);
		gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB5, 256, 64,
			GL_RGB, GL_UNSIGNED_BYTE, statusbar1);
	}

	if (statusdirty[1])
	{
		statusdirty[1] = false;
		glBindTexture(GL_TEXTURE_2D, statustex[1]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_LINEAR_MIPMAP_NEAREST);
		gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB5, 64, 64,
			GL_RGB, GL_UNSIGNED_BYTE, statusbar2);
	}
}

#define NUMREDSHIFTS	6
#define NUMWHITESHIFTS	6
#define WHITETICS		6

void UpdatePaletteShifts()
{
	extern int damagecount, bonuscount;
	int	red, white;
	float clr[4];

	if (bonuscount)
	{
		white = bonuscount/10 +1;
		if (white > NUMWHITESHIFTS)
			white = NUMWHITESHIFTS;
		bonuscount -= tics;
		if (bonuscount < 0)
			bonuscount = 0;
	}
	else
		white = 0;


	if (damagecount)
	{
		red = damagecount/10 +1;
		if (red > NUMREDSHIFTS)
			red = NUMREDSHIFTS;

		damagecount -= tics;
		if (damagecount < 0)
			damagecount = 0;
	}
	else
		red = 0;

	if (red)
	{
		clr[0] = 1.0f;
		clr[1] = 0.0f;
		clr[2] = 0.0f;
		clr[3] = (float)red/NUMREDSHIFTS;
	}
	else if (white)
	{
		clr[0] = 1.0f;
		clr[1] = 1.0f;
		clr[2] = 1.0f;
		clr[3] = (float)white/NUMWHITESHIFTS;
	}
	else
		return;

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glColor4fv(clr);
	glBegin(GL_QUADS);
	glVertex2i(0, 0);
	glVertex2i(320, 0);
	glVertex2i(320, 200);
	glVertex2i(0, 200);
	glEnd();
	glDisable(GL_BLEND);
}

void ThreeDRefresh()
{
	float tx,ty, aspect = (float)screen_width/screen_height;
	float color = 0;
	int texture, current = 0;
	int	xl,yl;
	xl = 160-viewwidth/2;
	yl = (200-STATUSLINES-viewheight)/2+STATUSLINES;

	glClearColor((float)gamepal[127*3]/255,
		(float)gamepal[127*3+1]/255, (float)gamepal[127*3+2]/255,1);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glDepthMask(GL_FALSE);
	glBegin(GL_QUADS);

	glColor3ubv(&gamepal[0x19*3]);
	glVertex2i(xl, yl);
	glVertex2i(xl+viewwidth, yl);
	glVertex2i(xl+viewwidth, yl+viewheight/2);
	glVertex2i(xl, yl+viewheight/2);

	glColor3ubv(&gamepal[vgaCeiling[gamestate.episode*10+mapon]*3]);
	glVertex2i(xl, yl+viewheight/2);
	glVertex2i(xl+viewwidth, yl+viewheight/2);
	glVertex2i(xl+viewwidth, yl+viewheight);
	glVertex2i(xl, yl+viewheight);

	glEnd();
	glDepthMask(GL_TRUE);

	glViewport(xl*screen_width/320, yl*screen_height/200, viewwidth*screen_width/320, viewheight*screen_height/200);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45, aspect, 0.1f, 25.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScalef(-1,1,1.2f);

	tx = 5*costable[player->angle];
	ty = -5*sintable[player->angle];
//	gluLookAt(player->x,player->y,20*TS, player->x,player->y,0, 1,0,0);
	gluLookAt(0,0,0, tx,ty,0, 0,0,1);
	glTranslated(-(float)player->x/(1<<16), -(float)player->y/(1<<16), -TS/2);

	glColor3f(1,1,1);
	glEnable(GL_TEXTURE_2D);

	WallRefresh();

	glAlphaFunc(GL_GREATER, r_alpharef);
	glEnable(GL_ALPHA_TEST);

	// draw all the scaled images
	DrawScaleds();			// draw scaled stuff
/*
	for (statptr = &statobjlist[0] ; statptr !=laststatobj ; statptr++)
	{
		long dx, dy;

		if (statptr->shapenum == -1)
			continue;	// object has been deleted

		if (TransformTile (statptr->tilex,statptr->tiley
			,0,0) && statptr->flags & FL_BONUS)
		{
			GetBonus (statptr);
			continue;
		}

		dx = ((long)statptr->tilex<<TILESHIFT)+TILEGLOBAL/2;
		dy = ((long)statptr->tiley<<TILESHIFT)+TILEGLOBAL/2;

		glPushMatrix();
		glTranslated(dx, dy, 0);
		glRotated(-player->angle, 0,0,1);

		if (texture_index[PMSpriteStart+statptr->shapenum] == 0)
			CreateSprite(PMSpriteStart+statptr->shapenum);
		else
			glBindTexture(GL_TEXTURE_2D, texture_index[PMSpriteStart+statptr->shapenum]);

		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3i(0,-TS/2,0);
		glTexCoord2f(1,0);
		glVertex3i(0, TS/2,0);
		glTexCoord2f(1,1);
		glVertex3i(0, TS/2,TS);
		glTexCoord2f(0,1);
		glVertex3i(0,-TS/2,TS);
		glEnd();

		glPopMatrix();
	}

	for (ob = player->next; ob; ob = ob->next)
	{
		if (!areabyplayer[ob->areanumber])
			continue;
		if (!ob->state->shapenum)
			continue;

		glPushMatrix();
		glTranslated(ob->x, ob->y, 0);
		glRotated(-player->angle, 0,0,1);

		if (texture_index[PMSpriteStart+ob->state->shapenum] == 0)
			CreateSprite(PMSpriteStart+ob->state->shapenum);
		else
			glBindTexture(GL_TEXTURE_2D, texture_index[PMSpriteStart+ob->state->shapenum]);

		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3i(0,-TS/2,0);
		glTexCoord2f(1,0);
		glVertex3i(0, TS/2,0);
		glTexCoord2f(1,1);
		glVertex3i(0, TS/2,TS);
		glTexCoord2f(0,1);
		glVertex3i(0,-TS/2,TS);
		glEnd();

		glPopMatrix();

	}
*/
	glViewport(0, 0, screen_width, screen_height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 320, 0, 200, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

//	DrawPlayerWeapon ();	// draw player's hands
/*
	// Player weapon
#ifndef SPEAR
	if (gamestate.victoryflag)
	{
		if (player->state == &s_deathcam && (TimeCount&32) )
			SimpleScaleShape(viewwidth/2,SPR_DEATHCAM,viewheight+1);
		return;
	}
#endif
*/
	if (gamestate.weapon != -1)
	{
		texture = PMSpriteStart+weaponscale[gamestate.weapon]+gamestate.weaponframe;
//		SimpleScaleShape(viewwidth/2,num,viewheight+1);

		if (texture_index[texture] == 0)
			CreateSprite(texture);
		else
			glBindTexture(GL_TEXTURE_2D, texture_index[texture]);

		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex2i(xl+viewwidth/4, yl);
		glTexCoord2f(1,0);
		glVertex2i(xl+3*viewwidth/4, yl);
		glTexCoord2f(1,1);
		glVertex2i(xl+3*viewwidth/4, yl+viewheight);
		glTexCoord2f(0,1);
		glVertex2i(xl+viewwidth/4, yl+viewheight);
		glEnd();
	}
/*
	if (demorecord || demoplayback)
		SimpleScaleShape(viewwidth/2,SPR_DEMO,viewheight+1);
*/

	glDisable(GL_ALPHA_TEST);

	DrawStatusBar();

	glDisable(GL_TEXTURE_2D);

	glBegin(GL_LINE_STRIP);
	glColor3ubv(&gamepal[125*3]);
	glVertex2i(xl, yl);
	glVertex2i(xl+viewwidth, yl);
	glVertex2i(xl+viewwidth, yl+viewheight);
	glEnd();

	glBegin(GL_LINE_STRIP);
	glColor3ubv(&gamepal[0]);
	glVertex2i(xl, yl);
	glVertex2i(xl, yl+viewheight);
	glVertex2i(xl+viewwidth, yl+viewheight);
	glEnd();

	Con_Draw();

	UpdatePaletteShifts();
	S_FinishDisplay();

/*
	// show screen and time last cycle
	if (fizzlein)
	{
		FizzleFade(bufferofs,displayofs+screenofs,viewwidth,viewheight,20,false);
		fizzlein = false;

		lasttimecount = TimeCount = 0;		// don't make a big tic count
	}
*/
	frameon++;
	PM_NextFrame();
}
