// WL_MAIN.C

#include <memory.h>
#include <string.h>
#include <stdlib.h>
#include "opengl.h"
#include "wl_def.h"
#include "system.h"
#include "console.h"

//	LOCAL CONSTANTS

#define FOCALLENGTH	(0x5700l)	// in global coordinates
#define VIEWGLOBAL	0x10000		// globals visable flush to wall

/*
#define VIEWWIDTH       256        // size of view window
#define VIEWHEIGHT      144
*/
//	GLOBAL VARIABLES

char		str[80],str2[20];

int			tedlevelnum;
wbool		tedlevel;
int	   dirangle[9] = {0,ANGLES/8,2*ANGLES/8,3*ANGLES/8,4*ANGLES/8,
	5*ANGLES/8,6*ANGLES/8,7*ANGLES/8,ANGLES};

// projection variables
fixed           focallength;
int             viewwidth;
int             viewheight;
int             centerx;
int             shootdelta; // pixels away from centerx a target can be
fixed           scale,maxslope;
long            heightnumerator;
int             minheightdiv;

void            Quit (char *error);

wbool			startgame, loadedgame;
int             mouseadjustment;

char			configname[13]="glconfig.";

int _argc;
char** _argv;

void ReadConfig()
{
	FILE*	file;
	SDMode	sd;
	SMMode	sm;
	SDSMode	sds;
	int i;

	if ((file = fopen(configname, "rb")) != 0)
	{
		// valid config file
		for (i = 0; i < MaxScores; i++)
		{
			fread(&Scores[i].name, MaxHighName+1, 1, file);
			fread(&Scores[i].score, 4, 1, file);
			fread(&Scores[i].completed, 2, 1, file);
			fread(&Scores[i].episode, 2, 1, file);
		}

		fread(&sd,sizeof(sd), 1, file);
		fread(&sm,sizeof(sm), 1, file);
		fread(&sds,sizeof(sds), 1, file);

		fread(&mouseenabled,sizeof(mouseenabled), 1, file);
		fread(&joystickenabled,sizeof(joystickenabled), 1, file);
		fread(&joypadenabled,sizeof(joypadenabled), 1, file);
		fread(&joystickport,sizeof(joystickport), 1, file);

		fread(&dirscan,sizeof(dirscan), 1, file);
		fread(&buttonscan,sizeof(buttonscan), 1, file);
		fread(&buttonmouse,sizeof(buttonmouse), 1, file);
		fread(&buttonjoy,sizeof(buttonjoy), 1, file);

		fread(&viewsize,sizeof(viewsize), 1, file);
		fread(&mouseadjustment,sizeof(mouseadjustment), 1, file);

		fclose(file);

		if (sd == sdm_AdLib && !SoundBlasterPresent)
		{
			sd = sdm_Off;
			sm = smm_Off;
		}

		if (sds == sds_SoundBlaster && !SoundBlasterPresent)
			sds = sds_Off;

		if (!MousePresent)
			mouseenabled = false;
		if (!JoysPresent[joystickport])
			joystickenabled = false;

		if (viewsize < 4 || viewsize > 19)
			viewsize = 15;

		MainMenu[6].active=1;
		MainItems.curpos=0;
	}
	else
	{
		// no config file, so select by hardware
		if (SoundBlasterPresent || AdLibPresent)
		{
			sd = sdm_AdLib;
			sm = smm_AdLib;
		}
		else
		{
			sd = sdm_Off;
			sm = smm_Off;
		}

		if (SoundBlasterPresent)
			sds = sds_SoundBlaster;
		else
			sds = sds_Off;

		if (MousePresent)
			mouseenabled = true;

		joystickenabled = false;
		joypadenabled = false;
		joystickport = 0;

		viewsize = 15;
		mouseadjustment=5;
	}

	SD_SetMusicMode(sm);
	SD_SetSoundMode(sd);
	SD_SetDigiDevice(sds);
}

void WriteConfig()
{
	FILE* file;
	int i;

	file = fopen(configname, "wb");

	if (file != 0)
	{
		// valid config file
		for (i = 0; i < MaxScores; i++)
		{
			fwrite(&Scores[i].name, MaxHighName+1, 1, file);
			fwrite(&Scores[i].score, 4, 1, file);
			fwrite(&Scores[i].completed, 2, 1, file);
			fwrite(&Scores[i].episode, 2, 1, file);
		}

		fwrite(&SoundMode, sizeof(SoundMode), 1, file);
		fwrite(&MusicMode, sizeof(MusicMode), 1, file);
		fwrite(&DigiMode, sizeof(DigiMode), 1, file);

		fwrite(&mouseenabled,sizeof(mouseenabled), 1, file);
		fwrite(&joystickenabled,sizeof(joystickenabled), 1, file);
		fwrite(&joypadenabled,sizeof(joypadenabled), 1, file);
		fwrite(&joystickport,sizeof(joystickport), 1, file);

		fwrite(&dirscan,sizeof(dirscan), 1, file);
		fwrite(&buttonscan,sizeof(buttonscan), 1, file);
		fwrite(&buttonmouse,sizeof(buttonmouse), 1, file);
		fwrite(&buttonjoy,sizeof(buttonjoy), 1, file);

		fwrite(&viewsize,sizeof(viewsize), 1, file);
		fwrite(&mouseadjustment,sizeof(mouseadjustment), 1, file);

		fclose(file);
	}
}

//===========================================================================

//	Set up new game to start from the beginning
void NewGame(int difficulty, int episode)
{
	memset (&gamestate,0,sizeof(gamestate));
	gamestate.difficulty = difficulty;
	gamestate.weapon = gamestate.bestweapon
		= gamestate.chosenweapon = wp_pistol;
	gamestate.health = 100;
	gamestate.ammo = STARTAMMO;
	gamestate.lives = 3;
	gamestate.nextextra = EXTRAPOINTS;
	gamestate.episode=episode;

	startgame = true;
}

//===========================================================================

void DiskFlopAnim(int x,int y)
{
/*
	static char which = 0;
	if (!x && !y)
		return;
	VWB_DrawPic(x,y,C_DISKLOADING1PIC+which);
	VW_UpdateScreen();
	which^=1;
*/
}


long DoChecksum(byte *source, unsigned short size, long checksum)
{
	unsigned short i;

	for (i=0;i<size-1;i++)
		checksum += source[i]^source[i+1];

	return checksum;
}

//===========================================================================

#define WriteByte(a) fwrite(&a, 1, 1, file)
#define WriteWord(a) fwrite(&a, 2, 1, file)
#define WriteLong(a) fwrite(&a, 4, 1, file)

wbool SaveTheGame(FILE* file, int x,int y)
{
	int i;
	objtype *ob, nullobj;
	long checksum;
/*
	struct diskfree_t dfree;
	long avail,size;

	if (_dos_getdiskfree(0,&dfree))
	  Quit("Error in _dos_getdiskfree call");

	avail = (long)dfree.avail_clusters *
			dfree.bytes_per_sector *
			dfree.sectors_per_cluster;

	size = 0;
	for (ob = player; ob ; ob=ob->next)
	  size += sizeof(*ob);
	size += sizeof(nullobj);

	size += sizeof(gamestate) +
			sizeof(LRstruct)*8 +
			sizeof(tilemap) +
			sizeof(actorat) +
			sizeof(laststatobj) +
			sizeof(statobjlist) +
			sizeof(doorposition) +
			sizeof(pwallstate) +
			sizeof(pwallx) +
			sizeof(pwally) +
			sizeof(pwalldir) +
			sizeof(pwallpos);

	if (avail < size)
	{
	 Message(STR_NOSPACE1"\n"
			 STR_NOSPACE2);
	 return false;
	}
*/
	checksum = 0;

	DiskFlopAnim(x,y);

	WriteWord(gamestate.difficulty);
	WriteWord(gamestate.mapon);
	WriteLong(gamestate.oldscore);
	WriteLong(gamestate.score);
	WriteLong(gamestate.nextextra);
	WriteWord(gamestate.lives);
	WriteWord(gamestate.health);
	WriteWord(gamestate.ammo);
	WriteWord(gamestate.keys);
	WriteWord(gamestate.bestweapon);
	WriteWord(gamestate.weapon);
	WriteWord(gamestate.chosenweapon);
	WriteWord(gamestate.faceframe);
	WriteWord(gamestate.attackframe);
	WriteWord(gamestate.attackcount);
	WriteWord(gamestate.weaponframe);
	WriteWord(gamestate.episode);
	WriteWord(gamestate.secretcount);
	WriteWord(gamestate.treasurecount);
	WriteWord(gamestate.killcount);
	WriteWord(gamestate.secrettotal);
	WriteWord(gamestate.treasuretotal);
	WriteWord(gamestate.killtotal);
	WriteLong(gamestate.TimeCount);
	WriteLong(gamestate.killx);
	WriteLong(gamestate.killy);
	WriteLong(gamestate.victoryflag);
	checksum = DoChecksum((byte*)&gamestate, sizeof(gamestate), checksum);

	DiskFlopAnim(x,y);
#ifdef SPEAR
	for (i = 0; i < 20; i++)
#else
	for (i = 0; i < 8; i++)
#endif
	{
		WriteWord(LevelRatios[i].kill);
		WriteWord(LevelRatios[i].secret);
		WriteWord(LevelRatios[i].treasure);
		WriteLong(LevelRatios[i].time);
		checksum = DoChecksum((byte*)&LevelRatios[i], sizeof(LRstruct), checksum);
	}

	DiskFlopAnim(x,y);
	CA_FarWrite(file, (void*)tilemap, sizeof(tilemap));
	checksum = DoChecksum((byte*)tilemap, sizeof(tilemap), checksum);
	DiskFlopAnim(x,y);
	CA_FarWrite(file, (void*)actorat, sizeof(actorat));
	checksum = DoChecksum((byte*)actorat, sizeof(actorat), checksum);

	CA_FarWrite(file, (void*)areaconnect, sizeof(areaconnect));
	CA_FarWrite(file, (void*)areabyplayer, sizeof(areabyplayer));

	for (ob = player; ob; ob = ob->next)
	{
		DiskFlopAnim(x,y);
		CA_FarWrite(file, (void*)ob, sizeof(*ob));
	}
	nullobj.active = ac_badobject;          // end of file marker
	DiskFlopAnim(x,y);
	CA_FarWrite (file, (void*)&nullobj, sizeof(nullobj));

	DiskFlopAnim(x,y);
	i = laststatobj  - statobjlist;
	CA_FarWrite (file, (void*)&i, sizeof(i));
	checksum = DoChecksum((byte*)&i, sizeof(i), checksum);

	DiskFlopAnim(x,y);
	CA_FarWrite (file, (void*)statobjlist, sizeof(statobjlist));
	checksum = DoChecksum((byte*)statobjlist, sizeof(statobjlist),checksum);

	DiskFlopAnim(x,y);
	CA_FarWrite(file,(void*)doorposition, sizeof(doorposition));
	checksum = DoChecksum((byte*)doorposition, sizeof(doorposition),checksum);
	DiskFlopAnim(x,y);
	CA_FarWrite(file,(void*)doorobjlist, sizeof(doorobjlist));
	checksum = DoChecksum((byte*)doorobjlist, sizeof(doorobjlist),checksum);

	DiskFlopAnim(x,y);
	CA_FarWrite (file, (void*)&pwallstate, sizeof(pwallstate));
	checksum = DoChecksum((byte*)&pwallstate, sizeof(pwallstate),checksum);
	CA_FarWrite (file, (void*)&pwallx, sizeof(pwallx));
	checksum = DoChecksum((byte*)&pwallx, sizeof(pwallx),checksum);
	CA_FarWrite (file, (void*)&pwally, sizeof(pwally));
	checksum = DoChecksum((byte*)&pwally, sizeof(pwally),checksum);
	CA_FarWrite (file, (void*)&pwalldir, sizeof(pwalldir));
	checksum = DoChecksum((byte*)&pwalldir, sizeof(pwalldir),checksum);
	CA_FarWrite (file, (void*)&pwallpos, sizeof(pwallpos));
	checksum = DoChecksum((byte*)&pwallpos, sizeof(pwallpos),checksum);

	// WRITE OUT CHECKSUM
	CA_FarWrite (file, (void*)&checksum, sizeof(checksum));

	return true;
}

//===========================================================================

#define ReadByte(a) fread(&a, 1, 1, file)
#define ReadWord(a) fread(&a, 2, 1, file)
#define ReadLong(a) fread(&a, 4, 1, file)

wbool LoadTheGame(FILE* file, int x, int y)
{
	int i;
	long checksum, oldchecksum;
	objtype nullobj;

	checksum = 0;

	DiskFlopAnim(x,y);

	ReadWord(gamestate.difficulty);
	ReadWord(gamestate.mapon);
	ReadLong(gamestate.oldscore);
	ReadLong(gamestate.score);
	ReadLong(gamestate.nextextra);
	ReadWord(gamestate.lives);
	ReadWord(gamestate.health);
	ReadWord(gamestate.ammo);
	ReadWord(gamestate.keys);
	ReadWord(gamestate.bestweapon);
	ReadWord(gamestate.weapon);
	ReadWord(gamestate.chosenweapon);
	ReadWord(gamestate.faceframe);
	ReadWord(gamestate.attackframe);
	ReadWord(gamestate.attackcount);
	ReadWord(gamestate.weaponframe);
	ReadWord(gamestate.episode);
	ReadWord(gamestate.secretcount);
	ReadWord(gamestate.treasurecount);
	ReadWord(gamestate.killcount);
	ReadWord(gamestate.secrettotal);
	ReadWord(gamestate.treasuretotal);
	ReadWord(gamestate.killtotal);
	ReadLong(gamestate.TimeCount);
	ReadLong(gamestate.killx);
	ReadLong(gamestate.killy);
	ReadLong(gamestate.victoryflag);
	checksum = DoChecksum((byte*)&gamestate,sizeof(gamestate),checksum);

	DiskFlopAnim(x,y);
#ifdef SPEAR
	for (i = 0; i < 20; i++)
#else
	for (i = 0; i < 8; i++)
#endif
	{
		ReadWord(LevelRatios[i].kill);
		ReadWord(LevelRatios[i].secret);
		ReadWord(LevelRatios[i].treasure);
		ReadLong(LevelRatios[i].time);
		checksum = DoChecksum((byte*)&LevelRatios[i], sizeof(LRstruct), checksum);
	}

	DiskFlopAnim(x,y);
	SetupGameLevel();

	DiskFlopAnim(x,y);
	CA_FarRead(file, (void*)tilemap, sizeof(tilemap));
	checksum = DoChecksum((byte*)tilemap,sizeof(tilemap), checksum);
	DiskFlopAnim(x,y);
	CA_FarRead(file, (void*)actorat, sizeof(actorat));
	checksum = DoChecksum((byte*)actorat, sizeof(actorat), checksum);

	CA_FarRead(file, (void*)areaconnect, sizeof(areaconnect));
	CA_FarRead(file, (void*)areabyplayer, sizeof(areabyplayer));

	InitActorList();
	DiskFlopAnim(x,y);
	CA_FarRead(file, (void*)player, sizeof(*player));

	while (1)
	{
		DiskFlopAnim(x,y);
		CA_FarRead(file, (void*)&nullobj, sizeof(nullobj));
/*
		ReadWord(nullobj.active);
		ReadWord(nullobj.ticcount);
		ReadWord(nullobj.obclass);
		ReadWord(nullobj.state); //	statetype	*state;
		ReadByte(nullobj.flags);
		ReadLong(nullobj.distance);
		ReadWord(nullobj.dir);
		ReadLong(nullobj.x);
		ReadLong(nullobj.y);
		ReadWord(nullobj.tilex);
		ReadWord(nullobj.tiley);
		ReadByte(nullobj.areanumber);
		ReadWord(nullobj.viewx);
		ReadWord(nullobj.viewheight);
		ReadLong(nullobj.transx);
		ReadLong(nullobj.transy);
		ReadWord(nullobj.angle);
		ReadWord(nullobj.hitpoints);
		ReadLong(nullobj.speed);
		ReadWord(nullobj.temp1);
		ReadWord(nullobj.temp2);
		ReadWord(nullobj.temp3);
		ReadWord((short)nullobj.next);
		ReadWord((short)nullobj.prev);
*/
		if (nullobj.active == ac_badobject)
			break;
		GetNewActor(); // don't copy over the links

		new->active = nullobj.active;
		new->ticcount = nullobj.ticcount;
		new->obclass = nullobj.obclass;
		new->state = nullobj.state;
		new->flags = nullobj.flags;
		new->distance = nullobj.distance;
		new->dir = nullobj.dir;
		new->x = nullobj.x;
		new->y = nullobj.y;
		new->tilex = nullobj.tilex;
		new->tiley = nullobj.tiley;
		new->areanumber = nullobj.areanumber;
		new->viewx = nullobj.viewx;
		new->viewheight = nullobj.viewheight;
		new->transx = nullobj.transx;
		new->transy = nullobj.transy;
		new->angle = nullobj.angle;
		new->hitpoints = nullobj.hitpoints;
		new->speed = nullobj.speed;
		new->temp1 = nullobj.temp1;
		new->temp2 = nullobj.temp2;
		new->temp3 = nullobj.temp3;

//		memcpy(new, &nullobj, sizeof(nullobj)-8);
	}

	DiskFlopAnim(x,y);
	CA_FarRead (file, (void*)&i, sizeof(i));
	checksum = DoChecksum((byte*)&i, sizeof(i), checksum);
	laststatobj = statobjlist + i;

	DiskFlopAnim(x,y);
	CA_FarRead (file, (void*)statobjlist, sizeof(statobjlist));
	checksum = DoChecksum((byte*)statobjlist, sizeof(statobjlist),checksum);

	DiskFlopAnim(x,y);
	CA_FarRead(file,(void*)doorposition, sizeof(doorposition));
	checksum = DoChecksum((byte*)doorposition, sizeof(doorposition),checksum);
	DiskFlopAnim(x,y);
	CA_FarRead(file,(void*)doorobjlist, sizeof(doorobjlist));
	checksum = DoChecksum((byte*)doorobjlist, sizeof(doorobjlist),checksum);

	DiskFlopAnim(x,y);
	CA_FarRead(file, (void*)&pwallstate, sizeof(pwallstate));
	checksum = DoChecksum((byte*)&pwallstate, sizeof(pwallstate),checksum);
	CA_FarRead(file, (void*)&pwallx, sizeof(pwallx));
	checksum = DoChecksum((byte*)&pwallx, sizeof(pwallx),checksum);
	CA_FarRead(file, (void*)&pwally, sizeof(pwally));
	checksum = DoChecksum((byte*)&pwally, sizeof(pwally),checksum);
	CA_FarRead(file, (void*)&pwalldir, sizeof(pwalldir));
	checksum = DoChecksum((byte*)&pwalldir,sizeof(pwalldir),checksum);
	CA_FarRead(file, (void*)&pwallpos, sizeof(pwallpos));
	checksum = DoChecksum((byte*)&pwallpos, sizeof(pwallpos),checksum);

	CA_FarRead(file, (void*)&oldchecksum, sizeof(oldchecksum));

	if (oldchecksum != checksum)
	{
/*
		Message(STR_SAVECHT1"\n"
			 STR_SAVECHT2"\n"
			 STR_SAVECHT3"\n"
			 STR_SAVECHT4);

		IN_ClearKeysDown();
		IN_Ack();
*/
		gamestate.score = 0;
		gamestate.lives = 1;
		gamestate.weapon =
		gamestate.chosenweapon =
		gamestate.bestweapon = wp_pistol;
		gamestate.ammo = 8;
	}

	return true;
}

//===========================================================================

// Shuts down all ID_?? managers
void ShutdownId()
{
	US_Shutdown();
	SD_Shutdown();
	PM_Shutdown();
	IN_Shutdown();
	VL_Shutdown();
	CA_Shutdown();
	MM_Shutdown();
	GL_Shutdown();
	S_Shutdown();
}

//===========================================================================

//	Calculates:
//	scale                 projection constant
//	sintable/costable     overlapping fractional tables

const float radtoint = (float)FINEANGLES/2/PI;

void BuildTables()
{
  int		i;
  float		angle, anglestep;
  double	tang;
  float		value;

	// calculate fine tangents
	for (i=0;i<FINEANGLES/8;i++)
	{
		tang = tan( (i+0.5)/radtoint);
		finetangent[i] = tang*TILEGLOBAL;
		finetangent[FINEANGLES/4-1-i] = 1/tang*TILEGLOBAL;
	}

	// costable overlays sintable with a quarter phase shift
	// ANGLES is assumed to be divisable by four
	angle = 0;
	anglestep = PI/2/ANGLEQUAD;
	for (i = 0; i <= ANGLEQUAD; i++)
	{
		value = sin(angle);
		sintable[i]=
			sintable[i+ANGLES]=
			sintable[ANGLES/2-i] = value;
		sintable[ANGLES-i]=
			sintable[ANGLES/2+i] = -value;
		angle += anglestep;
	}
}

//===========================================================================

//	Uses focallength
void CalcProjection(long focal)
{
	int     i;
	long    intang;
	float   angle;
	double  tang;
	int     halfview;
	double  facedist;

	focallength = focal;
	facedist = focal+MINDIST;
	// half view in pixels
	halfview = viewwidth/2;

	// calculate scale value for vertical height calculations
	// and sprite x calculations
	scale = halfview*facedist/(VIEWGLOBAL/2);

	// divide heightnumerator by a posts distance to get the posts height for
	// the heightbuffer.  The pixel height is height>>2
	heightnumerator = (TILEGLOBAL*scale)>>6;
	minheightdiv = heightnumerator/0x7fff +1;

	// calculate the angle offset from view angle of each pixel's ray
	for (i=0;i<halfview;i++)
	{
		// start 1/2 pixel over, so viewangle bisects two middle pixels
		tang = (long)i*VIEWGLOBAL/viewwidth/facedist;
		angle = atan(tang);
		intang = angle*radtoint;
		pixelangle[halfview-1-i] = intang;
		pixelangle[halfview+i] = -intang;
	}

	// if a point's abs(y/x) is greater than maxslope, the point is outside
	// the view area
	maxslope = finetangent[pixelangle[0]];
	maxslope >>= 8;
}

//===========================================================================

//	Map tile values to scaled pics
void SetupWalls ()
{
	int i;

	for (i = 1; i < MAXWALLTILES; i++)
	{
		horizwall[i]=(i-1)*2;
		vertwall[i]=(i-1)*2+1;
	}
}

//===========================================================================

void SignonScreen()
{
/*
	int	i, x, y;
	unsigned short width, height;
	long sx, sy;
	byte *source, *buffer;
	GLuint tex;
	int l, b;

	source = signon;
	sx = width = 320;
	sy = height = 200;
	l = b = 0;

	for (i = 0; ((i < 16) && (sx != 0)); i++)
		sx = sx >> 1;
	sx = (i != 0) ? 1 << i : 1;

	for (i = 0; ((i < 16) && (sy != 0)); i++)
		sy = sy >> 1;
	sy = (i != 0) ? 1 << i : 1;

	buffer = malloc(sx*sy*3);

	for (y = height-1; y >= 0; y--)
	for (x = 0; x < width; x++)
	{
		buffer[(y*sx+x)*3] = gamepal[(*source)*3];
		buffer[(y*sx+x)*3+1] = gamepal[(*source)*3+1];
		buffer[(y*sx+x)*3+2] = gamepal[(*source)*3+2];
		source++;
	}

	for (y = height; y < sy; y++)
	for (x = 0; x < sx; x++)
	{
		buffer[(y*sx+x)*3] = buffer[((y-1)*sx+x)*3];
		buffer[(y*sx+x)*3+1] = buffer[((y-1)*sx+x)*3+1];
		buffer[(y*sx+x)*3+2] = buffer[((y-1)*sx+x)*3+2];
	}

	for (x = width; x < sx; x++)
	for (y = 0; y < height; y++)
	{
		buffer[(y*sx+x)*3] = buffer[(y*sx+x-1)*3];
		buffer[(y*sx+x)*3+1] = buffer[(y*sx+x-1)*3+1];
		buffer[(y*sx+x)*3+2] = buffer[(y*sx+x-1)*3+2];
	}

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_NEAREST);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, sx, sy,
		GL_RGB, GL_UNSIGNED_BYTE, buffer);
	free(buffer);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 320, 0, 200, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glColor3f(1,1,1);
	glEnable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2i(l, b);
	glTexCoord2f((float)width/sx,0);
	glVertex2i(l+width, b);
	glTexCoord2f((float)width/sx,(float)height/sy);
	glVertex2i(l+width, b+height);
	glTexCoord2f(0, (float)height/sy);
	glVertex2i(l, b+height);
	glEnd();
	glDisable(GL_TEXTURE_2D);

	glDeleteTextures(1, &tex);
	S_FinishDisplay();

//	if (IN_UserInput(TickBase*10))
//		;
*/

/*	unsigned short  segstart,seglength;

	VL_SetVGAPlaneMode ();
	VL_TestPaletteSet ();
	VL_SetPalette (gamepal);

	VW_SetScreen(0x8000,0);
	VL_MungePic (&introscn,320,200);
	VL_MemToScreen (&introscn,320,200,0,0);
	VW_SetScreen(0,0);

	// reclaim the memory from the linked in signon screen
	segstart = FP_SEG(&introscn);
	seglength = 64000/16;
	if (FP_OFF(&introscn))
	{
		segstart++;
		seglength--;
	}
	MML_UseSpace (segstart,seglength);
*/
}

void FinishSignon()
{
/*
#ifndef SPEAR
	VW_Bar (0,189,300,11,peekb(0xa000,0));
	WindowX = 0;
	WindowW = 320;
	PrintY = 190;

	#ifndef JAPAN
	SETFONTCOLOR(14,4);

	#ifdef SPANISH
	US_CPrint ("Oprima una tecla");
	#else
	US_CPrint ("Press a key");
	#endif

	#endif

	if (!NoWait)
		IN_Ack ();

	#ifndef JAPAN
	VW_Bar (0,189,300,11,peekb(0xa000,0));

	PrintY = 190;
	SETFONTCOLOR(10,4);

	#ifdef SPANISH
	US_CPrint ("pensando...");
	#else
	US_CPrint ("Working...");
	#endif

	#endif

	SETFONTCOLOR(0,15);
#else
	if (!NoWait)
		VW_WaitVBL(3*70);
#endif
*/
}

//===========================================================================

wbool MS_CheckParm(char *check)
{
	int i;
	char *parm;

	for (i = 1; i < _argc; i++)
	{
		parm = _argv[i];

		// skip - / \ etc.. in front of parm
		while (!isalpha(*parm))
			if (!*parm++)
				break;

		if (!strcmp(check, parm))
			return true;
	}

	return false;
}

//===========================================================================

static  int     wolfdigimap[] =
		{
			// These first sounds are in the upload version
#ifndef SPEAR
			HALTSND,                0,
			DOGBARKSND,             1,
			CLOSEDOORSND,           2,
			OPENDOORSND,            3,
			ATKMACHINEGUNSND,       4,
			ATKPISTOLSND,           5,
			ATKGATLINGSND,          6,
			SCHUTZADSND,            7,
			GUTENTAGSND,            8,
			MUTTISND,               9,
			BOSSFIRESND,            10,
			SSFIRESND,              11,
			DEATHSCREAM1SND,        12,
			DEATHSCREAM2SND,        13,
			DEATHSCREAM3SND,        13,
			TAKEDAMAGESND,          14,
			PUSHWALLSND,            15,

			LEBENSND,               20,
			NAZIFIRESND,            21,
			SLURPIESND,             22,

			YEAHSND,				32,

#ifndef UPLOAD
			// These are in all other episodes
			DOGDEATHSND,            16,
			AHHHGSND,               17,
			DIESND,                 18,
			EVASND,                 19,

			TOT_HUNDSND,            23,
			MEINGOTTSND,            24,
			SCHABBSHASND,           25,
			HITLERHASND,            26,
			SPIONSND,               27,
			NEINSOVASSND,           28,
			DOGATTACKSND,           29,
			LEVELDONESND,           30,
			MECHSTEPSND,			31,

			SCHEISTSND,				33,
			DEATHSCREAM4SND,		34,	// AIIEEE
			DEATHSCREAM5SND,		35,	// DEE-DEE
			DONNERSND,				36,	// EPISODE 4 BOSS DIE
			EINESND,				37,	// EPISODE 4 BOSS SIGHTING
			ERLAUBENSND,			38,	// EPISODE 6 BOSS SIGHTING
			DEATHSCREAM6SND,		39,	// FART
			DEATHSCREAM7SND,		40,	// GASP
			DEATHSCREAM8SND,		41,	// GUH-BOY!
			DEATHSCREAM9SND,		42,	// AH GEEZ!
			KEINSND,				43,	// EPISODE 5 BOSS SIGHTING
			MEINSND,				44,	// EPISODE 6 BOSS DIE
			ROSESND,				45,	// EPISODE 5 BOSS DIE

#endif
#else
//
// SPEAR OF DESTINY DIGISOUNDS
//
			HALTSND,                0,
			CLOSEDOORSND,           2,
			OPENDOORSND,            3,
			ATKMACHINEGUNSND,       4,
			ATKPISTOLSND,           5,
			ATKGATLINGSND,          6,
			SCHUTZADSND,            7,
			BOSSFIRESND,            8,
			SSFIRESND,              9,
			DEATHSCREAM1SND,        10,
			DEATHSCREAM2SND,        11,
			TAKEDAMAGESND,          12,
			PUSHWALLSND,            13,
			AHHHGSND,               15,
			LEBENSND,               16,
			NAZIFIRESND,            17,
			SLURPIESND,             18,
			LEVELDONESND,           22,
			DEATHSCREAM4SND,		23,	// AIIEEE
			DEATHSCREAM3SND,        23,	// DOUBLY-MAPPED!!!
			DEATHSCREAM5SND,		24,	// DEE-DEE
			DEATHSCREAM6SND,		25,	// FART
			DEATHSCREAM7SND,		26,	// GASP
			DEATHSCREAM8SND,		27,	// GUH-BOY!
			DEATHSCREAM9SND,		28,	// AH GEEZ!
			GETGATLINGSND,			38,	// Got Gat replacement

#ifndef SPEARDEMO
			DOGBARKSND,             1,
			DOGDEATHSND,            14,
			SPIONSND,               19,
			NEINSOVASSND,           20,
			DOGATTACKSND,           21,
			TRANSSIGHTSND,			29,	// Trans Sight
			TRANSDEATHSND,			30,	// Trans Death
			WILHELMSIGHTSND,		31,	// Wilhelm Sight
			WILHELMDEATHSND,		32,	// Wilhelm Death
			UBERDEATHSND,			33,	// Uber Death
			KNIGHTSIGHTSND,			34,	// Death Knight Sight
			KNIGHTDEATHSND,			35,	// Death Knight Death
			ANGELSIGHTSND,			36,	// Angel Sight
			ANGELDEATHSND,			37,	// Angel Death
			GETSPEARSND,			39,	// Got Spear replacement
#endif
#endif
			LASTSOUND
};

void InitDigiMap()
{
	int *map;

	for (map = wolfdigimap; *map != LASTSOUND; map += 2)
		DigiMap[map[0]] = map[1];
}

/*
#ifndef SPEAR
CP_iteminfo	MusicItems={CTL_X,CTL_Y,6,0,32};
CP_itemtype MusicMenu[]=
	{
		{1,"Get Them!",0},
		{1,"Searching",0},
		{1,"P.O.W.",0},
		{1,"Suspense",0},
		{1,"War March",0},
		{1,"Around The Corner!",0},

		{1,"Nazi Anthem",0},
		{1,"Lurking...",0},
		{1,"Going After Hitler",0},
		{1,"Pounding Headache",0},
		{1,"Into the Dungeons",0},
		{1,"Ultimate Conquest",0},

		{1,"Kill the S.O.B.",0},
		{1,"The Nazi Rap",0},
		{1,"Twelfth Hour",0},
		{1,"Zero Hour",0},
		{1,"Ultimate Conquest",0},
		{1,"Wolfpack",0}
	};
#else
CP_iteminfo MusicItems={CTL_X,CTL_Y-20,9,0,32};
CP_itemtype MusicMenu[]=
   {
		{1,"Funky Colonel Bill",0},
		{1,"Death To The Nazis",0},
		{1,"Tiptoeing Around",0},
		{1,"Is This THE END?",0},
		{1,"Evil Incarnate",0},
		{1,"Jazzin' Them Nazis",0},
		{1,"Puttin' It To The Enemy",0},
		{1,"The SS Gonna Get You",0},
		{1,"Towering Above",0}
	};
#endif

#ifndef SPEARDEMO
void DoJukebox(void)
{
	int which,lastsong=-1;
	unsigned short start,songs[]=
		{
#ifndef SPEAR
			GETTHEM_MUS,
			SEARCHN_MUS,
			POW_MUS,
			SUSPENSE_MUS,
			WARMARCH_MUS,
			CORNER_MUS,

			NAZI_OMI_MUS,
			PREGNANT_MUS,
			GOINGAFT_MUS,
			HEADACHE_MUS,
			DUNGEON_MUS,
			ULTIMATE_MUS,

			INTROCW3_MUS,
			NAZI_RAP_MUS,
			TWELFTH_MUS,
			ZEROHOUR_MUS,
			ULTIMATE_MUS,
			PACMAN_MUS
#else
			XFUNKIE_MUS,             // 0
			XDEATH_MUS,              // 2
			XTIPTOE_MUS,             // 4
			XTHEEND_MUS,             // 7
			XEVIL_MUS,               // 17
			XJAZNAZI_MUS,            // 18
			XPUTIT_MUS,              // 21
			XGETYOU_MUS,             // 22
			XTOWER2_MUS              // 23
#endif
		};
// LZ
//	struct dostime_t time;



	IN_ClearKeysDown();
	if (!AdLibPresent && !SoundBlasterPresent)
		return;


	MenuFadeOut();

#ifndef SPEAR
#ifndef UPLOAD
// LZ
//	_dos_gettime(&time);
//	start = (time.hsecond%3)*6;
start = 0;
#else
	start = 0;
#endif
#else
	start = 0;
#endif


	CA_CacheGrChunk (STARTFONT+1);
#ifdef SPEAR
	CacheLump (BACKDROP_LUMP_START,BACKDROP_LUMP_END);
#else
	CacheLump (CONTROLS_LUMP_START,CONTROLS_LUMP_END);
#endif
	CA_LoadAllSounds ();

	fontnumber=1;
	ClearMScreen ();
	VWB_DrawPic(112,184,C_MOUSELBACKPIC);
	DrawStripes (10);
	SETFONTCOLOR (TEXTCOLOR,BKGDCOLOR);

#ifndef SPEAR
	DrawWindow (CTL_X-2,CTL_Y-6,280,13*7,BKGDCOLOR);
#else
	DrawWindow (CTL_X-2,CTL_Y-26,280,13*10,BKGDCOLOR);
#endif

	DrawMenu (&MusicItems,&MusicMenu[start]);

	SETFONTCOLOR (READHCOLOR,BKGDCOLOR);
	PrintY=15;
	WindowX = 0;
	WindowY = 320;
	US_CPrint ("Robert's Jukebox");

	SETFONTCOLOR (TEXTCOLOR,BKGDCOLOR);
	VW_UpdateScreen();
	MenuFadeIn();

	do
	{
		which = HandleMenu(&MusicItems,&MusicMenu[start],NULL);
		if (which>=0)
		{
			if (lastsong >= 0)
				MusicMenu[start+lastsong].active = 1;

			StartCPMusic(songs[start + which]);
			MusicMenu[start+which].active = 2;
			DrawMenu (&MusicItems,&MusicMenu[start]);
			VW_UpdateScreen();
			lastsong = which;
		}
	} while(which>=0);

	MenuFadeOut();
	IN_ClearKeysDown();
#ifdef SPEAR
	UnCacheLump (BACKDROP_LUMP_START,BACKDROP_LUMP_END);
#else
	UnCacheLump (CONTROLS_LUMP_START,CONTROLS_LUMP_END);
#endif
}
#endif
*/

//	Load a few things right away
void InitGame()
{
	int i;

	S_Startup();

	// so the signon screen can be freed
	MM_Startup();
	SignonScreen();

	VL_Startup();
	IN_Startup();
	PM_Startup();
	PM_UnlockMainMem();
	SD_Startup();
	CA_Startup();
	US_Startup();
	GL_Startup();
	Con_Init();

	// build some tables
	InitDigiMap();

	for (i = 0; i < MAPSIZE; i++)
	{
		nearmapylookup[i] = &tilemap[0][0]+MAPSIZE*i;
		farmapylookup[i] = i*64;
	}

	ReadConfig();

/*
	// HOLDING DOWN 'M' KEY?
#ifndef SPEARDEMO
	if (Keyboard[sc_M])
	  DoJukebox();
	else
#endif
*/
	// draw intro screen stuff
	IntroScreen();

	// load in and lock down some basic chunks
	CA_CacheGrChunk(STARTFONT);
	MM_SetLock (&grsegs[STARTFONT],true);

	LoadLatchMem();
	BuildTables();	// trig tables
	SetupWalls();

	NewViewSize(viewsize);

/*	// initialize variables
	FinishSignon();
*/
}

//===========================================================================

wbool SetViewSize (unsigned short width, unsigned short height)
{
	viewwidth = width&~15;	// must be divisable by 16
	viewheight = height&~1;	// must be even
	centerx = viewwidth/2-1;
	shootdelta = viewwidth/10;

	// calculate trace angles and projection constants
	CalcProjection (FOCALLENGTH);

	return true;
}

void NewViewSize (int width)
{
	CA_UpLevel();
	MM_SortMem();
	viewsize = width;
	SetViewSize(width*16,width*16*HEIGHTRATIO);
	CA_DownLevel();
}

//===========================================================================

void Quit(char *error)
{
/*
	unsigned short finscreen;
	memptr screen;
*/
	ClearMemory();

	if (error == 0)
	{
/*		#ifndef JAPAN
		CA_CacheGrChunk(ORDERSCREEN);
		screen = grsegs[ORDERSCREEN];
		#endif
*/		WriteConfig();
	}
	else
	{
	  printf(error);
/*		CA_CacheGrChunk(ERRORSCREEN);
		screen = grsegs[ERRORSCREEN];
*/	}

	ShutdownId();
/*
	if (error && *error)
	{
	  movedata ((unsigned short)screen,7,0xb800,0,7*160);
	  gotoxy (10,4);
	  puts(error);
	  gotoxy (1,8);
	  exit(1);
	}
	else if (!error || !(*error))
	{
		clrscr();
		#ifndef JAPAN
		movedata ((unsigned short)screen,7,0xb800,0,4000);
		gotoxy(1,24);
		#endif
	}
*/
	exit(0);
}

//===========================================================================

static char *ParmStrings[] = {"baby","easy","normal","hard",""};

void DemoLoop()
{
	static int LastDemo;
	int i,level;

	// check for launch from ted
	if (tedlevel)
	{
		NoWait = true;
		NewGame(1,0);

		for (i = 1; i < _argc; i++)
		{
			if ((level = US_CheckParm(_argv[i], ParmStrings)) != -1)
			{
				gamestate.difficulty=level;
				break;
			}
		}

#ifndef SPEAR
		gamestate.episode = tedlevelnum/10;
		gamestate.mapon = tedlevelnum%10;
#else
		gamestate.episode = 0;
		gamestate.mapon = tedlevelnum;
#endif
		GameLoop();
		Quit (NULL);
	}

	// main game cycle

#ifndef DEMOTEST

	#ifndef UPLOAD

		#ifndef GOODTIMES
		#ifndef SPEAR
		#ifndef JAPAN
		if (!NoWait)
			NonShareware();
		#endif
		#else

			#ifndef GOODTIMES
			#ifndef SPEARDEMO
			CopyProtection();
			#endif
			#endif

		#endif
		#endif
	#endif

	StartCPMusic(INTROSONG);

#ifndef JAPAN
	if (!NoWait)
		PG13();
#endif

#endif


	while (1)
	{
		while (!NoWait)
		{
			// title page
			MM_SortMem();
#ifndef DEMOTEST

#ifdef SPEAR
			GL_CacheScreen(TITLE1PIC, TITLEPALETTE);
			GL_FadeIn(0, 0, 0, 30, GL_DrawScreen);
			GL_DrawScreen();
			S_FinishDisplay();

			if (IN_UserInput(TickBase*15))
			{
				GL_FadeOut(0, 0, 0, 30, GL_DrawScreen);
				GL_FreeScreen();
				break;
			}
			GL_FadeOut(0, 0, 0, 30, GL_DrawScreen);
			GL_FreeScreen();
#else
			GL_CacheScreen(TITLEPIC, -1);
			GL_FadeIn(0, 0, 0, 30, GL_DrawScreen);
			GL_DrawScreen();
			S_FinishDisplay();

			if (IN_UserInput(TickBase*10))
			{
				GL_FadeOut(0, 0, 0, 30, GL_DrawScreen);
				GL_FreeScreen();
				break;
			}
			GL_FadeOut(0, 0, 0, 30, GL_DrawScreen);
			GL_FreeScreen();
#endif

			// credits page
			GL_CacheScreen(CREDITSPIC, -1);
			GL_FadeIn(0, 0, 0, 30, GL_DrawScreen);
			GL_DrawScreen();
			S_FinishDisplay();

			if (IN_UserInput(TickBase*10))
			{
				GL_FadeOut(0, 0, 0, 30, GL_DrawScreen);
				GL_FreeScreen();
				break;
			}
			GL_FadeOut(0, 0, 0, 30, GL_DrawScreen);
			GL_FreeScreen();

			// high scores
#ifndef SPEAR
			GL_CacheTexture(C_NAMEPIC);
			GL_CacheTexture(C_LEVELPIC);
			GL_CacheTexture(C_SCOREPIC);
#endif
			LoadFonts();

			GL_FadeIn(0, 0, 0, 30, DrawHighScores);
			DrawHighScores();
			S_FinishDisplay();

			if (IN_UserInput(TickBase*10))
			{
				GL_FadeOut(0, 0, 0, 30, DrawHighScores);
#ifndef SPEAR
				GL_UnCacheTexture(C_NAMEPIC);
				GL_UnCacheTexture(C_LEVELPIC);
				GL_UnCacheTexture(C_SCOREPIC);
#endif
				break;
			}
			GL_FadeOut(0, 0, 0, 30, DrawHighScores);
#ifndef SPEAR
			GL_UnCacheTexture(C_NAMEPIC);
			GL_UnCacheTexture(C_LEVELPIC);
			GL_UnCacheTexture(C_SCOREPIC);
#endif
#endif

			// demo
			#ifndef SPEARDEMO
			PlayDemo(LastDemo++%4);
			#else
			PlayDemo(0);
			#endif

			if (playstate == ex_abort)
				break;
			StartCPMusic(INTROSONG);
		}

/*		VW_FadeOut();
*/

#ifndef SPEAR
		if (Keyboard[sc_Tab] && MS_CheckParm("goobers"))
#else
		if (Keyboard[sc_Tab] && MS_CheckParm("debugmode"))
#endif
			RecordDemo ();
		else
			US_ControlPanel (0);

		if (startgame || loadedgame)
		{
			GameLoop();
/*			VW_FadeOut();
*/			StartCPMusic(INTROSONG);
		}
	}
}

//===========================================================================

int main(int argc, char* argv[])
{
	int i;

	_argc = argc;
	_argv = argv;

	for (i = 0; i < 768; i++)
		gamepal[i] *=4 ;

	chdir(DATADIR);
	CheckForEpisodes();
	InitGame();
	DemoLoop();
	Quit("Demo loop exited???");

	return 0;
}





