//
// Sound stuff for the linux port
//
// The mixing code was stolen from linuxdoom,
// I've just moved everything to a second thread
//

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/time.h>

#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <pthread.h>

// Linux voxware output.
#include <linux/soundcard.h>

#include <signal.h>

#include "wl_def.h"

#define MAXSNDS 256

typedef struct
{
  int length;
  void *data;
} SNDBUFFER;

SNDBUFFER snds[MAXSNDS];

static pthread_t sound_thread;
static int sound_alive = 0;

// How much do we have in the buffer
int audioleft;

// Update all 30 millisecs, approx. 30fps synchronized.
// Linux resolution is allegedly 10 millisecs,
//  scale is microseconds. LZ: was 500
#define SOUND_INTERVAL     500

// Get the interrupt. Set duration in millisecs.
int I_SoundSetTimer(int duration_of_tick);
void I_SoundDelTimer();

void I_SetChannels();

// The number of internal mixing channels,
//  the samples calculated for each mixing step,
//  the size of the 16bit, 2 hardware channel (stereo)
//  mixing buffer, and the samplerate of the raw data.


// Needed for calling the actual sound output.
#define SAMPLECOUNT	        1024
#define NUM_CHANNELS		4
// It is 2 for 16bit, and 2 for two channels.
#define BUFMUL                  4
#define MIXBUFFERSIZE		(SAMPLECOUNT*BUFMUL)

#define SAMPLERATE		11025	// Hz
#define SAMPLESIZE		2   	// 16bit

// The actual output device.
int	audio_fd;

// The global mixing buffer.
// Basically, samples from all active internal channels
//  are modifed and added, and stored in the buffer
//  that is submitted to the audio device.
signed short	mixbuffer[MIXBUFFERSIZE];

// The channel data pointers, start and end.
unsigned char*	channels[NUM_CHANNELS];
unsigned char*	channelsend[NUM_CHANNELS];

// Time/gametic that the channel started playing,
//  used to determine oldest, which automatically
//  has lowest priority.
// In case number of active sounds exceeds
//  available channels.
int		channelstart[NUM_CHANNELS];

// Volume lookups.
int		vol_lookup[128*256];

// Hardware left and right channel volume lookup.
int*		channelleftvol_lookup[NUM_CHANNELS];
int*		channelrightvol_lookup[NUM_CHANNELS];


int DigitalDevice;


// Safe ioctl, convenience.
static void myioctl(int fd, int command, int* arg)
{   
    int		rc;
    
    rc = ioctl(fd, command, arg);  
    if (rc < 0)
    {
	fprintf(stderr, "ioctl(dsp,%d,arg) failed\n", command);
	fprintf(stderr, "errno=%d\n", errno);
	exit(-1);
    }
}

//============================================================

int prebuffer = 0;
int playing = 0;
void UpdateSounds();

void* sound_loop(void *arg)
{
  while (sound_alive)
  {
    UpdateSounds();
    // Feed sound device if necesary.
    if (audioleft && !prebuffer)
    {
      playing = 1;
      //      printf("%d\n", audioleft);

      // Write it to DSP device.
      write(audio_fd, mixbuffer, audioleft*BUFMUL);
      audioleft = 0;
      playing = 0;
    }
    else
      usleep(10000);
  }

  pthread_exit(NULL);
}

//============================================================

void S_InitSound()
{
  int i;

  //  fprintf(stderr, "S_SoundSetTimer: %d microsecs\n", SOUND_INTERVAL);
  //  I_SoundSetTimer( SOUND_INTERVAL );

  // Secure and configure sound device first.
  audio_fd = open("/dev/dsp", O_WRONLY);
  if (audio_fd<0)
  {
    fprintf(stderr, "Could not open /dev/dsp\n");
    DigitalDevice = 0;
	return;
  }
  else
    DigitalDevice = 1;
                     
  i = 11 | (2<<16);                                           
  myioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &i);
  myioctl(audio_fd, SNDCTL_DSP_RESET, 0);
  
  i=SAMPLERATE;
  
  myioctl(audio_fd, SNDCTL_DSP_SPEED, &i);
  
  i=1;
  myioctl(audio_fd, SNDCTL_DSP_STEREO, &i);
  
  myioctl(audio_fd, SNDCTL_DSP_GETFMTS, &i);

  if (i&=AFMT_S16_LE)    
    myioctl(audio_fd, SNDCTL_DSP_SETFMT, &i);
  else
    fprintf(stderr, "Could not play signed 16 data\n");

  fprintf(stderr, "configured audio device\n" );
    
  // Now initialize mixbuffer with zero.
  for ( i = 0; i< MIXBUFFERSIZE; i++ )
    mixbuffer[i] = 0;

  // Whatever these did with DMX, these are rather dummies now.
  I_SetChannels();
  
  // Finished initialization.
  printf("sound module ready\n");
    
  for (i = 0; i < MAXSNDS; i++)
    snds[i].data = 0;

  sound_alive = 1;
  pthread_create(&sound_thread, NULL, sound_loop, NULL);
}

void S_ShutdownSound()
{
  int i;

  sound_alive = 0;
  pthread_join(sound_thread, NULL);   

  I_SoundDelTimer();
  
  // Cleaning up -releasing the DSP device.
  close(audio_fd);

  for (i = 0; i < MAXSNDS; i++)
    if (snds[i].data)
      free(snds[i].data);
}

//============================================================

// Starting a sound means adding it
//  to the current list of active sounds
//  in the internal channels.
// As the SFX info struct contains
//  e.g. a pointer to the raw data,
//  it is ignored.
void S_PlaySound(int id, int pan)
{
  static unsigned short handlenums = 0;

  int		i;
  int		rc = -1;
    
  int		oldest = TicCount;
  int		oldestnum = 0;
  int		slot;

  int		rightvol;
  int		leftvol;

  int seperation = (pan+15)*254/30; // 127
int volume = 64;

    // Loop all channels to find oldest SFX.
    for (i = 0; (i < NUM_CHANNELS) && (channels[i]); i++)
    {
	if (channelstart[i] < oldest)
	{
	    oldestnum = i;
	    oldest = channelstart[i];
	}
    }

    // Tales from the cryptic.
    // If we found a channel, fine.
    // If not, we simply overwrite the first one, 0.
    // Probably only happens at startup.
    if (i == NUM_CHANNELS)
	slot = oldestnum;
    else
	slot = i;

    // Okay, in the less recent channel,
    //  we will handle the new SFX.
    // Set pointer to raw data.
    channels[slot] = (unsigned char*) snds[id].data;
    // Set pointer to end of raw data.
    channelsend[slot] = channels[slot] + snds[id].length;

    // Reset current handle number, limited to 0..100.
    if (!handlenums)
	handlenums = 100;
/*
    // Set stepping???
    // Kinda getting the impression this is never used.
    channelstep[slot] = step;
    // ???
    channelstepremainder[slot] = 0;
*/    // Should be gametic, I presume.
    channelstart[slot] = TicCount;

    // Separation, that is, orientation/stereo.
    //  range is: 1 - 256
    seperation += 1;

    // Per left/right channel.
    //  x^2 seperation,
    //  adjust volume properly.
    leftvol =
	volume - ((volume*seperation*seperation) >> 16); ///(256*256);
    seperation = seperation - 257;
    rightvol =
	volume - ((volume*seperation*seperation) >> 16);	

    // Sanity check, clamp volume.
    if (rightvol < 0 || rightvol > 127)
	Quit("rightvol out of bounds");
    
    if (leftvol < 0 || leftvol > 127)
	Quit("leftvol out of bounds");

    // Get the proper lookup table piece
    //  for this volume level???
    channelleftvol_lookup[slot] = &vol_lookup[leftvol*256];
    channelrightvol_lookup[slot] = &vol_lookup[rightvol*256];
}

// This function loops all active (internal) sound
//  channels, retrieves a given number of samples
//  from the raw sound data, modifies it according
//  to the current (internal) channel parameters,
//  mixes the per channel samples into the global
//  mixbuffer, clamping it to the allowed range,
//  and sets up everything for transferring the
//  contents of the mixbuffer to the (two)
//  hardware channels (left and right, that is).
//
// This function currently supports only 16bit.
void S_UpdateSound()
{
}

void UpdateSounds()
{
  // Mix current sound data.
  // Data, from raw sound, for right and left.
  register unsigned int	sample;
  register int		dl;
  register int		dr;
  
  // Pointers in global mixbuffer, left, right, end.
  signed short*		leftout;
  signed short*		rightout;
  signed short*		leftend;
  // Step in mixbuffer, left and right, thus two.
  int				step;

  // Mixing channel index.
  int				chan;
  int added;

  if (playing)
    return;

  // Left and right channel
  //  are in global mixbuffer, alternating.
  leftout = mixbuffer;
  rightout = mixbuffer+1;
  step = 2;

  // Determine end, for left channel only
  //  (right channel is implicit).
  leftend = mixbuffer + SAMPLECOUNT*step;

  prebuffer = 1;

  // Mix sounds into the mixing buffer.
  // Loop over step*SAMPLECOUNT,
  //  that is 512 values for two channels.
  while (leftout != leftend)
  {
	// Reset left/right value. 
	dl = 0;
	dr = 0;
	added = 0;

	// Love thy L2 chache - made this a loop.
	// Now more channels could be set at compile time
	//  as well. Thus loop those  channels.
	for ( chan = 0; chan < NUM_CHANNELS; chan++ )
	{
	    // Check channel, if active.
	    if (channels[ chan ])
	    {
	      added++;
		// Get the raw data from the channel. 
		sample = *channels[ chan ];
		// Add left and right part
		//  for this channel (sound)
		//  to the current data.
		// Adjust volume accordingly.
		dl += channelleftvol_lookup[ chan ][sample];
		dr += channelrightvol_lookup[ chan ][sample];

		// Check whether we are done.
		channels[ chan ]++;
		if (channels[ chan ] >= channelsend[ chan ])
		    channels[ chan ] = 0;
	    }
	}

	if (!added)
	  break;
	
	// Clamp to range. Left hardware channel.
	// Has been char instead of short.
	// if (dl > 127) *leftout = 127;
	// else if (dl < -128) *leftout = -128;
	// else *leftout = dl;

	if (dl > 0x7fff)
	    *leftout = 0x7fff;
	else if (dl < -0x8000)
	    *leftout = -0x8000;
	else
	    *leftout = dl;

	// Same for right hardware channel.
	if (dr > 0x7fff)
	    *rightout = 0x7fff;
	else if (dr < -0x8000)
	    *rightout = -0x8000;
	else
	    *rightout = dr;

	// Increment current pointers in mixbuffer.
	leftout += step;
	rightout += step;
	audioleft++;
    }

  prebuffer = 0;

    if (audioleft > MIXBUFFERSIZE)
      fprintf(stderr, "I_SoundUpdate: missed some buffer writes\n");
}

void S_LoadSound(int id, void* data, long length)
{
  unsigned char* paddedsfx;
  int i, paddedsize;

  if (snds[id].data == 0)
  {
    // Pads the sound effect out to the mixing buffer size.
    // The original realloc would interfere with zone memory.
    paddedsize = ((length-8 + (SAMPLECOUNT-1)) / SAMPLECOUNT) * SAMPLECOUNT;

    snds[id].data = paddedsfx = (unsigned char*)malloc(paddedsize+8);

    // Now copy and pad.
    memcpy(paddedsfx, data, length);
    for (i = length; i < paddedsize+8; i++)
        paddedsfx[i] = 128;

    // Preserve padded length.
    snds[id].length = paddedsize;
  }
}

// SFX API
// Note: this was called by S_Init.
// However, whatever they did in the
// old DPMS based DOS version, this
// were simply dummies in the Linux
// version.
// See soundserver initdata().
void I_SetChannels()
{
  // Init internal lookups (raw data, mixing buffer, channels).
  // This function sets up internal lookups used during
  //  the mixing process. 
  int i, j;

  // Okay, reset internal mixing channels to zero.
  for (i=0; i<NUM_CHANNELS; i++)
    channels[i] = 0;
  audioleft = 0;

  // Generates volume lookup tables
  //  which also turn the unsigned samples
  //  into signed samples.
  for (i=0 ; i<128 ; i++)
    for (j=0 ; j<256 ; j++)
      vol_lookup[i*256+j] = (i*(j-128)*256)/127;
}	

//
// Experimental stuff.
// A Linux timer interrupt, for asynchronous
//  sound output.
// I ripped this out of the Timer class in
//  our Difference Engine, including a few
//  SUN remains...
//  
#ifdef sun
    typedef     sigset_t        tSigSet;
#else    
    typedef     int             tSigSet;
#endif


// We might use SIGVTALRM and ITIMER_VIRTUAL, if the process
//  time independend timer happens to get lost due to heavy load.
// SIGALRM and ITIMER_REAL doesn't really work well.
// There are issues with profiling as well.
static int /*__itimer_which*/  itimer = ITIMER_REAL;

static int sig = SIGALRM;

// Interrupt handler.
void I_HandleSoundTimer( int ignore )
{
  // Feed sound device if necesary.
  if (audioleft && !prebuffer)
  {
    playing = 1;
    printf("%d\n", audioleft);
    // Write it to DSP device.
    write(audio_fd, mixbuffer, audioleft*BUFMUL);
    audioleft = 0;
    playing = 0;
  }
}

// Get the interrupt. Set duration in millisecs.
int I_SoundSetTimer( int duration_of_tick )
{
  // Needed for gametick clockwork.
  struct itimerval    value;
  struct itimerval    ovalue;
  struct sigaction    act;
  struct sigaction    oact;

  int res;

  // This sets to SA_ONESHOT and SA_NOMASK, thus we can not use it.
  //     signal( _sig, handle_SIG_TICK );
  
  // Now we have to change this attribute for repeated calls.
  act.sa_handler = I_HandleSoundTimer;
#ifndef sun    
  //ac	t.sa_mask = _sig;
#endif
  act.sa_flags = SA_RESTART;
  
  sigaction( sig, &act, &oact );

  value.it_interval.tv_sec    = 0;
  value.it_interval.tv_usec   = duration_of_tick;
  value.it_value.tv_sec       = 0;
  value.it_value.tv_usec      = duration_of_tick;

  // Error is -1.
  res = setitimer( itimer, &value, &ovalue );

  // Debug.
  if ( res == -1 )
    fprintf( stderr, "I_SoundSetTimer: interrupt n.a.\n");
  
  return res;
}


// Remove the interrupt. Set duration to zero.
void I_SoundDelTimer()
{
  // Debug.
  if ( I_SoundSetTimer( 0 ) == -1)
    fprintf( stderr, "I_SoundDelTimer: failed to remove interrupt. Doh!\n");
}







