#include <stdio.h>
#include <sys/time.h>
#include <X11/keysym.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include "wl_def.h"

int screen_width, screen_height;
Display *dpy;
Window win;

static int convertkey(int key)
{
  int rc;

  switch(key)
  {
  case XK_Left:      rc = sc_LeftArrow;     break;
  case XK_Right:     rc = sc_RightArrow;    break;
  case XK_Down:      rc = sc_DownArrow;     break;
  case XK_Up:        rc = sc_UpArrow;       break;
  case XK_Escape:    rc = sc_Escape;        break;
  case XK_Return:    rc = sc_Return;        break;
  case XK_Tab:       rc = sc_Tab;           break;
  case XK_F1:        rc = sc_F1;            break;
  case XK_F2:        rc = sc_F2;            break;
  case XK_F3:        rc = sc_F3;            break;
  case XK_F4:        rc = sc_F4;            break;
  case XK_F5:        rc = sc_F5;            break;
  case XK_F6:        rc = sc_F6;            break;
  case XK_F7:        rc = sc_F7;            break;
  case XK_F8:        rc = sc_F8;            break;
  case XK_F9:        rc = sc_F9;            break;
  case XK_F10:       rc = sc_F10;           break;
  case XK_F11:       rc = sc_F11;           break;
  case XK_F12:       rc = sc_F12;           break;
  case XK_BackSpace: rc = sc_BackSpace;     break;
  case XK_Delete:    rc = sc_Delete;        break;

  case XK_space:     rc = sc_Space;            break;
  case XK_Caps_Lock: rc = sc_CapsLock;         break;
  case XK_Insert:    rc = sc_Insert;           break;
  case XK_Home:      rc = sc_Home;             break;
  case XK_End:       rc = sc_End;              break;
  case XK_Prior:     rc = sc_PgUp;             break;
  case XK_Next:      rc = sc_PgDn;             break;

  case XK_Shift_L:   rc = sc_LShift;           break;
  case XK_Shift_R:   rc = sc_RShift;           break;
  case XK_Control_L: rc = sc_Control;          break;
  case XK_Control_R: rc = sc_Control;          break;
  case XK_Alt_L:     rc = sc_Alt;              break;
  case XK_Meta_L:    rc = sc_Alt;              break;
  case XK_Alt_R:     rc = sc_Alt;              break;
  case XK_Meta_R:    rc = sc_Alt;              break;

  case XK_A: case XK_a: rc = sc_A;             break;
  case XK_B: case XK_b: rc = sc_B;             break;
  case XK_C: case XK_c: rc = sc_C;             break;
  case XK_D: case XK_d: rc = sc_D;             break;
  case XK_E: case XK_e: rc = sc_E;             break;
  case XK_F: case XK_f: rc = sc_F;             break;
  case XK_G: case XK_g: rc = sc_G;             break;
  case XK_H: case XK_h: rc = sc_H;             break;
  case XK_I: case XK_i: rc = sc_I;             break;
  case XK_J: case XK_j: rc = sc_J;             break;
  case XK_K: case XK_k: rc = sc_K;             break;
  case XK_L: case XK_l: rc = sc_L;             break;
  case XK_M: case XK_m: rc = sc_M;             break;
  case XK_N: case XK_n: rc = sc_N;             break;
  case XK_O: case XK_o: rc = sc_O;             break;
  case XK_P: case XK_p: rc = sc_P;             break;
  case XK_Q: case XK_q: rc = sc_Q;             break;
  case XK_R: case XK_r: rc = sc_R;             break;
  case XK_S: case XK_s: rc = sc_S;             break;
  case XK_T: case XK_t: rc = sc_T;             break;
  case XK_U: case XK_u: rc = sc_U;             break;
  case XK_V: case XK_v: rc = sc_V;             break;
  case XK_W: case XK_w: rc = sc_W;             break;
  case XK_X: case XK_x: rc = sc_X;             break;
  case XK_Y: case XK_y: rc = sc_Y;             break;
  case XK_Z: case XK_z: rc = sc_Z;             break;

  case XK_0: rc = sc_0;                       break;
  case XK_1: rc = sc_1;                       break;
  case XK_2: rc = sc_2;                       break;
  case XK_3: rc = sc_3;                       break;
  case XK_4: rc = sc_4;                       break;
  case XK_5: rc = sc_5;                       break;
  case XK_6: rc = sc_6;                       break;
  case XK_7: rc = sc_7;                       break;
  case XK_8: rc = sc_8;                       break;
  case XK_9: rc = sc_9;                       break;

  default:   rc = 0;                          break;
  }

    return rc; 
}

static Window make_rgb_db_window(unsigned int width, unsigned int height)
{
   int attrib[] = { GLX_RGBA,
                    GLX_DOUBLEBUFFER,
		    GLX_DEPTH_SIZE, 16,
                    None };
   int scrnum;
   XSetWindowAttributes attr;
   unsigned long mask;
   Window root;
   GLXContext ctx;
   XVisualInfo *visinfo;

   scrnum = DefaultScreen( dpy );
   root = RootWindow( dpy, scrnum );

   visinfo = glXChooseVisual( dpy, scrnum, attrib );
   if (!visinfo) {
      printf("Error: couldn't get an RGB, Double-buffered visual\n");
      exit(1);
   }

   /* window attributes */
   attr.background_pixel = 0;
   attr.border_pixel = 0;
   attr.colormap = XCreateColormap( dpy, root, visinfo->visual, AllocNone);
   attr.event_mask = StructureNotifyMask | ExposureMask | KeyPressMask | KeyReleaseMask;
   mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;

   win = XCreateWindow( dpy, root, 0, 0, width, height,
                        0, visinfo->depth, InputOutput,
                        visinfo->visual, mask, &attr );
   XStoreName(dpy, win, "WolfGL");
   ctx = glXCreateContext( dpy, visinfo, NULL, True );

   glXMakeCurrent( dpy, win, ctx );

   return win;
}

void S_Startup()
{
  char* parm;
  int i, tmp, width = 400, height = 300;
  extern int _argc;
  extern char** _argv;

  printf("WolfGL startup\n");

  dpy = XOpenDisplay(NULL);

  for (i = 1; i < _argc; i++)
  {
    parm = _argv[i];

    // skip - / \ etc.. in front of parm
    while (!isalpha(*parm))
      if (!*parm++)
	break;

    if (strcmp("width", parm) == 0)
    {
      i++;
      parm = _argv[i];
      if (sscanf(parm, "%d", &tmp) == 1)
	width = tmp;
    }
    else if (strcmp("height", parm) == 0)
    {
      i++;
      parm = _argv[i];
      if (sscanf(parm, "%d", &tmp) == 1)
	height = tmp;
    }
  }
   
  win = make_rgb_db_window(width, height);
  XMapWindow( dpy, win );
  S_InitSound();
}

void S_Shutdown()
{
  S_ShutdownSound();
}

int S_FileExists(const char* name)
{
  FILE* f = fopen(name, "r");

  if (f != 0)
  {
    fclose (f);
    return 1;
  }
  else
    return 0;
}

unsigned long S_FileLength(FILE* file)
{
  unsigned long nLen, nCur;

  nCur = ftell(file);
  fseek(file, 0, SEEK_END);
  nLen = ftell(file);
  fseek(file, nCur, SEEK_SET);

  return nLen;
}

void S_ProcessEvents()
{
  XEvent event;
  int key;

  while (XPending(dpy))
  {
    XNextEvent(dpy, &event);

    switch (event.type)
    {
    case KeyPress:
      key = convertkey(XKeycodeToKeysym(dpy, event.xkey.keycode, 0));

      if (key)
	IN_KeyEvent(key);
      break;

    case KeyRelease:
      key = convertkey(XKeycodeToKeysym(dpy, event.xkey.keycode, 0));

      if (key)
      {
	key |= 0x80;
	IN_KeyEvent(key);
      }
      break;

    case Expose:
      //      redraw( dpy, event.xany.window );
      break;
    
    case ConfigureNotify:
      screen_width = event.xconfigure.width;
      screen_height = event.xconfigure.height;
      break;
    }
  }
}

int S_GetNewTics()
{
    struct timeval      tp;
    struct timezone     tzp;
    int                 newtics;
    static int          basetime=0;
    static int          lasttic;
    int ret = 0;

    gettimeofday(&tp, &tzp);
    if (!basetime)
        basetime = tp.tv_sec;
    newtics = (tp.tv_sec-basetime)*70 + tp.tv_usec*70/1000000;

    if (newtics != lasttic)
    {
      ret = newtics - lasttic;
      lasttic = newtics;
    }

    return ret;
}

void S_FinishDisplay()
{
/*
  static int basetime = 0;
  static int count;
  struct timeval      tp;
  struct timezone     tzp;

  gettimeofday(&tp, &tzp);
  if (basetime != tp.tv_sec)
  {
    basetime = tp.tv_sec;
    printf("%d\n", count);
    count = 0;
  }
  else
    count++;
*/
  glXSwapBuffers( dpy, win );
}





